#!/usr/bin/clisp -ansi -modern -q -q

;;;;--------------------------------------------------------------------------+
;;;; WP-MIRROR is a free utility for mirroring a set of Wikipedias.           |
;;;;--------------------------------------------------------------------------+


(defconstant *copyright-license-and-disclaimer*
"Copyright (C) 2011,2012 Dr. Kent L. Miller.  All rights reserved.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY\; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>")


(defconstant *version* 
"wp-mirror 0.3")


;;;;--------------------------------------------------------------------------+
;;;; Internationalization:                                                    |
;;;;--------------------------------------------------------------------------+


(setf (i18n:textdomaindir "wp-mirror") "./")
(defun _ (msgid) (i18n:gettext msgid "wp-mirror"))


;;;;--------------------------------------------------------------------------+
;;;; Packages:                                                                |
;;;;  asdf   - process other packages                                         |
;;;;  getopt - process command-line arguments                                 |
;;;;  md5    - compute md5sum                                                 |
;;;;--------------------------------------------------------------------------+

;;
;; Design note:
;;
;; It can happen that `clisp' is unable to find any packages.  
;; The error message will be similar to
;;
;; *** - read from #<input buffered file-stream character
;; #P"./wp-mirror-0.2.lisp" @47>: there is no package with name "ASDF"
;;
;; Most likely, the `cl-asdf' package was either not installed or not
;; configured.  Its documentation is probably under
;; `/usr/share/doc/cl-asdf/'.  Look for the file `README.Debian' or a
;; directory containing HTML pages.  They should have instructions for
;; configuring ASDF.  For Debian GNU/Linux (and its derivatives) the
;; next two lines should work.

(load "/usr/share/common-lisp/source/cl-asdf/asdf.lisp")
(push "/usr/share/common-lisp/systems/" asdf:*central-registry*)

;; If you are installing your own packages elsewhere, then provide
;; additional lines that push the relevant path(s) onto
;; `custom:*load-paths*'.  Note that a directory name must have a
;; trailing slash.

;;(push "/usr/share/common-lisp/systems/" custom:*load-paths*)

(setf assert-asdf
      (_ "
CATEGORY:    software prerequisite                                     
ASSERT:      assert-asdf                                               
ABSTRACT:    The installed version of `clisp' is missing key features. 
DESCRIPTION: Like all languages, `clisp' comes with extensions and     
             packages that offer features well beyond the minimum for  
             compliance with standards.  `wp-mirror' relies on:        
             (:asdf2 :asdf :clisp :clx :gettext :i18n :loop :regexp    
             :screen :syscalls)                                        
             This error message occurred because the `cl-asdf' package 
             has not been installed, or has not been configured.       
             Please run                                                
                                                                       
               (shell) # aptitude install cl-asdf                      
                                                                       
             Then read the documentation, which can usually be found   
             under `/usr/share/doc/cl-asdf/'.  Information on          
             configuration is usually found in the file `README.Debian'
             or in a directory containing HTML pages.  Configuration   
             may entail adding the following lines to `/root/.clisprc': 
                                                                       
             (load \"/usr/share/common-lisp/source/cl-asdf/asdf.lisp\")
             (push \"/usr/share/common-lisp/systems/\" asdf:*central-registry*)
                                                                       
GNU/LINUX:   To see the features your `clisp' installation offers, try:
               (shell) $ clisp -q                                      
               [1]> *features*                                         
REFERENCE:   See the `wp-mirror' README for more on installation.      
"))

(unless (member :asdf2 *features*)
  (error assert-asdf))
  
(asdf:operate 'asdf:compile-op :getopt :verbose nil)
(asdf:operate 'asdf:load-op    :getopt :verbose nil)
;;(asdf:operate 'asdf:test-op    :getopt)

;;; The md5 package runs far too slowly, so 
;;; o we use it only for short strings (such as file names),
;;; o we do not use it for file contents (we use shell commands instead).
(asdf:operate 'asdf:compile-op :md5    :verbose nil)
(asdf:operate 'asdf:load-op    :md5    :verbose nil)

;;; Debian packages: cl-sql, cl-sql-uffi, and cl-sql-mysql all seem
;;; broken, so for database use shell commands instead.


;;;;--------------------------------------------------------------------------+
;;;; Constants:                                                               |
;;;;   *author*                                                               |
;;;;   *config-default*                                                       |
;;;;   *config-local*                                                         |
;;;;   *copyright-license-and-disclaimer* (see above)                         |
;;;;   *help*                                                                 |
;;;;   *version* (see above)                                                  |
;;;;   *wtf*                                                                  |
;;;; The values of these constants are used by the build process.             |
;;;; The utility `help2man' generates the `wp-mirror.1.gz' man page out       |
;;;; of `--help' and `--version' output.  See also the functions              |
;;;; `put-help-and-die' and `put-version-and-die' below.                      |
;;;;--------------------------------------------------------------------------+


(defconstant *author*
"Written by Dr. Kent L. Miller.")  


(defconstant *help* 
"Wp-mirror is a free utility for building and maintaining a set of 
mirrors of Wikipedias (aka a Wikipedia farm).  The WikiMedia Foundation 
offers Wikipedias in nearly 300 languages.  The user chooses the set of
languages to be mirrored.

Usage: wp-mirror [OPTIONS] [CONFIGFILE]

Options:

  --debug            run in debug mode (more verbose)
  --delete-dump language  delete dump file and state for given language, and exit
  --help             display this help, and exit
  --mirror           force mirror mode (overrides monitor)
  --version          display version and copyright, and exit

Options for monitor mode:

  --gui              operate monitor in gui mode (default)
  --monitor          operate in monitor mode (gui default)
  --screen           operate monitor in screen mode
  --text             operate monitor in text mode

Options used during build and install:

  --config-default   display default config file, and exit
  --config-local     display local config file, and exit
  --copyright        display copyright file, and exit
  --mw-import        download latest version of `Import.php' from SVN


Files:

  /etc/cron.d/wp-mirror  weekly cron job

  /etc/logrotate.d/wp-mirror  logrotate configuration

  /etc/mediawiki/LocalSettings_wpmirror.php  configuration file for building and using a wikipedia farm

  /etc/wp-mirror/default.conf  default configuration file which lists all configurable parameters (do not edit this file)

  /etc/wp-mirror/local.conf.template  local configuration file (rename this file to local.conf)

  /etc/wp-mirror/local.conf  local configuration file (edit this file instead)

  /usr/share/mediawiki/includes/Import.php  patched version of Import.php

  /usr/share/mediawiki/maintenance/template.sql  template for wikimedia databases

  /usr/share/mediawiki/maintenance/importDump_farm.php  wikipedia farm wrapper for importDump.php

  /usr/share/mediawiki/maintenance/rebuildImages_farm.php  wikipedia farm wrapper for rebuildImages.php

  /usr/share/mediawiki/maintenance/update_farm.php  wikipedia farm wrapper for update.php

  /var/lib/mediawiki/images/  images are stored here

  /var/lib/mediawiki/images/wp-mirror/  work files are placed here

  /var/lib/mysql/wikidb  empty wikimedia database (dumped to make template.sql)

  /var/lib/mysql/wpmirror  database for wp-mirror state information

  /var/lib/mysql/xxwiki  one database for each mirror in the wikipedia farm (where `xx' is replaced by the language code, such as `en' or `simple')

  /var/log/wp-mirror.log  log file

  /var/run/wp-mirror.pid  pid file

Examples: 

Installation is described in `/usr/share/doc/wp-mirror/README'.  It
should be considered essential reading.

Wp-mirror uses two configuration files: `/etc/wp-mirror/default.conf'
and `/etc/wp-mirror/local.conf'.  The first shows all the configurable
parameters.  The second is where you should make your edits (if any).
By default, `wp-mirror' builds a mirror of the `simple' wiki.  The
`simple' wiki is the collection of articles written using Simple
English. That is, the articles are written using shorter sentences, as
would be appropriate for those for whom English is a Second Language
(ESL).  The `simple' wiki is also suitable for children.  If you wish
for a different wiki, or a set of wiki's, then edit
`/etc/wp-mirror/local.conf'.

After you edit /etc/wp-mirror/local.conf' you should run as root:

      # wp-mirror

You may run multiple instances of `wp-mirror' concurrently.  The first
instance runs in mirror mode, downloading and installing the articles
and images.  Each additional instance of `wp-mirror' runs in monitor
mode.  Instances running in monitor mode display the state of each
mirror.  If a suitable windowing system is present, progress bars are
rendered using graphics in a separate window, and otherwise using
ASCII characters in a console.  If X complains that you lack
permission, first try:

      $ xhost local:

You may control the mode of each instance of `wp-mirror' by providing
either the `--mirror' or the `--monitor' option at the command line.
For a desktop PC, it is reasonable to run two or three instances in
mirror mode together with an instance in monitor mode.  However, for a
laptop PC, it is best to run just one instance in mirror mode together
with one instance in monitor mode.

Wp-mirror normally runs in background as a weekly cron job, updating
the mirrors whenever the WikiMedia Foundation posts new dump files.

Wp-mirror configures a virtual-host for your web-server:

      <VirtualHost *:80>

      ServerAdmin webmaster@localhost

      ServerName mediawiki.site

      ServerAlias *.mediawiki.site

      DocumentRoot /var/lib/mediawiki

      CustomLog ${APACHE_LOG_DIR}/mediawiki.site-access.log combined

      ErrorLog ${APACHE_LOG_DIR}/mediawiki.site-error.log

      </VirtualHost>

Add the following line to `/etc/hosts':

      ::1 mediawiki.site simple.mediawiki.site

Then just open a web browser to <http://simple.mediawiki.site/> and
enjoy.

Report bugs in `wp-mirror' to <wp-mirror-list@nongnu.org>.")


;;;;--------------------------------------------------------------------------+
;;;; Any number of mirrors and monitors can communicate state via database    |
;;;;                                                                          |
;;;;     +--------+                      +---------+                          |
;;;;     | mirror |------          ------| monitor |                          |
;;;;     +--------+      \        /      +---------+                          |
;;;;     +--------+     +----------+                                          |
;;;;     | mirror |-----| wpmirror |                                          |
;;;;     +--------+     +----------+                                          |
;;;;     +--------+      /        \      +---------+                          |
;;;;     | mirror |------          ------| monitor |                          |
;;;;     +--------+                      +---------+                          |
;;;;                                                                          |
;;;;--------------------------------------------------------------------------+


(defconstant *wtf*
  "For help, please try:
    $ wp-mirror --help")


;;;;--------------------------------------------------------------------------+
;;;; Print utilities:                                                         |
;;;;   The idea is to give all the messages a common look and feel, and to    |
;;;;   have few format strings outside of this section.                       |
;;;;   Note:  `put-and-log-string' should be the only way to *standard-output*|
;;;;--------------------------------------------------------------------------+


(defun put-and-log-string (str)
  "Showing given string, and copying it to log file (if any)."
  (format *standard-output* "~a~%" str)
  (unless (null *log-enabled-p*)
    (with-open-file (f *whereis-file-log*
		       :direction :output
		       :if-exists :append
		       :if-does-not-exist :create)
      (format f "[~d]~a~%" (os:process-id) str))))

(defun die ()
  "Exiting immediately."
  (ext:exit 0))

(defun abort-and-die ()
  "Showing abort message, and die."
  (put-and-log-string (format nil "~a." (_ "Aborting")))
  (die))

(defun debug-message (message)
  "Showing given message when non-zero debug-level."
  (when (>= *debug-level* 1)
    (put-message (concatenate 'string (_ ";debug: ") message))))

(defun debug-message-done (message)
  "Showing given message followed by [DONE] when non-zero debug-level."
  (when (>= *debug-level* 1)
    (put-message-done (concatenate 'string (_ ";debug: ") message))))

(defun debug-message-start (message)
  "Showing given message followed by ellipses when non-zero debug-level."
  (when (>= *debug-level* 1)
    (put-message-start (concatenate 'string (_ ";debug: ") message))))

(defun debug-message-value (message value)
  "Showing given message and value when non-zero debug-level."
  (when (>= *debug-level* 1)
    (put-message-value (concatenate 'string (_ ";debug: ") message) value)))

(defun format-integer-for-human (n)
  "Formatting a number for human (e.g. 4096 --> 4.1K).  
Return four character string."
  (cond ((<= n 999)             (format nil "~3d " n))
	((<= n 9949)            (format nil "~3,1,-3fK" n))
	((<= n 999499)          (format nil "~3dK" (round (/ n 1000))))
	((<= n 9949999)         (format nil "~3,1,-6fM" n))
	((<= n 999499999)       (format nil "~3dM" (round (/ n 1000000))))
	((<= n 9949999999)      (format nil "~3,1,-9fG" n))
	((<= n 999499999999)    (format nil "~3dG" (round (/ n 1000000000))))
	((<= n 9949999999999)   (format nil "~3,1,-9fT" n))
	((<= n 999499999999999) 
	 (format nil "~3dT" (round (/ n 1000000000000))))
	((<= n 9949999999999999)   
	 (format nil "~3,1,-9fP" n))
	((<= n 999499999999999999)
	 (format nil "~3dP" (round (/ n 1000000000000000))))
	((<= n 9949999999999999999)   
	 (format nil "~3,1,-12fE" n))
	((<= n 999499999999999999999)
	 (format nil "~3dE" (round (/ n 1000000000000000000))))
	((<= n 9949999999999999999999)   
	 (format nil "~3,1,-15fZ" n))
	((<= n 999499999999999999999999)
	 (format nil "~3dZ" (round (/ n 1000000000000000000000))))
	((<= n 9949999999999999999999999)   
	 (format nil "~3,1,-18fY" n))
	((<= n 999499999999999999999999999)
	 (format nil "~3dY" (round (/ n 1000000000000000000000000))))
	(t                      (format nil "~d" n))))

(defun format-progress-bar-pulse ()
  "Formatting a progress bar (pulse type) as a string"
  (format-progress-bar-string 0.5 :pulse t))

(defun format-progress-bar-string (val &key (pulse nil))
  "Formatting a progress bar as a string"
  (let* ((len-string        35)
	 (s                 (make-string len-string :initial-element #\Space))
	 ;; if pulse, then use clock to generate pulse motion
	 (val-float         (cond (pulse (float (/ (get-decoded-time) 60)))
				  ((null val) 0)
				  ((numberp val) 
				   (max 0 val (min val 1)))  ; [0,1]
				  (t val)))
	 (val-percent       (* val-float 100))
	 (val-chars         (floor (* val-float len-string)))
	 (pos-equal-left    0)
	 (pos-arrow         (1- (+ pos-equal-left val-chars)))
	 (pos-equal-right   (1- pos-arrow)))
    (when (>= pos-arrow 0)
      (loop 
	for i of-type integer from pos-equal-left to pos-equal-right by 1
	do (setf (aref s i) #\=))
      (setf (aref s pos-arrow) #\>))
    (cond (pulse               (format nil "[~a] pulse" s))
          ((= val-percent   0) (format nil "[~a] wait " s))
          ((= val-percent 100) (format nil "[~a] done " s))
          (t                   (format nil "[~a] ~5,2f%" s val-percent)))))

(defmacro put-message (message)
  "Showing given message"
  `(put-and-log-string (format nil "~a" ,message)))

(defun put-message-and-abort (message)
  "Showing given message, and abort."
  (put-message message)
  (abort-and-die))

(defun put-message-and-die (message)
  "Showing given message, and die."
  (put-message message)
  (die))

(defun put-message-done (message)
  "Showing given message followed by [DONE]."
  (put-message-result message :DONE))

(defun put-message-fail (message)
  "Showing given message followed by [FAIL]."
  (put-message-result message :FAIL))

(defun put-message-fail-and-abort (message)
  "Showing given message followed by [FAIL], and abort."
  (put-message-result message :FAIL)
  (abort-and-die))

(defun put-message-fail-and-die (message)
  "Showing given message followed by [FAIL], and die."
  (put-message-result message :FAIL)
  (die))

(defun put-message-pass (message)
  "Showing given message followed by [PASS]."
  (put-message-result message :PASS))

(defmacro put-message-result (message result)
  "Showing given message followed by a result"
  `(put-and-log-string (format nil "~70a[~a]" ,message ,result)))

(defun put-message-start (message)
  "Showing given message followed by ellipses."
  (put-and-log-string (format nil "~a ..." message)))

(defun put-message-value (message value)
  "Showing given message and a value."
  (put-and-log-string (format nil "~34a: ~a" message value)))

(defun put-message-value-done (message value)
  "Showing given message and value, followed by [DONE]."
  (put-message-value-result message value :DONE))

(defun put-message-value-fail (message value)
  "Showing given message and value, followed by [FAIL]."
  (put-message-value-result message value :FAIL))

(defun put-message-value-fail-and-abort (message value)
  "Showing given message and value, followed by [FAIL], and abort."
  (put-message-value-result message value :FAIL)
  (abort-and-die))

(defun put-message-value-fail-and-die (message value)
  "Showing given message and value, followed by [FAIL], and die."
  (put-message-value-result message value :FAIL)
  (die))

(defun put-message-value-pass (message value)
  "Showing given message and value, followed by [PASS]."
  (put-message-value-result message value :PASS))

(defmacro put-message-value-result (message value result)
  "Showing given message and value, followed by a result"
  `(put-and-log-string (format nil "~34a: ~34a[~a]" ,message ,value ,result)))

(defun put-symbol-value (symbol)
  "Showing symbol and its value."
  (put-and-log-string (format nil "~34s: ~a" symbol (symbol-value symbol))))

(defun put-symbol-value-fail (symbol)
  "Showing symbol, its value, and [FAIL]."
  (put-symbol-value-result symbol :FAIL))

(defun put-symbol-value-pass (symbol)
  "Showing symbol, its value, and [PASS]."
  (put-symbol-value-result symbol :PASS))

(defmacro put-symbol-value-result (symbol result)
  "Showing symbol, its value, and a result."
  `(put-and-log-string (format nil "~34s: ~34a[~a]" 
			       ,symbol (symbol-value ,symbol) ,result)))

(defun put-symbol-value-table (symbol-list)
  "Showing table of symbols and their values."
  (dolist (symbol symbol-list t)
    (put-symbol-value symbol)))


;;;;--------------------------------------------------------------------------+
;;;; Parameters:                                                              |
;;;;   May be changed by: 1) config file `/etc/wp-mirror/local.conf', or      |
;;;;                      2) by command-line options.                         |
;;;;--------------------------------------------------------------------------+

;;; Was `--mirror' given as command line options?
(defparameter *cmd-force-mirror*        nil)
;;; Were any of `--gui', `--monitor', `--screen', or `--text' 
;;; given as command line options?
(defparameter *cmd-force-monitor*       nil)

;;; Database administrator and wikiuser account credentials will be
;;; grepped from *whereis-wikimedia-adminsettings* and
;;; *whereis-wikimedia-localsettings*
(defparameter *db-admin-user*           nil)    ; expect `root'
(defparameter *db-admin-password*       nil)    ;
(defparameter *db-name*                 nil)    ; expect `wikidb'
(defparameter *db-server*               nil)    ; expect `localhost'
(defparameter *db-type*                 nil)    ; expect `mysql'
(defparameter *db-wiki-user*            nil)    ; expect `wikiuser'
(defparameter *db-wiki-password*        nil)    ; 
(defparameter *db-wp-mirror-name*       "wpmirror") ; state

(defparameter *debug-level*             0)
(defparameter *log-enabled-p*           nil)

;;; Concurrency of mirroring
(defparameter *mirrors*                 4)      ; not implemented
(defparameter *mirror-concurrency-limit-xchunk* 1)

;;; Do we download and validate image files?  
(defparameter *mirror-image-download-p*                   t)
(defparameter *mirror-image-use-wikix-p*                nil)
(defparameter *mirror-image-validate-p*                   t)
(defparameter *mirror-rebuildimages-threshold*        20000)
(defparameter *mirror-rebuildimages-timeout-sec-min*   6000)
(defparameter *mirror-rebuildimages-timeout-sec-inc*    200)
(defparameter *mirror-rebuildimages-timeout-sec-max*  20000)
(defparameter *mirror-objectcache-threshold*          10000)
(defparameter *mirror-objectcache-delete-limit*        1000)

;;; Monitors (if any) should poll mirror status every so many seconds.
(defparameter *monitor-gui-font-preference-list* ; most preferred first
'("-*-helvetica-medium-r-*-*-12-*-*-*-*-*-*"    
  "-*-lucida-medium-r-*-*-12-*-*-*-*-*-*"
  "-*-lucida-medium-r-*-*-14-*-*-*-*-*-*"))
(defparameter *monitor-gui-one-window-per-language*    t) ;
(defparameter *monitor-mode*          :auto)    ; (:auto, :gui, :screen, :text)
(defparameter *monitor-poll-sec*         10)
(defparameter *monitor-poll-sec-min*     10)
(defparameter *monitor-poll-sec-max*   1000)
(defparameter *monitor-screen-stream*   nil)    ; if choose :screen
(defparameter *monitor-screen-height*   nil)    ; if choose :screen
(defparameter *monitor-screen-width*    nil)    ; if choose :screen

;;; Constants
(defconstant *kibibyte*        (expt 1024 1) "KiB")
(defconstant *mebibyte*        (expt 1024 2) "MiB")
(defconstant *gibibyte*        (expt 1024 3) "GiB")
(defconstant *tebibyte*        (expt 1024 4) "TiB")
(defconstant *kibobyte*        (expt 1000 1) "K or KB")
(defconstant *megabyte*        (expt 1000 2) "M or MB")
(defconstant *gigabyte*        (expt 1000 3) "G or GB")
(defconstant *terabyte*        (expt 1000 4) "T or TB")

;;; System resources minimums
(defparameter *system-cpu*                 1)
;;; Assert adequate disk and memory capacity for largest wiki's
(defparameter *system-disk-space-min*      (* 250 *gigabyte*))
(defparameter *system-physical-memory-min* (*   4 *gigabyte*))
(defparameter *wikimedia-large-wiki-list*
  '("en" "de" "fa" "nl" "it" "es" "pl" "ru" "ja" "pt"))

;;; Most of the heavy lifting is done by invoking existing utilities.
(defparameter *whereis-apache2ctl*     "/usr/sbin/apache2ctl")
(defparameter *whereis-bashrc*         "/etc/bash.bashrc")
(defparameter *whereis-bunzip2*        "/bin/bunzip2")
(defparameter *whereis-cp*             "/bin/cp")
(defparameter *whereis-chown*          "/bin/chown")
(defparameter *whereis-chmod*          "/bin/chmod")
(defparameter *whereis-cron*           "/etc/cron.d/wp-mirror")
(defparameter *whereis-curl*           "/usr/bin/curl")
(defparameter *whereis-curlrc*         "/root/.curlrc")
(defparameter *whereis-echo*           "/bin/echo")
(defparameter *whereis-gm*             "/usr/bin/gm")
(defparameter *whereis-grep*           "/bin/grep")
(defparameter *whereis-hdparm*         "/sbin/hdparm")
(defparameter *whereis-ls*             "/bin/ls")
(defparameter *whereis-md5sum*         "/usr/bin/md5sum")
(defparameter *whereis-file-dev-null*  "/dev/null")
(defparameter *whereis-file-etc-hosts* "/etc/hosts")
(defparameter *whereis-file-mediawiki-adminsettings*      "AdminSettings.php")
(defparameter *whereis-file-mediawiki-localsettings*      "LocalSettings.php")
(defparameter *whereis-file-mediawiki-localsettings-wpmirror*
  "LocalSettings_wpmirror.php")
(defparameter *whereis-file-mediawiki-import*        "Import.php")
(defparameter *whereis-file-mediawiki-importdump*    "importDump.php")
(defparameter *whereis-file-mediawiki-rebuildimages* "rebuildImages.php")
(defparameter *whereis-file-mediawiki-update*        "update.php")
(defparameter *whereis-file-mediawiki-farm-importdump*
  "importDump_farm.php")
(defparameter *whereis-file-mediawiki-farm-rebuildimages* 
  "rebuildImages_farm.php")
(defparameter *whereis-file-mediawiki-farm-update*        "update_farm.php")
(defparameter *whereis-mv*             "/bin/mv")
(defparameter *whereis-mysql*          "/usr/bin/mysql")
(defparameter *whereis-mysqldump*      "/usr/bin/mysqldump")
(defparameter *whereis-mysql-custom*   "/etc/mysql/conf.d/custom.cnf")
(defparameter *whereis-openssl*        "/usr/bin/openssl")
(defparameter *whereis-pidfile*        "/var/run/wp-mirror.pid")
(defparameter *whereis-php*            "/usr/bin/php")
(defparameter *whereis-php5-suhosin*   "/etc/php5/conf.d/suhosin.ini")
(defparameter *whereis-rm*             "/bin/rm")
(defparameter *whereis-rsvg*           "/usr/bin/rsvg")
(defparameter *whereis-wc*             "/usr/bin/wc")
(defparameter *whereis-wget*           "/usr/bin/wget")
(defparameter *whereis-wgetrc*         "/etc/wgetrc")
(defparameter *whereis-wikix*          "/usr/bin/wikix")
(defparameter *whereis-x*              "/usr/bin/X")

;;; Need directories and files
(defparameter *whereis-directory-apache-sites-available*
  "/etc/apache2/sites-available/")
(defparameter *whereis-directory-apache-sites-enabled*
  "/etc/apache2/sites-enabled/")
(defparameter *whereis-directory-configuration* "/etc/wp-mirror/")
(defparameter *whereis-directory-images*        "/var/lib/mediawiki/images/")
(defparameter *whereis-directory-images-bad* 
  "/var/lib/mediawiki/images/bad-images/")
(defparameter *whereis-directory-mediawiki*     "/etc/mediawiki/")
(defparameter *whereis-directory-mediawiki-extensions*
  "/etc/mediawiki-extensions/")
(defparameter *whereis-directory-mediawiki-extensions-available*
  "/etc/mediawiki-extensions/extensions-available/")
(defparameter *whereis-directory-mediawiki-extensions-enabled*
  "/etc/mediawiki-extensions/extensions-enabled/")
(defparameter *whereis-directory-mediawiki-includes*
  "/usr/share/mediawiki/includes/")
(defparameter *whereis-directory-mediawiki-maintenance*
  "/usr/share/mediawiki/maintenance/")
(defparameter *whereis-directory-mysql*         "/var/lib/mysql/")
(defparameter *whereis-directory-working*
  "/var/lib/mediawiki/images/wp-mirror/")
(defparameter *whereis-directory-tmp*           "/tmp/")

;;; Need files
(defparameter *whereis-file-config-default*     "default.conf")
(defparameter *whereis-file-config-local*       "local.conf")
(defparameter *whereis-file-log*                "/var/log/wp-mirror.log")
(defparameter *whereis-file-virtual-host*       "mediawiki.site.conf")
(defparameter *whereis-file-wikidb-template*    "template.sql")
(defparameter *whereis-tmp-http*                "/tmp/http")

;;; templates and URLs for wikimedia and wikix
(defparameter *whereis-file-checksums-template*
  "xxwiki-latest-md5sums.txt")
(defparameter *whereis-file-dump-template*
  "xxwiki-yyyymmdd-pages-articles.xml.bz2")
(defparameter *wikimedia-checksums-template*
  "xxwiki/latest/xxwiki-latest-md5sums.txt")
(defparameter *wikimedia-dump-template*
  "xxwiki/yyyymmdd/xxwiki-yyyymmdd-pages-articles.xml.bz2")
(defparameter *wikimedia-import*     "http://svn.wikimedia.org/viewvc/mediawiki/trunk/phase3/includes/Import.php?view=co")
(defparameter *wikimedia-site*       "http://dumps.wikimedia.org/")
(defparameter *wikimedia-wikix*      "http://wikix.ngen-cast.de/wikix.tar.gz")

;;; Used for splitting dump into small XML files `xchunks'
(defparameter *xchunk-page-count*               1000)
(defparameter *ichunk-curl-retry*                  0)   ; [0,1] are reasonable

;;; This is the key parameter
(defparameter *mirror-languages*     '("simple"))
(defparameter *virtual-host-name*    "mediawiki.site")


;;;;--------------------------------------------------------------------------+
;;;; Configuration File:                                                      |
;;;;   Read config file, apply sanity checks, redefine parameters             |
;;;;--------------------------------------------------------------------------+


(defconstant *config-default*
";;;;-------------------------------------------------------------------------+
;;;; This is the default configuration file for wp-mirror.                   |
;;;;-------------------------------------------------------------------------+
;;;
;;;    Make your customizations by editting `/etc/wp-mirror/local.conf'.
;;;    Do not make changes to this file.

;;; 0) Initialization.
;;;    Wp-mirror starts by checking for prerequisite software.
;;;    If any are not installed, please install them.
;;;    If any are installed, but in a directory other than the default, 
;;;    please uncomment and edit the relevant line.
;;(defparameter *whereis-apache2ctl*    \"/usr/sbin/apache2ctl\")
;;(defparameter *whereis-bashrc*        \"/etc/bash.bashrc\")
;;(defparameter *whereis-bunzip2*       \"/bin/bunzip2\")
;;(defparameter *whereis-cp*            \"/bin/cp\")
;;(defparameter *whereis-chown*         \"/bin/chown\")
;;(defparameter *whereis-chmod*         \"/bin/chmod\")
;;(defparameter *whereis-cron*          \"/etc/cron.d/wp-mirror\")
;;(defparameter *whereis-curl*          \"/usr/bin/curl\")
;;(defparameter *whereis-curlrc*        \"/root/.curlrc\")
;;
;;(defparameter *whereis-directory-apache-sites-available* 
;;  \"/etc/apache2/sites-available/\")
;;(defparameter *whereis-directory-apache-sites-enabled* 
;;  \"/etc/apache2/sites-enabled/\")
;;(defparameter *whereis-directory-configuration* \"/etc/wp-mirror/\")
;;(defparameter *whereis-directory-images*
;;  \"/var/lib/mediawiki/images/\")
;;(defparameter *whereis-directory-images-bad* 
;;  \"/var/lib/mediawiki/images/bad-images/\")
;;(defparameter *whereis-directory-mediawiki*     \"/etc/mediawiki/\")
;;(defparameter *whereis-directory-mediawiki-includes*
;;  \"/usr/share/mediawiki/includes/\")
;;(defparameter *whereis-directory-mediawiki-maintenance*
;;  \"/usr/share/mediawiki/maintenance/\")
;;(defparameter *whereis-directory-mysql*         \"/var/lib/mysql/\")
;;(defparameter *whereis-directory-working*
;;  \"/var/lib/mediawiki/images/wp-mirror/\")
;;
;;(defparameter *whereis-echo*                         \"/bin/echo\")
;;
;;(defparameter *whereis-file-config-default*          \"default.conf\")
;;(defparameter *whereis-file-config-local*            \"local.conf\")
;;(defparameter *whereis-file-dev-null*                \"/dev/null\")
;;(defparameter *whereis-file-etc-hosts*               \"/etc/hosts\")
;;(defparameter *whereis-file-log*                \"/var/log/wp-mirror.log\")
;;(defparameter *whereis-file-mediawiki-adminsettings* \"AdminSettings.php\")
;;(defparameter *whereis-file-mediawiki-localsettings* \"LocalSettings.php\")
;;(defparameter *whereis-file-mediawiki-localsettings-wpmirror*
;;  \"LocalSettings_wpmirror.php\")
;;(defparameter *whereis-file-mediawiki-import*        \"Import.php\")
;;(defparameter *whereis-file-mediawiki-importdump*    \"importDump.php\")
;;(defparameter *whereis-file-mediawiki-rebuildimages* \"rebuildImages.php\")
;;(defparameter *whereis-file-mediawiki-update*        \"update.php\")
;;(defparameter *whereis-file-mediawiki-farm-importdump*
;;  \"importDump_farm.php\")
;;(defparameter *whereis-file-mediawiki-farm-rebuildimages* 
;;  \"rebuildImages_farm.php\")
;;(defparameter *whereis-file-mediawiki-farm-update*   \"update_farm.php\")
;;(defparameter *whereis-file-virtual-host*       \"mediawiki.site.conf\")
;;(defparameter *whereis-file-wikidb-template*    \"template.sql\")
;;(defparameter *whereis-gm*            \"/usr/bin/gm\")
;;(defparameter *whereis-grep*          \"/bin/grep\")
;;(defparameter *whereis-hdparm*        \"/sbin/hdparm\")
;;(defparameter *whereis-ls*            \"/bin/ls\")
;;(defparameter *whereis-md5sum*        \"/usr/bin/md5sum\")
;;(defparameter *whereis-mv*            \"/bin/mv\")
;;(defparameter *whereis-mysql*         \"/usr/bin/mysql\")
;;(defparameter *whereis-mysqldump*     \"/usr/bin/mysqldump\")
;;(defparameter *whereis-mysql-custom*  \"/etc/mysql/conf.d/custom.cnf\")
;;(defparameter *whereis-openssl*       \"/usr/bin/openssl\")
;;(defparameter *whereis-php*           \"/usr/bin/php\")
;;(defparameter *whereis-php5-suhosin*  \"/etc/php5/conf.d/suhosin.ini\")
;;(defparameter *whereis-pidfile*       \"/var/run/wp-mirror.pid\")
;;(defparameter *whereis-rm*            \"/bin/rm\")
;;(defparameter *whereis-rsvg*          \"/usr/bin/rsvg\")
;;(defparameter *whereis-wc*            \"/usr/bin/wc\")
;;(defparameter *whereis-wget*          \"/usr/bin/wget\")
;;(defparameter *whereis-wgetrc*        \"/etc/wgetrc\")
;;(defparameter *whereis-x*             \"/usr/bin/X\")

;;;    Wp-mirror can generate `i-chunk's by using either `wikix' or a
;;;    built-in algorythm.  If you want `wikix', uncomment and set to
;;;    `t' (actually, do it in `local.conf', not here).
;;(defparameter *mirror-image-use-wikix-p*         nil)
;;;
;;;    If you want to use `wikix', note that it might not be available
;;;    as a package (e.g. deb, rpm).  Rather, it must be downloaded,
;;;    built, and installed.
;;(defparameter *whereis-wikix*         \"/usr/bin/wikix\")
;;(defparameter *wikimedia-wikix*  
;;  \"http://wikix.ngen-cast.de/wikix.tar.gz\")

;;; Wp-mirror also checks that the system has adequate resources
;;; for the largest wiki's
;;(defparameter *system-disk-space-min*      (* 250 *gigabyte*))
;;(defparameter *system-physical-memory-min* (*   4 *gigabyte*))
;;(defparameter *wikimedia-large-wiki-list*
;;  '(\"en\" \"de\" \"fa\" \"nl\" \"it\" \"es\" \"pl\" \"ru\" \"ja\" \"pt\"))


;;; 1) Dumpfile.
;;; a) Templates.  `Wp-mirror' downloads the latest checksum file.  From 
;;;    the checksum file it identifies the latest dump file, and downloads
;;;    that.  Site and file names are like the following templates.  Do
;;;    not edit these unless the folks at the Wikimedia Foundation change
;;;    something on their end.
;;(defparameter *whereis-file-checksums-template*
;;  \"xxwiki-latest-md5sums.txt\")
;;(defparameter *whereis-file-dump-template*
;;  \"xxwiki-yyyymmdd-pages-articles.xml.bz2\")
;;(defparameter *wikimedia-checksums-template*
;;  \"xxwiki/latest/xxwiki-latest-md5sums.txt\")
;;(defparameter *wikimedia-dump-template*
;;  \"xxwiki/yyyymmdd/xxwiki-yyyymmdd-pages-articles.xml.bz2\")
;;(defparameter *wikimedia-import*       \"http://svn.wikimedia.org/viewvc/mediawiki/trunk/phase3/includes/Import.php?view=co\")
;;(defparameter *wikimedia-site*         \"http://dumps.wikimedia.org/\")
;;; b) Languages.  The default is Simple English (shorter sentences),
;;;    which at 40G (as of 2011) fits nicely on a laptop.
;;(defparameter *mirror-languages* '(\"simple\"))


;;; 2) X-Chunks.
;;;    A dump file is in XML format, and may contain millions of pages
;;;    (`en' wiki has over 11 million as of 2011).  Unfortunately,
;;;    there are numerous pages that can cause the importation
;;;    process to fail, leaving all subsequent pages unimported.
;;;    Power failures can have a similar effect.  To make the process
;;;    much more robust (and easy to resume if broken off for some
;;;    reason), wp-mirror therefore splits the dump into many small
;;;    chunks containing 1000 (default) pages each.  These small files
;;;    are called `x-chunks'.  An unimportable page will then only
;;;    prevent importation of the subsequent pages within its own
;;;    x-chunk.  You can change this parameter.  However, keep in mind
;;;    the limitations of your file system.  If you are using
;;;    EXT2/EXT3, where the reference counter is only two bytes, then
;;;    you will not be able to create more than 64K files per
;;;    directory.  It is recommended that you use filesystems that
;;;    have no such limitation (e.g. ReiserFS).
;;(defparameter *xchunk-page-count*         1000)


;;; 3) I-Chunks.
;;;    Wikimedia.org no longer provides dump files of images.  Legal
;;;    issues and bandwidth have been cited as reasons.  However, the
;;;    software `wikix' is able to scan a dump file, find image names,
;;;    generate URLs for the images, and finally to generate a BASH
;;;    script that uses `curl' to download them.  These script files
;;;    are called i-chunks.  Wp-mirror generates an i-chunk for each
;;;    x-chunk.  If you do not have enough free disk space under
;;;    `/var/lib/mediawiki/images/' then set *mirror-image-download-p*
;;;    to `nil'.
;;;    Also, it is best to download images then import pages.  However,
;;;    if you wish to do the reverse, then: 
;;;        a) set `*mirror-image-download-p* to `nil', 
;;;        b) run `wp-mirror' once to import the pages, 
;;;        c) set `*mirror-image-download-p* to `t', and
;;;        d) run `wp-mirror' again to download the images.
;;(defparameter *mirror-image-download-p*            t)
;;(defparameter *ichunk-curl-retry*                  0)  ; [0,1] are reasonable

;;;    Occasionally, `curl' downloads an image file partially, and
;;;    then times-out, thereby providing a corrupt image file.
;;;    Therefore, after processing all the i-chunks, it is a good idea
;;;    to validate the downloaded image files and sequester any that
;;;    are corrupt.
;;(defparameter *mirror-image-validate-p*                   t)
;;(defparameter *mirror-rebuildimages-threshold*        20000)
;;(defparameter *mirror-rebuildimages-timeout-sec-min*   6000)
;;(defparameter *mirror-rebuildimages-timeout-sec-inc*    200)
;;(defparameter *mirror-rebuildimages-timeout-sec-max*  20000)
;;(defparameter *mirror-objectcache-threshold*          10000)
;;(defparameter *mirror-objectcache-delete-limit*        1000)


;;; 4) Mirrors
;;;    `Wp-mirror' can fork multiple concurrent mirror processes.
;;;    Warning: this feature is NOT IMPLEMENTED yet
;;(defparameter *mirrors*                      1) ; [1,3] are reasonable


;;; 5) Monitors.
;;;    A wp-mirror process can run in one of two modes: mirror and monitor.
;;; a) Mirror.  When a wp-mirror process begins, it looks for a pid-file.
;;;    If it finds none, then it sets one, and starts the mirroring process.
;;; b) Monitor.  If the user creates an additional instance of wp-mirror, 
;;;    the later instance will detect the pid-file of the first instance, 
;;;    and then monitor the progress of the first instance.  A monitor will
;;;    show a number of progress bars.
;;; c) Poll.  Monitors (if any) should poll the mirror status every 10
;;;    seconds (default).  There is little point to having a faster polling
;;;    rate given that mirroring a top ten wikipedia can take weeks to 
;;;    complete (although the small wikipedias will be done over-night).
;;(defparameter *monitor-poll-sec*          10)
;;(defparameter *monitor-poll-sec-min*      10)
;;(defparameter *monitor-poll-sec-max*    1000)
;;; d) GUI.  If the user has a suitable windowing system, then a monitor
;;;    may use it to render colored progress bars.  Otherwise, it uses a 
;;;    terminal to render characters.  If you have a preferred font, add
;;;    it to the front of the font preference list.
;;(defparameter *monitor-gui-font-preference-list* ; most preferred first
;;'(\"-*-helvetica-medium-r-*-*-12-*-*-*-*-*-*\"    
;;  \"-*-lucida-medium-r-*-*-12-*-*-*-*-*-*\"
;;  \"-*-lucida-medium-r-*-*-14-*-*-*-*-*-*\"))
;;(defparameter *monitor-gui-one-window-per-language*   nil)
;;(defparameter *monitor-mode*           :auto) ; (:auto, :gui, :screen, :text)
")

(defconstant *config-local*
";;;;-------------------------------------------------------------------------+
;;;; This is the local configuration file for wp-mirror.                     |
;;;;-------------------------------------------------------------------------+

;;;    This file was generated during `wp-mirror' installation.  If
;;;    you make manual changes, please provide comments (e.g. put 
;;;    your name and date above them) in case you need to revert them
;;;    later.
;;;
;;;    See `/etc/wp-mirror/default.conf' for all configurable
;;;    parameters and their default values.  However, be sure to make
;;;    changes here in this file, and not there.

;;;    Language.  By default, `wp-mirror' downloads the latest Simple
;;;    English language dump containing the most recent revision of
;;;    articles, namely, `simplewiki-yyyymmdd-pages-articles.xml.bz2'
;;;    (where `yyyymmdd' is a date).  If you wish to mirror different
;;;    languages (e.g. `de', `es', `fr', etc.), then first visit
;;;    <http://meta.wikimedia.org/wiki/List_of_Wikipedias> to find the
;;;    correct language codes.  Then uncomment and edit a
;;;    *mirror-languages* parameter below.  Here are some suggestions:

;;;    1) For LAPTOP:
;;;
;;;    Currently, a laptop with a single 500G HDD is not large enough
;;;    to mirror any of the top ten largest wiki's (see DESKTOP
;;;    below).  It will however do nicely for the others.
;;;
;;;    The default is Simple English (sentences are shorter).
;;;    You can expect 100ks (one day) run time for the initial build.
;;(defparameter *mirror-languages* '(\"simple\"))
;;;
;;;    If you want to get involved with the Wikimedia Founation, then
;;;    uncomment this line.
;;(defparameter *mirror-languages* '(\"meta\" \"simple\"))
;;;
;;;    If English is your Second Language (ESL), first visit
;;;    <http://meta.wikimedia.org/wiki/List_of_Wikipedias> to find your
;;;    language code.  Then uncomment and edit one of these:
;;(defparameter *mirror-languages* '(\"ar\" \"simple\"))        ;arabic
;;(defparameter *mirror-languages* '(\"hi\" \"simple\"))        ;hindi
;;(defparameter *mirror-languages* '(\"vi\" \"simple\"))        ;vietnamese
;;(defparameter *mirror-languages* '(\"zh\" \"simple\"))        ;chinese
;;;
;;;    If you study languages, uncomment and edit one of these:
;;(defparameter *mirror-languages* '(\"el\" \"la\" \"simple\")) ;greek,latin
;;(defparameter *mirror-languages* '(\"he\" \"yi\" \"simple\")) ;hebrew,yiddish
;;(defparameter *mirror-languages* '(\"simple\" \"zh-classical\")) ;chinese

;;;    2) For DESKTOP:
;;;
;;;    If you have 2T free disk space, then you should be able to
;;;    mirror one or more of the top ten largest wiki's
;;;    (currently en de fr nl it es pl ru ja pt).
;;;    But first read the `/usr/share/doc/wp-mirror/README' file.  
;;;
;;;    For the `en' wiki (the most demanding case) uncomment, and
;;;    expect 5Ms (two months) for the initial build.
;;(defparameter *mirror-languages* '(\"en\"))
;;;
;;;    If you are updating a large existing mirror (e.g. `en' wiki)
;;;    then uncomment this:
;;(defparameter *mirror-rebuildimages-timeout-sec-inc* 2000)
;;(defparameter *system-disk-space-min*      (* 100 *gigabyte*))
")


(defun put-parameters ()
  "Showing parameter values"
  (put-symbol-value-table
   '(*debug-level* 
		   *mirrors*
		   *mirror-concurrency-limit-xchunk*
		   *mirror-image-download-p* *mirror-image-validate-p*
		   *mirror-languages* *monitor-mode* *monitor-poll-sec*
		   *mirror-rebuildimages-threshold*
		   *mirror-rebuildimages-timeout-sec-min*
		   *mirror-rebuildimages-timeout-sec-inc*
		   *mirror-rebuildimages-timeout-sec-max*
		   *monitor-poll-sec-min* *monitor-poll-sec-max*
		   *system-cpu*
		   *system-disk-space-min* *system-physical-memory-min*
		   *wikimedia-large-wiki-list*
		   *whereis-bashrc* *whereis-bunzip2*
		   *whereis-cp* *whereis-chown* *whereis-chmod*
		   *whereis-cron* *whereis-curl* 
		   *whereis-curlrc*
		   *whereis-directory-apache-sites-available*
		   *whereis-directory-apache-sites-enabled*
		   *whereis-directory-configuration*
		   *whereis-directory-images* *whereis-directory-images-bad*
		   *whereis-directory-mediawiki-includes*
		   *whereis-directory-mediawiki-maintenance*
		   *whereis-directory-working* *whereis-directory-tmp*
		   *whereis-echo*
		   *whereis-file-checksums-template*
		   *whereis-file-config-default* *whereis-file-config-local* 
		   *whereis-file-dump-template*
		   *whereis-file-log*
		   *whereis-file-mediawiki-adminsettings*
		   *whereis-file-mediawiki-farm-importdump*
		   *whereis-file-mediawiki-farm-rebuildimages*
		   *whereis-file-mediawiki-farm-update*
		   *whereis-file-mediawiki-import*
		   *whereis-file-mediawiki-importdump*
		   *whereis-file-mediawiki-localsettings*
		   *whereis-file-mediawiki-localsettings-wpmirror*
		   *whereis-file-mediawiki-rebuildimages*
		   *whereis-file-mediawiki-update*
		   *whereis-file-virtual-host*
		   *whereis-file-wikidb-template*
		   *whereis-gm* *whereis-hdparm*
		   *whereis-ls*
		   *whereis-md5sum*
		   *whereis-mv* *whereis-mysql* *whereis-mysqldump*
		   *whereis-mysql-custom* 
		   *whereis-openssl*
		   *whereis-pidfile* *whereis-php*
		   *whereis-rm* *whereis-rsvg*
		   *whereis-tmp-http*
		   *whereis-wc* *whereis-wget* *whereis-wgetrc*
		   *whereis-wikix* *whereis-x*
		   *wikimedia-checksums-template*
		   *wikimedia-site* *wikimedia-wikix*
		   *xchunk-page-count*))
  t)

(defun process-configuration-files-or-die ()
  "Loading configuration file"
  ;; 0) load config files
  (let ((path-default (merge-pathnames
		       (parse-namestring *whereis-directory-configuration*)
		       (parse-namestring *whereis-file-config-default*)))
	(path-local   (merge-pathnames
		       (parse-namestring *whereis-directory-configuration*)
		       (parse-namestring *whereis-file-config-local*))))
    ;; default config file
    (put-message-value (_ "probing") path-default)
    (unless (file-exists-p path-default)
      (put-message-fail-and-abort (_ "file not found")))
    (load path-default :if-does-not-exist nil)
    (put-message-done (_ "default config file loaded"))
    ;; local config file
    (put-message-value (_ "probing") path-local)
    (unless (file-exists-p path-local)
      (put-message-fail-and-abort (_ "file not found")))
    (load path-local :if-does-not-exist nil)
    (put-message-done (_ "local config file loaded")))

  ;; 1) keep the monitor polling rate within reasonable bounds
  (put-message-start (_ "checking range *monitor-poll-sec*"))
  (put-message-value (_ "reasonable bounds would be")
		     (format nil "[~d,~d]" *monitor-poll-sec-min* 
			     *monitor-poll-sec-max*))
  (cond ((< *monitor-poll-sec* *monitor-poll-sec-min*)
	 (setq *monitor-poll-sec* *monitor-poll-sec-min*))
	((> *monitor-poll-sec* *monitor-poll-sec-max*)
	 (setq *monitor-poll-sec* *monitor-poll-sec-max*))
	(t t))
  (put-message-value-done (_ "checking range *monitor-poll-sec*")
			  *monitor-poll-sec*)
  ;; 2) return
  (let ((path-default (merge-pathnames
		       (parse-namestring *whereis-directory-configuration*)
		       (parse-namestring *whereis-file-config-default*))))
    (file-exists-p path-default)))


;;;;--------------------------------------------------------------------------+
;;;; Command Line:                                                            |
;;;;   Read command line options, apply sanity checks, redefine parameters    |
;;;;--------------------------------------------------------------------------+


(defun put-config-default-and-die ()
  "Showing default configuration file `/etc/wp-mirror/default.conf', and die."
  (put-message-and-die *config-default*))

(defun put-config-local-and-die ()
  "Showing local configuration file `/etc/wp-mirror/local.conf', and die."
  (put-message-and-die *config-local*))

(defun put-copyright-license-and-disclaimer-and-die ()
  "Showing copyright, license, and disclaimer, and die."
  (put-message-and-die *copyright-license-and-disclaimer*))

(defun put-help-and-die ()
  "Showing help message, and die."
  (put-message-and-die *help*))

(defun put-mw-import-and-die ()
  "Downloading latest version of `Import.php' from SVN"
  (download-file *wikimedia-import* "Import.php")
  (die))

(defun put-version-and-die ()
  "Showing version, copyright, and author, and die."
  (put-message *version*)
  (put-message " ")
  (put-message *copyright-license-and-disclaimer*)
  (put-message " ")
  (put-message-and-die *author*))

(defun put-wtf-and-die ()
  "Telling user how to get help, and die."
  (put-message-and-die *wtf*))

(defun process-command-line-arguments-or-die ()
  "Processing any command line arguments using getopt."
  (multiple-value-bind (filtered opt-args unexpected)
      ;; 0) parse command-line
      ;;    o FILE (if given) should be in `filtered',
      ;;    o options (and their arguments) should be in `*opt-args*', and
      ;;    o anything invalid should be in `unexpected'
      (getopt:getopt ext:*args*
		     '(("config-default"         :none     nil)
		       ("config-local"           :none     nil)
		       ("copyright"              :none     nil)
		       ("debug"                  :none     nil)
		       ("delete-dump"            :optional nil)
		       ("gui"                    :none     nil)
		       ("help"                   :none     nil)
		       ("mirror"                 :none     nil)
		       ("monitor"                :none     nil)
		       ("mw-import"              :none     nil)
		       ("screen"                 :none     nil)
		       ("text"                   :none     nil)
		       ("version"                :none     nil)
		       ))
    (when (assoc "debug"            opt-args :test #'string=) ; --debug
      (defparameter *debug-level* (1+ *debug-level*)))
    (debug-message-start (_ "Processing command line arguments"))
    (debug-message-start (_ "List packages and globals "))
    (debug-message-value "custom:*ansi*" custom:*ansi*)
    (debug-message-value "*package*"    *package*)
    (debug-message-value "*print-case*" *print-case*)
    (debug-message-start (_ "Parse command line"))
    (debug-message-value (_ "ext:*args*") ext:*args*)
    (debug-message-value (_ "filtered")   filtered)
    (debug-message-value (_ "opt-args")   opt-args)
    (debug-message-value (_ "unexpected") unexpected)
    (debug-message-done  (_ "Parse command line"))
    ;; 1) detect user who needs help and die
    (when (assoc "config-default"   opt-args :test #'string=) ; --config-default
      (put-config-default-and-die))
    (when (assoc "config-local"     opt-args :test #'string=) ; --config-local
      (put-config-local-and-die))
    (when (assoc "copyright"        opt-args :test #'string=) ; --copyright
      (put-copyright-license-and-disclaimer-and-die))
    (when (assoc "delete-dump"      opt-args :test #'string=) ; --delete-dump
      (put-message (assoc "delete-dump" opt-args :test #'string=))
      (let ((language (cdr (assoc "delete-dump" opt-args :test #'string=))))
	(dump-file-remove-all language)
	(die)))
    (when (assoc "help"             opt-args :test #'string=) ; --help
      (put-help-and-die))
    (when (assoc "mw-import"        opt-args :test #'string=) ; --mw-import
      (put-mw-import-and-die))
    (when (assoc "version"          opt-args :test #'string=) ; --version
      (put-version-and-die))
    (when unexpected
      (put-message-start (_ "Error: command line"))
      (put-message-value-fail (_ "unexpected option") unexpected)
      (put-wtf-and-die))
    (when (> (length filtered) 1)                      ; more than one FILE
      (put-message-start (_ "Error: command line"))
      (put-message-value-fail (_ "saw more than one file") filtered)
      (put-wtf-and-die))
    (when (assoc "monitor"          opt-args :test #'string=) ; --monitor
      (setq *monitor-mode*          :auto)
      (setq *cmd-force-monitor*     t))
    (when (assoc "gui"              opt-args :test #'string=) ; --gui
      (setq *monitor-mode*          :gui)
      (setq *cmd-force-monitor*     t))
    (when (assoc "screen"           opt-args :test #'string=) ; --screen
      (setq *monitor-mode*          :screen)
      (setq *cmd-force-monitor*     t))
    (when (assoc "text"             opt-args :test #'string=) ; --text 
      (setq *monitor-mode*          :text)
      (setq *cmd-force-monitor*     t))
    (when (assoc "mirror"           opt-args :test #'string=) ; --mirror
      (setq *cmd-force-mirror*      t)
      (setq *cmd-force-monitor*     nil))                     ; override
    (when (assoc "debug"            opt-args :test #'string=) ; --debug
      (setq *monitor-mode*          :text))                   ; forces --text
    ;; 2) set configuration file (if given on command-line)
    (when filtered
      (defparameter *whereis-file-config-local* (first filtered))
      (assert-file-exists-p *whereis-file-config-local*))
  (put-message-done (_ "processing command line arguments"))))


;;;;--------------------------------------------------------------------------+
;;;; Shell utilities:                                                         |
;;;;   Some tasks are more easily done with existing utilities.               |
;;;;   Avoid reinventing the wheel (absent compelling reasons).               |
;;;;--------------------------------------------------------------------------+


(defun shell-apache2ctl-grep-vhost-p ()
  "Determining if Apache2 was loaded with mediawiki virtual host"
  (let* ((command-string (format nil "~a -S |& ~a ~s"
				 *whereis-apache2ctl*
				 *whereis-grep*
				 *virtual-host-name*))
	 (result-stream  (ext:run-shell-command command-string
						:output :stream))
	 (lines          nil))
    (debug-message command-string)
    (if (null result-stream)
	(put-message-value-fail-and-abort (_ "shell command") command-string)
      (with-open-stream (s result-stream)
	(do ((l (read-line s nil 'eof) (read-line s nil 'eof)))
	    ((eq l 'eof) (nreverse lines))
	  (push l lines))))))

(defun shell-bash (file-name)
  "Running a shell script"
  (let ((result (ext:run-shell-command
		 (format nil "./~a" file-name))))
    (if (null result)
	(debug-message-value (_ "bash script returned") result)
      (put-message-value-fail (_ "bash script returned") result))
    result))

(defun shell-bunzip2 (file-name)
  "Decompressing `bzip2' file"
  (let* ((command-string (format nil "~a --force --keep ~a"
				 *whereis-bunzip2* file-name))
	 (result         (ext:run-shell-command command-string)))
    (if (null result)
	(debug-message-value (_ "bunzip2 returned") result)
      (put-message-value-fail-and-abort (_ "bunzip2 returned") result))
    result))

(defun shell-cat-grep (file-name pattern)
  "Grepping a file for a line matching pattern"
  (let* ((command-string (format nil "cat ~s | ~a ~s" 
				 file-name
				 *whereis-grep*
				 pattern))
	 (result-stream  (ext:run-shell-command command-string
						:output :stream))
	 (lines          nil))
    (debug-message command-string)
    (if (null result-stream)
	(put-message-value-fail-and-abort (_ "shell command") command-string)
      (with-open-stream (s result-stream)
	(do ((l (read-line s nil 'eof) (read-line s nil 'eof)))
	    ((eq l 'eof) (nreverse lines))
	  (push l lines))))))

(defun shell-chmod (file-name)
  "Chmod file to be owner executable"
  (let ((result	(ext:run-program *whereis-chmod*
				 :arguments (list "744" file-name))))
    (if (null result)
	(debug-message-value (_ "chmod returned") result)
      (put-message-value-fail (_ "chmod returned") result))
    result))

(defun shell-chown-dir (dir-name)
  "Chown directory recursively"
  (put-message-start (format nil "chown ~a" dir-name))
  (let ((result (ext:run-program 
		 *whereis-chown*
		 :arguments (list "--recursive" "www-data:www-data"
				  dir-name))))
    (if (null result)
	(put-message-value-done (_ "chown returned") result)
      (put-message-value-fail (_ "chown returned") result))
    result))

(defun shell-count-dir (dir-name)
  "Counting files in directory"
  (let* ((command (format nil "~a -U ~a | ~a -l"
			  *whereis-ls*
			  dir-name
			  *whereis-wc*))
	 (ls-stream (ext:run-shell-command command :output :stream))
	 (result    (with-open-stream (s ls-stream)
		      (read s))))
    (debug-message command)
    (if result
	(put-message-value (format nil "shell-count-dir ~a" dir-name) result)
      (put-message-value-fail (_ "shell-count-dir returned") result))
    result))

(defun shell-count-images ()
  "Counting images"
  (let* ((command (format nil "~a -U ~a/[0-9a-f]/* | ~a -l"
			  *whereis-ls*
			  *whereis-directory-images*
			  *whereis-wc*))
	 (ls-stream (ext:run-shell-command command :output :stream))
	 (result    (with-open-stream (s ls-stream)
		      (read s))))
    (debug-message command)
    (if result
	(debug-message-value (_ "shell-count-images") result)
      (put-message-value-fail (_ "shell-count-images returned") result))
    result))

(defun shell-copy-file (file-name dir-name)
  "Copying file to destination directory"
  (debug-message-value (_ "shell-copy-file: file-name") file-name)
  (debug-message-value (_ "shell-copy-file: dir-name") dir-name)
  (let* ((command  (format nil "~a -a ~a ~a"
			   *whereis-cp*
			   file-name
			   dir-name))
	(result (ext:run-shell-command command)))
    (debug-message command)
    (if (null result)
	(debug-message-value (_ "cp returned") result)
      (put-message-value-fail-and-abort (_ "cp returned") result))
    result))

(defun shell-curl-http-head (url file-name &key (put nil))
  "Downloading HTTP head from URL, result stored in `file-name.head'"
  (let* ((arguments (list "--head"
			  "--fail" url))
	 (message (append (list "curl") arguments))
	 (result  1)) ; assume abnormal process termination
    (unless (null put)
      (put-message message))
    (setq result
	  (ext:run-program *whereis-curl*
			   :arguments arguments
			   :output (format nil "~a.head" file-name)))
    (if (null result)
	(unless (null put)
	  (put-message-value (_ "curl --head returned") result))
      (put-message-value-fail (_ "curl --head returned") result))
    result))

(defun shell-curl-http-file (url file-name &key (put nil))
  "Downloading file from URL, save as file-name"
  (let* ((arguments (list "--retry" "1"
			  "--fail"
			  "--max-time" "5000"
			  url
			  "--output" file-name
			  ))
	 (message   (append (list "curl") arguments))
	 (wip-file  (format nil "~a.wip" file-name))
	 (result    1)) ; assume abnormal process termination
    (unless (null put)
      (put-message message))
    ;; 1) create a temporary file that we can use to detect partial downloads
    (with-open-file (f wip-file :direction :probe :if-does-not-exist :create))
    (unless (ext:probe-pathname wip-file)
      (put-message-value-fail-and-abort (_ "unable to create wip file") 
					wip-file))
    ;; 2) download file
    (setq result 
	  (ext:run-program *whereis-curl*
			   :arguments arguments))
    ;; 3) if success, remove temporary file
    (if (null result)
	(progn
	  (shell-remove-file wip-file)
	  (unless (null put)
	    (put-message-value-done (_ "curl returned") result)))
      (progn
	(put-message-value-fail (_ "curl returned") result)
	(when (file-exists-p file-name)
	  (put-message-value (_ "removing file fragment") file-name)
	  (shell-remove-file file-name)
	  (put-message-value-done (_ "removing file fragment") file-name)
	  (put-message-value (_ "removing wip file") file-name)
	  (shell-remove-file wip-file)
	  (put-message-value-done (_ "removing wip file") file-name))))
    result))

(defun shell-count-image (file-name)
  "Counting images `<title>File:' in file"
  (let* ((image-count    0)
	 (file-type      (sql-select-file-type file-name))
	 (command-string (cond
			  ((string= file-type "ichunk")
			   (format nil "cat ~a | ~a \"^if\" | wc -l"  
				   file-name
				   *whereis-grep*))
			  ((string= file-type "xchunk")
			   (format nil "cat ~a | ~a \"<title>File:\" | wc -l"
				   file-name
				   *whereis-grep*))
			  ((string= file-type "xml")
			   (format nil "cat ~a | ~a \"<title>File:\" | wc -l"
				   file-name
				   *whereis-grep*))
			  (t nil)))
	 (result-stream  (ext:run-shell-command command-string
						:output :stream)))
    (debug-message command-string)
    (if (null result-stream)
	(put-message-value-fail-and-abort (_ "shell command") command-string)
      (with-open-stream (s result-stream)
	(setq image-count (read s))))))

(defun shell-echo-md5sum (string)
  "Computing md5sum of a string"
  ;; Bug: fails if string contains a backquote
  (let* ((command-string (format nil "~a -n ~s | ~a md5"
				 *whereis-echo*
				 string
				 *whereis-openssl*))
	 (result-stream  nil))
    (debug-message command-string)
    (setq result-stream  (ext:run-shell-command command-string
						:output :stream))
    (if (null result-stream)
	(put-message-value-fail-and-abort (_ "shell command") command-string)
      (with-open-stream (s result-stream)
	(read-line s)))))

(defun shell-find-1k (dir-name)
  "Finding files smaller than 1k in given directory"
  (let* ((command-string (format nil "find ~a -size 1k" dir-name))
	 (result-stream  (ext:run-shell-command command-string
						:output :stream))
	 (lines          nil))
    (debug-message command-string)
    (if (null result-stream)
	(put-message-value-fail-and-abort (_ "shell command") command-string)
      (with-open-stream (s result-stream)
	(do ((l (read-line s nil 'eof) (read-line s nil 'eof)))
	    ((eq l 'eof) (nreverse lines))
	  (push l lines))))))

(defun shell-ls-grep (dir-name pattern)
  "Grepping a directory listing for file names matching a pattern"
  (let* ((command-string (format nil "~a -U ~a | ~a ~s" 
				 *whereis-ls* dir-name
				 *whereis-grep* pattern))
	 (result-stream  (ext:run-shell-command command-string
						:output :stream))
	 (lines          nil))
    (debug-message command-string)
    (if (null result-stream)
	(put-message-value-fail-and-abort (_ "shell command") command-string)
      (with-open-stream (s result-stream)
	(do ((l (read-line s nil 'eof) (read-line s nil 'eof)))
	    ((eq l 'eof) (nreverse lines))
	  (push l lines))))))

(defun shell-count-page (file-name)
  "Counting pages `<page>' in file"
  (let* ((page-count     0)
	 (file-type      (sql-select-file-type file-name))
	 (command-string (cond
			  ((string= file-type "xchunk")
			   (format nil "cat ~a | ~a \"  <page>\" | wc -l"  
				   file-name
				   *whereis-grep*))
			  ((string= file-type "xml")
			   (format nil "cat ~a | ~a \"  <page>\" | wc -l"  
				   file-name
				   *whereis-grep*))
			  (t nil)))
	 (result-stream  (ext:run-shell-command command-string
						:output :stream)))
    (debug-message command-string)
    (if (null result-stream)
	(put-message-value-fail-and-abort (_ "shell command") command-string)
      (with-open-stream (s result-stream)
	  (setq page-count (read s))))))

(defun shell-gm-identify (pathname &key (put nil))
  "Determining if image file is incomplete or corrupt"
  (unless (null put)
    (put-message-value (_ "shell gm") pathname))
  (let ((command-string (format nil
				"~a identify -verbose ~s |& ~a ~s"
				*whereis-gm*   (namestring pathname)
				*whereis-grep* "gm identify"))
	(result         nil))
    (unless (null put)
      (put-message command-string))
    (with-open-stream (s (ext:run-shell-command command-string
						:output :stream))
      (setq result (read-line s nil nil))
      (unless (or (null result) (null put))
	(put-message result)))
    result))   ; return nil means OK, a string may mean problems

(defun shell-hdparm-write-cache-disabled-p (block-device)
  "Determining if hard drive write cache has been disabled"
  (let ((result-stream (ext:run-program *whereis-hdparm*
					:arguments (list "-W" block-device)
					:output :stream))
	(line          nil))
    (if (null result-stream)
	(put-message-value-fail-and-abort (_ "hdparm -W") block-device)
      (with-open-stream (s result-stream)
	(read-line s)   ; throw away the first line  (which is blank)
	(read-line s)   ; throw away the second line (like '/dev/sda:')
	(setq line (read-line s)))) ; like ' write-caching =  0 (off)'
    (if (regexp:match "off" line) t nil))) ; t means disabled, nil means on

(defun shell-md5sum-and-return-stream (file-name)
  "Computing file checksum"
  (let ((s (ext:run-program *whereis-md5sum*
			    :arguments (list file-name)
			    :output :stream)))
    (if (null s)
	(put-message-value-fail-and-abort (_ "md5sum") file-name)
      s)))

(defun shell-md5sum (file-name)
  "Computing file checksum"
  (let* ((result   (shell-md5sum-and-return-stream file-name))
	 (l        (with-open-stream (s result)
		     (read-line s)))
	 (checksum (first (regexp:regexp-split " " l))))
    checksum))

(defun virtual-server-name (language-code)
  "Determining virtual server name given language-code"
  (format nil "~a.~a" language-code *virtual-host-name*))

(defun shell-mediawiki-farm-importdump (file-name)
  "Importing xchunk into `xxwiki' database"
  (let* ((language (sql-select-file-language file-name))
	 (code-path (merge-pathnames
		 (parse-namestring *whereis-directory-mediawiki-maintenance*)
		 (parse-namestring *whereis-file-mediawiki-farm-importdump*)))
	 (command (format nil "~a ~a ~a ~a" *whereis-php*
			  (namestring code-path)
			  file-name
			  (virtual-server-name language)))
	 (message (format nil "~a returned" command))
	 (result  nil))
    (put-message command)
    (setq result (ext:run-shell-command command))
    (if (null result)
	(debug-message-value message result)
      (put-message-value-fail message result))
    result))

(defun shell-mediawiki-farm-rebuildimages (file-name &key (time-out nil))
  "Rebuilding image table in `xxwiki' database"
  ;;
  ;; Design note:
  ;;
  ;; Problem.  `rebuildImages.php' seems to have a memory leak.  As it
  ;; processes images it occupies ever more SDRAM.  This can be seen
  ;; by running `top' and looking under the `RES' column.  For every
  ;; 20,000 images it occupies 1G of SDRAM, until a memory shortage
  ;; results in GNU/Linux randomly killing off processes, and the
  ;; system hangs.
  ;;
  ;; Solution.  We impose a timeout on the `rebuildImages.php' process
  ;; by killing it, if it is still running, after a period of time
  ;; (say 1ks).
  ;;
  ;; We spawn a sub-process and get its PID: 
  ;; 
  ;;     (php rebuildImages_farm.php lang --missing) & pid=$!
  ;;
  ;; We spawn a sub-process to timeout the PID:
  ;;
  ;;     (sleep 1000 && kill -9 $pid)
  ;;
  ;; We have the main process `wp-mirror' wait for it:
  ;;
  ;;     wait 1000
  ;;
  (let* ((language  (sql-select-file-language file-name))
	 (code-path (merge-pathnames
	       (parse-namestring *whereis-directory-mediawiki-maintenance*)
	       (parse-namestring *whereis-file-mediawiki-farm-rebuildimages*)))
	 (command   (if (null time-out)
			(format nil
	      "~a ~a ~a --missing > ~a" 
			  *whereis-php*
			  (namestring code-path)
			  (virtual-server-name language)
			  *whereis-file-dev-null*)
		      (format nil
"(~a ~a ~a --missing > ~a) & pid=$! && (sleep ~d && kill -9 $pid) & wait $pid" 
			  *whereis-php*
			  (namestring code-path)
			  (virtual-server-name language)
			  *whereis-file-dev-null*
			  time-out)))
	 (message   (format nil "~a returned" command))
	 (result    nil))
    (put-message command)
    (setq result (ext:run-shell-command command))
    (if (null result)
	(debug-message-value message result)
      (put-message-value-fail message result))
    result))

(defun shell-mediawiki-farm-update (file-name)
  "Updating `xxwiki' database"
  (let ((language (sql-select-file-language file-name)))
    (shell-phpscript-language-filename
     *whereis-file-mediawiki-farm-update* language file-name)))

(defmacro shell-phpscript-language-filename (php-script language file-name
							&key (args nil))
  "Calling a php script given language and file-name"
  `(let* ((path (merge-pathnames
		 (parse-namestring *whereis-directory-mediawiki-maintenance*)
		 (parse-namestring ,php-script)))
	  (message (format nil "~a ~a returned"
			   ,php-script
			   ,language))
	  (result (ext:run-program *whereis-php*
		    :arguments (append 
				(list path
				      (virtual-server-name ,language))
				,args)
		    :input  ,file-name
		    :output *whereis-file-dev-null*)))
     (if (null result)
	 (debug-message-value message result)
       (put-message-value-fail message result))
     result))

(defun shell-mediawiki-rebuildimages ()
  "Rebuilding `image' table in `wikidb' database"
  (shell-phpscript-filename *whereis-file-mediawiki-rebuildimages* nil
			    :args '("--missing")))

(defun shell-mediawiki-importdump (file-name)
  "Importing xchunk into `wikidb' database"
  (shell-phpscript-filename *whereis-file-mediawiki-importdump* file-name))

(defun shell-mediawiki-update ()
  "Updating `wikidb' database, deleting `objectcache' table"
  (shell-phpscript-filename *whereis-file-mediawiki-update* file-name))

(defmacro shell-phpscript-filename (php-script file-name &key (args nil))
  "Calling a php script given a file-name"
  `(let* ((message     (format nil "~a returned" ,php-script))
	  (script-path (namestring
			(merge-pathnames
			 (parse-namestring
			  *whereis-directory-mediawiki-maintenance*)
			 (parse-namestring ,php-script))))
	  (result (ext:run-program *whereis-php*
		    :arguments (append (list script-path) args)
		    :input  ,file-name
		    :output *whereis-file-dev-null*)))
     (if (null result)
	 (debug-message-value message result)
       (put-message-value-fail message result))
     result))

(defun shell-rename-file (file-name-1 file-name-2)
  "Renaming file"
  ;; directory is not changed, only the file is renamed
  (let ((result (ext:run-program
		 *whereis-mv*
		 :arguments (list file-name-1 file-name-2))))
    (if (null result)
	(debug-message-value (_ "mv returned") result)
      (put-message-value-fail-and-abort (_ "mv returned") result))
    result))

(defun shell-mysql-and-return-stream (arguments)
  "Querying MySQL database"
  (let ((s (ext:run-program *whereis-mysql*
			    :arguments arguments
			    :output :stream)))
  (if (null s)
      (put-message-value-fail-and-abort (_ "mysql query") arguments)
    s)))

(defun shell-mysql-load-database (db-name file-path)
  "Loading SQL dump file into database"
  (let* ((arguments (list 
		     (concatenate 'string "--host="     *db-server*)
		     (concatenate 'string "--user="     *db-admin-user*)
		     (concatenate 'string "--password=" *db-admin-password*)
		     (concatenate 'string "--database=" db-name)))
	 (result (ext:run-program *whereis-mysql*
				  :arguments arguments
				  :input file-path)))
    (if (null result)
	(debug-message-value (_ "mysql query") arguments)
      (put-message-value-fail-and-abort (_ "mysql query") arguments))
    result))

(defun shell-mysqldump-to-file (db-name file-path)
  "Dumping MySQL database"
  (let* ((arguments (list 
		     (concatenate 'string "--host="     *db-server*)
		     (concatenate 'string "--user="     *db-admin-user*)
		     (concatenate 'string "--password=" *db-admin-password*)
		     db-name))
	 (result (ext:run-program *whereis-mysqldump*
				  :arguments arguments
				  :output file-path)))
    (if (null result)
	(debug-message-value (_ "mysqldump query") arguments)
      (put-message-value-fail-and-abort (_ "mysqldump query") arguments))
    result))

(defun shell-remove-file (file-name)
  "Removing file"
  (let* ((command-string (format nil "~a -f ~a" *whereis-rm* file-name))
	 (result (ext:run-shell-command command-string
					:output *whereis-file-dev-null*)))
    (if (null result)
	(debug-message-value (_ "rm returned") result)
      (put-message-value-fail (_ "rm returned") result))
    result))

(defun shell-remove-all-working-files (language)
  "Removing all working files"
  (put-message-start (_ "cd to working directory"))
  (ext:cd *whereis-directory-working*)
  (put-message-value-done (_ "cd to working directory")
			  (ext:cd))
  (put-message-start (_ "removing old working files"))
  (put-message-start (_ "removing *"))
  (shell-remove-file (format nil "~a*" language))
  (put-message-value-done  (_ "removing *") language)
  (put-message-done  (_ "removing old working files")))

(defun shell-remove-old-working-files (language)
  "Removing decompressed dump file, xchunks, and ichunks given language"
  (put-message-start (_ "cd to working directory"))
  (ext:cd *whereis-directory-working*)
  (put-message-value-done (_ "cd to working directory")
			  (ext:cd))
  (put-message-start (_ "removing old working files"))
  (put-message-start (_ "removing *log"))
  (shell-remove-file (format nil "~a*log" language))
  (put-message-value-done  (_ "removing *log") language)
  (put-message-start (_ "removing *sh"))
  (shell-remove-file (format nil "~a*sh" language))
  (put-message-value-done  (_ "removing *sh") language)
  (put-message-start (_ "removing *xml"))
  (shell-remove-file (format nil "~a*xml" language))
  (put-message-value-done  (_ "removing *xml") language)
  (put-message-done  (_ "removing old working files")))

(defun shell-show-engine-innodb-status-grep (pattern)
  "Grepping `show engine innodb status' output"
  (let* ((command-string (concatenate 'string
			    (format nil "~a " *whereis-mysql*)
			    (format nil "--host=~a "     *db-server*)
			    (format nil "--user=~a "     *db-admin-user*)
			    (format nil "--password=~a " *db-admin-password*)
                            "--execute=\"SHOW ENGINE INNODB STATUS\\G\" "
			    "| "
			    (format nil "~a ~s" *whereis-grep* pattern)))
	 (result-stream (ext:run-shell-command command-string
					       :output :stream)))
    (if (null result-stream)
	(put-message-value-fail (_ "show engine innodb status") pattern)
      (with-open-stream (s result-stream)
	(read-line s nil nil))))) ; return string, or nil

(defun shell-wikix (xchunk-name ichunk-name)
  "Generating `ichunk' from `xchunk'"
  (let ((result (ext:run-program *whereis-wikix*
				 :input  xchunk-name
				 :output ichunk-name
				 :if-output-exists :overwrite)))
    (if (null result)
	(debug-message-value (_ "wikix returned") result)
      (put-message-value-fail (_ "wikix returned") result))
    result))


;;;;--------------------------------------------------------------------------+
;;;; Database utilities:                                                      |
;;;;   Debian packages cl-sql, cl-mysql, cl-uffi seem broken.                 |
;;;;   So 'query-database-*' are written using 'ext:run-program'.             |
;;;;--------------------------------------------------------------------------+


(defun grep-first (pattern file-name)
  "Grepping text file for first line matching pattern or nil"
  (first (shell-cat-grep file-name pattern)))

(defun grep-first-non-comment (pattern file-name)
  "Grepping text file for first non-comment line matching pattern or nil"
  (with-open-file (f file-name
		     :direction :input
		     :if-does-not-exist nil)
    (do ((l (read-line f nil 'eof) (read-line f nil 'eof)))
	((eq l 'eof) nil)
      (unless (regexp:match "^[#;]" l)
	(when (regexp:match pattern l)
	  (return l))))))

(defun query-database-and-return-stream (sql-string
					 &key 
					 (user     *db-wiki-user*)
					 (database *db-wp-mirror-name*)
					 (put      t))
  "Submitting SQL query, returning a stream containing the results"
  (unless (null put)
    (put-message-start (_ "query database return stream"))
    (put-message sql-string))
  (let* ((password  (cond ((eql user *db-wiki-user*)  *db-wiki-password*)
			  ((eql user *db-admin-user*) *db-admin-password*)
			  (t nil)))
	 (arguments (list 
		     (concatenate 'string "--host="     *db-server*)
		     (concatenate 'string "--user="     user)
		     (concatenate 'string "--password=" password)
		     ;; --batch separates cells with Tab, rows with Newline
		     ;; we need Tabs for reliable parsing
		     "--batch"
		     "--skip-column-names"
		     ;; --raw means no escaping special characters
		     "--raw"
		     (concatenate 'string "--database=" database)
		     (concatenate 'string "--execute="  sql-string))))
    (unless (null put)
      (put-message (format nil "~a" (append (list "mysql") arguments))))
    (shell-mysql-and-return-stream arguments)))

(defun query-database-and-return-list-of-strings (sql-string
				&key
				(user     *db-wiki-user*)
				(database *db-wp-mirror-name*)
				(put      nil))
  "Submitting SQL query, returning a list of strings where 
each string is a row in the table"
  (let ((table-list nil))
    (with-open-stream (s (query-database-and-return-stream sql-string
							   :user     user
							   :database database
							   :put      put))
      (do ((l (read-line s nil 'eof) (read-line s nil 'eof)))
	  ((eq l 'eof) (nreverse table-list))
	(push l table-list)))))

(defun query-database-and-return-list-of-lists (sql-string
				&key
				(user     *db-wiki-user*)
				(database *db-wp-mirror-name*)
				(put      nil))
  "Submitting SQL query, returning a list of lists of objects, 
where each object is an element in the table"  
  ;; this will misalign if there are NULL valued table elements not in
  ;; the last column (that is, NULL is not parsed)
  (let ((table-list nil))
    (with-open-stream (s (query-database-and-return-stream sql-string
							   :user     user
							   :database database
							   :put      put))
      ;; --batch option separates cells with Tab, ends line with Newline
      (regexp:with-loop-split (row s (string #\Tab))
        ;; row is bound to a list of substrings split from one line of s
	(loop
	  with new-row = nil
	  for cell    in row
	  as  elem    = (if (regexp:match ":" cell) ; we do not read packages
			    cell
			  (with-input-from-string (s cell) (read s nil 'eof)))
	  do (setq new-row (cond ((numberp elem) (push elem new-row))
				 ((symbolp elem) (push cell new-row))
				 (t              (push cell new-row))))
	  finally (setq table-list (push (nreverse new-row) table-list)))))
    (nreverse table-list)))

(defun sql-create-database (db-name &key (put nil))
  "Creating a database"
  (let ((create-database (format nil "CREATE DATABASE `~a`;" db-name)))
    (query-database-and-return-stream create-database
				      :database nil
				      :user     *db-admin-user*
				      :put      put)))

(defun sql-create-table-file (&key (put nil))
  "Creating table `file'"
  (let ((create-table    (concatenate 'string
        "CREATE TABLE IF NOT EXISTS file ("
	"timestamp TIMESTAMP DEFAULT CURRENT_TIMESTAMP "
	"ON UPDATE CURRENT_TIMESTAMP, "
	"language  VARCHAR(20) NOT NULL DEFAULT 'error', " ; e.g. `simple'
	"date      VARCHAR(8)  NOT NULL DEFAULT 'error', " ; e.g. `20111007'
	"name      VARCHAR(80) NOT NULL DEFAULT 'error', " 
	"size      BIGINT UNSIGNED NOT NULL DEFAULT 0, "   ; [0-18e18]
	"md5sum    VARCHAR(32) NOT NULL DEFAULT 'error', "
	"type      ENUM('checksum','dump','xml','xchunk','ichunk','images','error') "
	"NOT NULL DEFAULT 'error', "
	"state     ENUM('start','created','valid','pending','done','fail','error') "
        "NOT NULL DEFAULT 'error', "
	"page      INTEGER UNSIGNED NOT NULL DEFAULT 0, "
	"pages     INTEGER UNSIGNED NOT NULL DEFAULT 0, " ; pages
	"images    INTEGER UNSIGNED NOT NULL DEFAULT 0, " ; images
	"updates   INTEGER UNSIGNED NOT NULL DEFAULT 0, " ; 
	"semaphore INTEGER UNSIGNED NOT NULL DEFAULT 1, " ; [0-1]
	"PRIMARY KEY (name)) "
        "ENGINE=InnoDB;")))                               ; ACID compliant
    (query-database-and-return-stream create-table :put put)))

(defun sql-create-table-image (&key (put nil))
  "Creating table `image'"
  (let ((create-table    (concatenate 'string
        "CREATE TABLE IF NOT EXISTS image ("
	"timestamp  TIMESTAMP DEFAULT CURRENT_TIMESTAMP "
	"ON UPDATE  CURRENT_TIMESTAMP, "
	"img_md5sum VARCHAR(32) NOT NULL DEFAULT 'error', "
	"img_state  ENUM('created','valid','error') "
        "NOT NULL DEFAULT 'error', "
	"PRIMARY KEY (img_md5sum)) "
        "ENGINE=InnoDB;")))                               ; ACID compliant
    (query-database-and-return-stream create-table :put put)))

(defun sql-delete-from (language &key (put nil))
  "Deleting all rows of given language from `wpmirror.file' table"
  (let ((delete-from (format nil "DELETE FROM file WHERE language='~a' AND type <> 'checksum';" 
			     language)))
    (query-database-and-return-list-of-strings delete-from :put put)))

(defun sql-delete-from-objectcache (language &key (put nil))
  "Deleting a few rows from `objectcache' table in `xxwiki'"
  ;; we should routinely delete a few rows from the `objectcache'
  ;; table to avoid an accumulation of millions of lines, followed by
  ;; an hour long delay (which causes other queries to time-out) when
  ;; MediaWiki purges the cache.
  (unless (null put)
    (put-message-start (_ "delete from `objectcache' limit ")))
  (let* ((object-count (sql-select-count-xxwiki-objectcache language :put put))
	 (delete-from  (format nil "DELETE FROM `~awiki`.objectcache LIMIT ~d;"
			       (language-to-file-name language)
			       *mirror-objectcache-delete-limit*))
	 (threshold-p  (> object-count *mirror-objectcache-threshold*))
	 (result       nil))
    (if threshold-p
	(put-message-value (_ "objectcache above threshold")
			   (format nil "~d > ~d, deleting ~d"
				   object-count 
				   *mirror-objectcache-threshold*
				   *mirror-objectcache-delete-limit*))
      (put-message-value (_ "objectcache below threshold")
			   (format nil "~d <= ~d, skipping" 
				   object-count 
				   *mirror-objectcache-threshold*)))
    (when threshold-p
	(setq result (query-database-and-return-list-of-strings 
		      delete-from
		      :database *db-name*
		      :put put)))
    (unless (null put)
      (put-message-value-done (_ "delete from `objectcache' limit ") result))
    result))

(defun sql-drop-table (&key (put nil))
  "Dropping `file' table in *db-wp-mirror-name*"
  (let ((drop-table  "DROP TABLE file;"))
    (query-database-and-return-list-of-strings drop-table :put put)))

(defun sql-grant-all (db-name &key (put nil))
  "Granting permissions to *db-wiki-user* for given database"
  ;; error if db-name contains a `-' (e.g. `zh-classical')
  (let ((grant (format nil "GRANT ALL ON `~a`.* TO '~a'@'~a';" 
		       db-name *db-wiki-user* *db-server*)))
    (query-database-and-return-stream grant
				      :user     *db-admin-user*
				      :database nil
				      :put      put)))

(defun sql-insert-file (dump-language dump-date file-name file-type
				 &key (size nil) (md5sum nil)
				 (page nil) (pages nil) (images nil) 
				 (put nil))
  "Inserting new file"
  (let* ((insert-values (concatenate 'string
	   "INSERT INTO file "
	   "(language,date,name,size,md5sum,type,state,page,pages,images,updates,semaphore) "
	   "VALUES "
	   (format nil "('~a','~a','~a',~d,'~a','~a','~a',~d,~d,~d,~d,~d) "
		   dump-language
		   dump-date
		   file-name
		   (if (null size) 0 size)               ; size
		   (if (null md5sum) "0" md5sum)         ; md5sum
		   file-type                             ;
		   "start"                               ;
		   (if (null page) 0 page)               ; page start
		   (if (null pages) 0 pages)             ; pages
		   (if (null images) 0 images)           ; images
		   0                                     ; updates
		   1)                                    ; semaphore
	   "ON DUPLICATE KEY UPDATE "
	   (format nil "type='~a', " file-type)
	   (format nil "state='start';")))
	 (result (query-database-and-return-stream insert-values :put put)))
    (unless (null put) (put-message insert-values))
    t))

(defun sql-insert-image (md5sum &key (state nil) (put nil))
  "Inserting new image"
  (let* ((insert-values (concatenate 'string
	   "INSERT INTO image "
	   "(img_md5sum,img_state) "
	   "VALUES "
	   (format nil "('~a','~a');" 
		   md5sum
		   (if (null state) "created" state))))
	 (result (query-database-and-return-stream insert-values :put put)))
    (unless (null put) (put-message insert-values))
    t))

(defun sql-select-image (md5sum &key (put nil))
  "Determining if image-file is in `wpmirror.image'"
  (let* ((select (format nil "SELECT COUNT(*) FROM `wpmirror`.image WHERE img_md5sum='~a';" md5sum))
	 (result (sql-select-numeric select :put put)))
    ;; return t if found, nil otherwise
    (> result 0)))

(defun sql-select-image-state (md5sum &key (put nil))
  "Determining if image-file is in `wpmirror.image'"
  (let* ((select (format nil "SELECT img_state FROM `wpmirror`.image WHERE img_md5sum='~a';" md5sum)))
    (unless (null put) (put-message select))
    (first
     (first
      (query-database-and-return-list-of-lists select :put put)))))

(defun sql-select-languages (&key (put nil))
  "Determining languages seen in `wpmirror.file'"
  (let* ((select "SELECT language FROM `wpmirror`.file GROUP BY language;")
	 (result (query-database-and-return-list-of-strings select :put put)))
    (debug-message select)
    (debug-message result)
    result))

(defmacro sql-select-numeric (select &key (put nil))
  "Returning a numeric value from a database query"
  `(let ((result (query-database-and-return-list-of-lists ,select :put ,put)))
     (unless (null ,put) (put-message ,select))
     (cond ((null result) 0)
	   ((and (stringp (first (first result)))
		 (string= (first (first result)) "NULL")) 0)
	   (t (first (first result))))))

(defun sql-select-count (language &key (type nil) (state nil)
				  (semaphore nil) (put nil))
  "Counting records in `wpmirror.file'"
  ;; we centralize code common to most `SELECT COUNT(*)' queries
  (let ((select (concatenate 'string
			     "SELECT COUNT(*) FROM file "
			     (format nil "WHERE language='~a' " language)
			     (unless (null type)
			       (format nil "AND type='~a' " type))
			     (unless (null state)
			       (format nil "AND state='~a' " state))
			     (unless (null semaphore)
			       (format nil "AND semaphore='~d' " semaphore))
			     ";")))
     (sql-select-numeric select :put put)))

(defun sql-select-count-file (language &key (put nil))
  "Counting records in `wpmirror.file' of given language"
  (sql-select-count language :put put))

(defun sql-select-count-ichunk (language &key (put nil))
  "Counting `ichunk's given language"
  (sql-select-count language :type "ichunk" :put put))

(defun sql-select-count-ichunk-done (language &key (put nil))
  "Counting `ichunk's that are done"
  (sql-select-count language :type "ichunk" :state "done" :put put))

(defun sql-select-count-ichunk-fail (language &key (put nil))
  "Counting `ichunk's that have failed"
  (sql-select-count language :type "ichunk" :state "fail" :put put))

(defun sql-select-count-ichunk-remaining (language &key (put nil))
  "Counting `ichunk's remaining to be processed"
  (- (sql-select-count-ichunk      language :put nil)
     (sql-select-count-ichunk-done language :put nil)
     (sql-select-count-ichunk-fail language :put nil)))

(defun sql-select-count-ichunk-running (language &key (put nil))
  "Counting `ichunks' that are currently being processed"
  (sql-select-count language :type "ichunk" :semaphore 0 :put put))

(defun sql-select-count-image-valid (&key (put nil))
  "Counting validated image-files in `wpmirror.image'"
  (let* ((select (format nil "SELECT COUNT(*) FROM `wpmirror`.image WHERE img_state='valid';"))
	 (result (sql-select-numeric select :put put)))
    ;; return t if found, nil otherwise
    result))

(defun sql-select-count-images (language &key (put nil))
  "Counting `images' records"
  (sql-select-count language :type "images" :put put))

(defun sql-select-count-images-start (language &key (put nil))
  "Counting `images' records of state `start'"
  (sql-select-count language :type "images" :state "start" :put put))

(defun sql-select-count-images-created (language &key (put nil))
  "Counting `images' records of state `created'"
  (sql-select-count language :type "images" :state "created" :put put))

(defun sql-select-count-images-valid (language &key (put nil))
  "Counting `images' records of state `valid'"
  (sql-select-count language :type "images" :state "valid" :put put))

(defun sql-select-count-xxwiki-objectcache (language &key (put nil))
  "Counting `objectcache' records from `xxwiki'"
  (let* ((select (format nil "SELECT COUNT(*) FROM `~awiki`.objectcache;"
			 (language-to-file-name language)))
	 (result (query-database-and-return-list-of-lists select
							  :put put)))
    (unless (null put) (put-message select))
    (if (null result)
	0
      (first (first result)))))

(defun sql-select-count-xxwiki-image (language &key (put nil))
  "Counting `image' records from `xxwiki'"
  (let* ((select (format nil "SELECT COUNT(*) FROM `~awiki`.image;"
			 (language-to-file-name language)))
	 (result (query-database-and-return-list-of-lists select
							  :put put)))
    (unless (null put) (put-message select))
    (if (null result)
	0
      (first (first result)))))

(defun sql-select-count-xxwiki-page (language &key (put nil))
  "Counting `image' records from `xxwiki'"
  (let* ((select (format nil "SELECT COUNT(*) FROM `~awiki`.page WHERE page_namespace=0;"
			 (language-to-file-name language)))
	 (result (query-database-and-return-list-of-lists select
							  :put put)))
    (unless (null put) (put-message select))
    (if (null result)
	0
      (first (first result)))))

(defun sql-select-count-xchunk (language &key (put nil))
  "Counting xchunks"
  (sql-select-count language :type "xchunk" :put put))

(defun sql-select-count-xchunk-done (language &key (put nil))
  "Counting xchunks that are done"
  (sql-select-count language :type "xchunk" :state "done" :put put))

(defun sql-select-count-xchunk-fail (language &key (put nil))
  "Counting xchunks that have failed"
  (sql-select-count language :type "xchunk" :state "fail" :put put))

(defun sql-select-count-xchunk-remaining (language &key (put nil))
  "Counting `xchunk's remaining to be processed"
  (- (sql-select-count-xchunk      language :put nil)
     (sql-select-count-xchunk-done language :put nil)
     (sql-select-count-xchunk-fail language :put nil)))

(defun sql-select-count-xchunk-running (language &key (put nil))
  "Counting xchunks that are currently being processed"
  (sql-select-count language :type "xchunk" :semaphore 0 :put put))

(defun sql-select-dump-name (language &key (put nil))
  "Selecting name of dump file of given language"
  (let* ((select (format nil "SELECT name FROM file WHERE type='dump' and language='~a';" language))
	 (result (query-database-and-return-list-of-lists select :put put)))
    (unless (null put) (put-message select))
    (if (null result)
	nil
      (first (first result)))))

(defun sql-select-dump-xml-name (language &key (put nil))
  "Selecting name of the `xml' file of a given language"
  (let* ((select (format nil "SELECT name FROM file WHERE type='xml' and language='~a';" language))
	 (result (query-database-and-return-list-of-lists select :put put)))
    (unless (null put) (put-message select))
    (if (null result)
	nil
      (first (first result)))))

(defun sql-select-file (file-name &key (put nil))
  "Selecting file, returning list of lists"
  (let ((select (format nil "SELECT language,date,name,size,md5sum,type,state,page,pages,images,updates,semaphore FROM file WHERE name='~a';" file-name)))
    (unless (null put) 
      (put-message select))
    (query-database-and-return-list-of-lists select :put put)))

(defun sql-select-file-all (&key (put nil))
  "Selecting all, returning list of strings"
  (let ((select "SELECT * FROM file;"))
    (unless (null put)
      (put-message select))
    (query-database-and-return-list-of-strings select :put put)))
  
(defun sql-select-file-date (file-name &key (put nil))
  "Selecting file language, return string or nil"
  (let ((select (format nil "SELECT date FROM file WHERE name='~a';"
			file-name)))
    (first
     (query-database-and-return-list-of-strings select :put put))))
  
(defun sql-select-file-images (file-name &key (put nil))
  "Selecting images in file, return number or nil"
  (let ((select (format nil "SELECT images FROM file WHERE name='~a';"
			file-name)))
    (sql-select-numeric select :put put)))
  
(defun sql-select-file-language (file-name &key (put nil))
  "Selecting file language, return string or nil"
  (let ((select (format nil "SELECT language FROM file WHERE name='~a';"
			file-name)))
    (first
     (query-database-and-return-list-of-strings select :put put))))
  
(defun sql-select-file-md5sum (file-name &key (put nil))
  "Selecting file md5sum, return string or nil"
  (let ((select (format nil "SELECT md5sum FROM file WHERE name='~a';"
			file-name)))
    (first
     (query-database-and-return-list-of-strings select :put put))))
  
(defun sql-select-file-page (file-name &key (put nil))
  "Selecting file page, return number or nil"
  (let ((select (format nil "SELECT page FROM file WHERE name='~a';"
			file-name)))
    (sql-select-numeric select :put put)))
  
(defun sql-select-file-pages (file-name &key (put nil))
  "Selecting file pages in file, return number or nil"
  (let ((select (format nil "SELECT pages FROM file WHERE name='~a';"
			file-name)))
    (sql-select-numeric select :put put)))
  
(defun sql-select-file-semaphore (file-name &key (put nil))
  "Selecting file semaphore, return 0, 1, or nil"
  (let ((select (format nil "SELECT semaphore FROM file WHERE name='~a';"
			 file-name)))
    (sql-select-numeric select :put put)))

(defun sql-select-file-size (file-name &key (put nil))
  "Selecting file size, return number or nil"
  (let ((select (format nil "SELECT size FROM file WHERE name='~a';"
			 file-name)))
    (sql-select-numeric select :put put)))

(defun sql-select-file-state (file-name &key (put nil))
  "Selecting file state, return string or nil"
  (let ((select (format nil
			"SELECT state FROM file WHERE name='~a';"
			file-name)))
    (first
     (first
      (query-database-and-return-list-of-lists select :put put)))))

(defun sql-select-file-type (file-name &key (put nil))
  "Selecting file type, return string or nil"
  (let ((select (format nil
			"SELECT type FROM file WHERE name='~a';"
			file-name)))
    (first
     (first
      (query-database-and-return-list-of-lists select :put put)))))

(defun sql-select-file-updates (file-name &key (put nil))
  "Selecting file updates, return number or nil"
  (let ((select (format nil "SELECT updates FROM file WHERE name='~a';"
			file-name)))
    (sql-select-numeric select :put put)))

(defun sql-select-sum-images-images (language &key (put nil))
  "Counting images in `images' records"
  (let ((select (format nil "SELECT SUM(images) FROM file WHERE language='~a' AND type='images';" language)))
    (sql-select-numeric select :put put)))

(defun sql-select-sum-images-pages (language &key (put nil))
  "Counting validated images in `images' records"
  (let ((select (format nil "SELECT SUM(pages) FROM file WHERE language='~a' AND type='images';" language)))
    (sql-select-numeric select :put put)))

(defun sql-select-next-file-type-state (file-type file-state &key (put nil))
  "Selecting next file of given type and state, return filename"
  ;; 
  ;; Design note:
  ;;
  ;; Grabbing the lock by one mirror must be Isolated (the `I' in
  ;; `ACID') from attempts to grab the lock by other mirrors.  
  ;;
  ;; For this reason we use `SELECT ... FOR UPDATE' followed by
  ;; `UPDATE ...'.  The first identifies the qualifying row and
  ;; request InnoDB to establish an `IX' lock on that row, while the
  ;; second grabs the lock.
  ;;
  ;; Wrapping the transaction with `BEGIN TRANSACTION' and `COMMIT'
  ;; makes the entire transaction Atomic (the `A' in `ACID')
  ;;
  (let* ((select (concatenate 'string
	       "START TRANSACTION; "                 ; <--- atomic
	       "SELECT @name := name AS name,@semaphore := semaphore AS semaphore FROM file "
	       (format nil "WHERE type='~a' " file-type)
	       (format nil "AND state='~a' " file-state)
	       "AND semaphore=1 "                    ; <--- free
	       "ORDER BY name,state "
	       "LIMIT 1 "
	       "FOR UPDATE; "                        ; <--- atomic
	       "UPDATE file SET semaphore=0 "        ; <--- locked
	       "WHERE name=@name; "
	       "COMMIT;"))                           ; <--- atomic
	 (result (query-database-and-return-list-of-lists select :put put))
	 (file-name (first (first result))))
    (unless (null put)
      (put-message-value (_ "sql-select-next-file-type-state select") select)
      (put-message-value (_ "sql-select-next-file-type-state result") result)
      (put-message-value (_ "sql-select-next-file-type-state file-name") file-name))
    (unless (null result) file-name)))

(defun sql-select-next-file (&key (put nil))
  "Selecting next file for processing by the finite state machine"
  ;; Each file has a semaphore to assure Isolation (the `I' in ACID)
  (let ((message    (_ "selecting next file for processing"))
	(file-name  nil))
    (put-timestamp-message-start message)
    ;; select highest priority file
    (setq file-name
	  (loop 
	     with file-name = nil
	     for (type state concurreny non-image) in *type-state-priority*
	       when (and (or (i-am-the-first-mirror-p) concurreny)
			 (or *mirror-image-download-p* non-image))
	       do (setq file-name (sql-select-next-file-type-state 
				   type state :put put))
	     thereis file-name))
    ;;    notify of grab lock
    (if (null file-name)
	(progn
	  (put-message (_ "no file ready for processing"))
	  (sleep 1))
      (progn
	(put-message (format nil "~a" file-name))
	(put-message-value (_ "grab lock: set semaphore to")
			   (sql-select-file-semaphore file-name))
	(put-timestamp-message-done message)))
    file-name))

(defun sql-select-ichunk-wip (language &key (put nil))
  "Selecting ichunk in progess (if any)"
  (let* ((select (format nil "SELECT name FROM file WHERE language='~a' AND type='ichunk' AND state='valid';" language))
	 (result (query-database-and-return-list-of-strings select :put put)))
    (unless (null put)
      (if (null result)
	  (put-message-value (_ "no ichunks in progress") result)
	(put-message-value (_ "ichunks in progress") (cdr result))))
    (cdr result)))

(defun sql-select-xchunk-wip (language &key (put nil))
  "Selecting xchunk in progess (if any)"
  (let* ((select (format nil "SELECT name FROM file WHERE language='~a' AND type='xchunk' AND state='valid';" language))
	 (result (query-database-and-return-list-of-strings select :put put)))
    (unless (null put)
      (if (null result)
	  (put-message-value (_ "no xchunks in progress") result)
	(put-message-value (_ "xchunks in progress") (cdr result))))
    (cdr result)))

(defun sql-show-databases (&key (user *db-wiki-user*) 
				(put  nil))
  "Showing all databases"
  (let ((show-databases "SHOW DATABASES;"))
    (query-database-and-return-list-of-strings show-databases
					       :user     user
					       :database nil
					       :put      put)))

(defun sql-show-tables (db-name &key (user *db-wiki-user*)
			     (put  nil))
  "Showing tables in *db-wp-mirror-name*"
  (let ((show-tables  (format nil "SHOW TABLES FROM `~a`;" db-name)))
    (query-database-and-return-list-of-strings show-tables
					       :user user
					       :put  put)))

(defun sql-show-innodb-buffer-pool-pages-dirty (&key (put nil))
  "Showing number of buffer pool pages that are modified"
  (let* ((show   "SHOW STATUS LIKE 'Innodb_buffer_pool_pages_dirty';")
	 (result (query-database-and-return-list-of-lists show)))
    (if (null result)
	0
      (second (first result)))))

(defun sql-show-innodb-buffer-pool-pages-insert (&key (put nil))
  "Showing number of buffer pool pages used for insert buffer"
  ;;
  ;; Design note:
  ;;
  ;; As of MySQL 5.1, InnoDB does not expose any status variables
  ;; regarding the insert buffer.  So we use a shell script.
  ;;

  ;(let ((result (query-database-and-return-list-of-lists
  ;		 "SHOW STATUS LIKE 'Innodb_buffer_pool_pages_insert';")))
  ;  (if (null result)
  ;    0
  ;    (second (first result)))))

  (let* ((pattern  "Ibuf: size")
	 (l        (shell-show-engine-innodb-status-grep pattern)))
    (if (null l)
	0
      (with-input-from-string (s (third (regexp:regexp-split " " l)))
			      (read s)))))

(defun sql-show-innodb-buffer-pool-pages-misc (&key (put nil))
  "Showing number of buffer pool pages used for miscellaneous purposes"
  (let ((result (query-database-and-return-list-of-lists
		 "SHOW STATUS LIKE 'Innodb_buffer_pool_pages_misc';")))
    (if (null result)
	0
      (second (first result)))))

(defun sql-show-innodb-buffer-pool-pages-total (&key (put nil))
  "Showing number of buffer pool pages"
  (let ((result (query-database-and-return-list-of-lists
		 "SHOW STATUS LIKE 'Innodb_buffer_pool_pages_total';")))
    (if (null result)
	0
      (second (first result)))))

(defun sql-show-innodb-data-file-path (&key (put nil))
  "Showing filepath to InnoDB table space"
  (put-message (_ "Showing InnoDB data file path"))
  (let* ((result    (query-database-and-return-list-of-lists
		     "SHOW VARIABLES LIKE 'innodb_data_file_path';"))
	 (file-path nil))
    (if (null result)
	(put-message-value-fail-and-abort (_ "show innodb_data_file_path returned") result)
      (setq file-path (second (first result))))
    (put-message-value (_ "innodb_data_file_path") file-path)
    ;; this file path may contain a colon ':' when a raw partition is used
    ;; (e.g. 'ibdata1:10M:autoextend' (default))
    ;; (e.g. '/dev/mapper/vg0-ibdata0:300000Mraw')
    (setq file-path (first (regexp:regexp-split ":" file-path)))
    (put-message-value (_ "innodb_data_file_path trimmed") file-path)
    ;; the default case only gives the file-name, so merge directory path
    (let ((path-dir (pathname-directory (parse-namestring file-path))))
      (when (null path-dir)
	(put-message-value (_ "table space in directory")
			   *whereis-directory-mysql*)
	(setq file-path (merge-pathnames file-path *whereis-directory-mysql*))
	(put-message-value (_ "path to table space") file-path)))
    file-path))

(defun sql-show-innodb-file-format (&key (put nil))
  "Showing InnoDB file format (`Antelope' or `Barracuda')"
  (let ((result (query-database-and-return-list-of-lists
		 "SHOW VARIABLES LIKE 'innodb_file_format';")))
    (if (null result)
	0
      (second (first result)))))

(defun sql-update-fail-to-pending (&key (put nil))
  "Updating all failed `xchunk's and `ichunk's"
  (put-message-start (_ "updating state from `fail' to `pending'"))
  (put-message (_ "to give failed chunks as second chance"))
  (let ((update "UPDATE file SET state='pending' WHERE state='fail' AND (type='xchunk' OR type='ichunk');"))
    (unless (null put) (put-message update))
    (query-database-and-return-stream update :put put)
    (put-message-done (_ "updating state from `fail' to `pending'"))
    t))

(defun sql-update-images-to-pending (&key (put nil))
  "Updating all `images' to `pending'"
  (put-message-start (_ "updating state from `done' to `pending'"))
  (put-message (_ "to have `chown' run again"))
  ;; do just once (not once for each language)
  (let ((update (format nil "UPDATE file SET state='pending' WHERE language='~a' AND state='done' AND type='images';"
			(first *mirror-languages*))))
    (unless (null put) (put-message update))
    (query-database-and-return-stream update :put put)
    (put-message-done (_ "updating state from `done' to `pending'"))
    t))

(defun sql-update-file (file-name &key (size nil) (md5sum nil) (state nil)
				  (page nil) (pages nil) (images nil)
				  (semaphore nil)
				  (put nil))
  "Updating file size, md5sum, state, page, pages, images, updates"
  (let* ((old-state   (sql-select-file-state   file-name :put put))
	 (old-updates (sql-select-file-updates file-name :put put))
	 (new-updates (1+ old-updates))
	 (update (concatenate 'string
	           "UPDATE file SET "
		   (unless (null size)     (format nil "size=~d,"       size))
		   (unless (null md5sum)   (format nil "md5sum='~a'," md5sum))
		   (unless (null state)    (format nil "state='~a',"   state))
		   (unless (null page)     (format nil "page=~d,"       page))
		   (unless (null pages)    (format nil "pages=~d,"     pages))
		   (unless (null images)   (format nil "images=~d,"   images))
		   (unless (null semaphore)
		     (format nil "semaphore=~d," semaphore))
		   (format nil "updates=~d " new-updates)
		   (format nil "WHERE name='~a';" file-name))))
    (debug-message-value (_ "old-state") old-state)
    (debug-message-value (_ "old-updates") old-updates)
    (debug-message-value (_ "new-updates") new-updates)
    (unless (null put) (put-message update))
    (query-database-and-return-stream update :put put)
    t))

(defun sql-update-file-images (file-name images &key (put nil))
  "Updating images per file"
  (sql-update-file file-name :images images :put put))

(defun sql-update-file-md5sum (file-name md5sum &key (put nil))
  "Updating file md5sum"
  (sql-update-file file-name :md5sum md5sum :put put))

(defun sql-update-file-page (file-name page &key (put nil))
  "Updating file start page"
  (sql-update-file file-name :page page :put put))

(defun sql-update-file-pages (file-name pages &key (put nil))
  "Updating pages per file"
  (sql-update-file file-name :pages pages :put put))

(defun sql-update-file-semaphore (file-name semaphore &key (put nil))
  "Updating semaphore"
  ;; Only use this function for releasing the lock (setting to
  ;; semaphore to 1).  Grabbing the lock must be isolated from other
  ;; transactions (see sql-select-next-file-type for details)
  (let ((old-semaphore (sql-select-file-semaphore file-name :put put)))
    (sql-update-file file-name :semaphore semaphore :put put)
    ;(sleep 1)
    (let ((new-semaphore (sql-select-file-semaphore file-name :put put)))
      (if (eql new-semaphore 0)
	  (put-message-value (_ "grab lock - update semaphore") 
		     (format nil "~d-->~d" old-semaphore new-semaphore))
	(put-message-value (_ "release lock - update semaphore") 
		   (format nil "~d-->~d" old-semaphore new-semaphore))))))

(defun sql-update-file-semaphore-all (semaphore &key (put nil))
  "Updating all semaphores"
  (let ((update (format nil "UPDATE file SET semaphore=~d;" semaphore)))
    (unless (null put) (put-message-value (_ "update") update))
    (query-database-and-return-stream update :put put)
    t))

(defun sql-update-file-size (file-name &key (put nil))
  "Updating file size"
  (let ((record-size (sql-select-file-size file-name :put put))
	(actual-size nil))
    (unless (null put) (put-message-value (_ "record size") record-size))
    (if (ext:probe-pathname file-name)
	(progn
	  (setq actual-size (posix:file-size file-name))
	  (unless (null put) (put-message-value (_ "actual size") actual-size))
	  (unless (eql record-size actual-size)
	    (sql-update-file file-name :size actual-size :put put)))
      (put-message-value-fail-and-abort (_ "file not found") file-name)))
  t)

(defun sql-update-file-state (file-name state &key (put nil))
  "Updating file state"
  (sql-update-file file-name :state state :put put))

  
;;;;--------------------------------------------------------------------------+
;;;; Www utilities:                                                           |
;;;;--------------------------------------------------------------------------+


(defun download-http-head-and-file (url file-name)
  "Downloading file and its HTTP head from URL"
  (and (null (shell-curl-http-head url file-name :put t))
       (null (shell-curl-http-file url file-name :put t))))

(defun download-file (url file-name)
  "Downloading file and its HTTP head from given URL.
Cache header to avoid retrieving existing file."
  (let ((file-head (format nil "~a.head" file-name))
	(tmp-head  (format nil "~a.head" *whereis-tmp-http*))
	(wip-file  (format nil "~a.wip" file-name)))
    (if (and (ext:probe-pathname file-name)
	     (ext:probe-pathname file-head)
	     (not (ext:probe-pathname wip-file)))
	;; if both exist (and no `.wip' as evidence of partial download), 
	;; then download tmp-head, and compare ETags
	(when (null (shell-curl-http-head url *whereis-tmp-http* :put t))
	  (let* ((file-head-etag (grep-first "ETag" file-head))
		 (tmp-head-etag  (grep-first "ETag" tmp-head )))
	    ;; if ETags do not match, then download file and its head
	    (put-message-value (_ "ETag from head of server file")  
			       tmp-head-etag)
	    (put-message-value (_ "ETag from head of local file") 
			       file-head-etag)
	    (if (equal file-head-etag tmp-head-etag)
		(put-message (_ "server file no newer than local file---not retrieving"))
	      (download-http-head-and-file url file-name))))
      ;; If file or head missing, then download both
      (progn
	(put-message "file or head missing---try to retrieve both")
	(download-http-head-and-file url file-name)))))


;;;;--------------------------------------------------------------------------+
;;;; Error messages detailed:                                                 |
;;;;--------------------------------------------------------------------------+


(defparameter *error-messages* (make-hash-table :test 'equal))

(setf (gethash 'assert-clisp-features-p *error-messages*)
      (_ "
+-----------------------------------------------------------------------------+
| CATEGORY:    software prerequisite                                          |
| ASSERT:      assert-clisp-features-p                                        |
| ABSTRACT:    The installed version of `clisp' is missing key features.      |
| DESCRIPTION: Like all languages, `clisp' comes with extensions and packages |
|              that offer features well beyond the minimum for compliance with|
|              standards.  `wp-mirror' relies on these:                       |
|              (:asdf2 :asdf :clisp :clx :gettext :i18n :loop :regexp         |
|              :screen :syscalls)                                             |
| GNU/LINUX:   To see what features your `clisp' installation offers, try:    |
|                $ clisp -q -q                                                |
|                [1]> *features*                                              |
| REFERENCE:   See the `wp-mirror' README for more on planning your software. |
+-----------------------------------------------------------------------------+
"))

(setf (gethash 'assert-concurrency-limit-xchunk-p *error-messages*)
      (_ "
+-----------------------------------------------------------------------------+
| CATEGORY:    software prerequisite                                          |
| ASSERT:      assert-concurrency-limit-xchunk-p                              |
| ABSTRACT:    Unknown `innodb-file-format'                                   |
| DESCRIPTION: InnoDB has two file formats (as of 2011):                      |
|              o `Antelope' is uncompressed and handles concurrency well.     |
|                We limit concurrency to lesser of number of CPUs and three.  |
|              o `Barracuda' is compressed using `zlib' and does NOT handle   |
|                concurrency well (frequent deadlocks).                       |
|                We limit concurrency to one (that is, we process `x-chunk's  |
|                one-by-one).                                                 |
| REFERENCE:   See the `wp-mirror' README for more on InnoDB.                 |
+-----------------------------------------------------------------------------+
"))

(setf (gethash 'assert-curl-max-time-p *error-messages*)
      (_ "
+-----------------------------------------------------------------------------+
| CATEGORY:    software prerequisite                                          |
| ASSERT:      assert-curl-max-time-p                                         |
| ABSTRACT:    Unable to find timeout option in configuration file for `curl'.|
| DESCRIPTION: `Wp-mirror' uses `curl' to download files.  Sometimes however  |
|              only a partial file is served.  In this case `curl' blocks.    |
|              However, `curl' does offer a time-out to break the block.      |
| GNU/LINUX:   The configuration file for `curl' is                           |
|                `/root/.curlrc'                                              |
|              In this file you should append                                 |
|                `--max-time 1000'                                            |
| REFERENCE:   See `wp-mirror' README for more on planning your system.       |
+-----------------------------------------------------------------------------+
"))

(setf (gethash 'assert-database-administrator-credentials-p *error-messages*)
      (_ "
+-----------------------------------------------------------------------------+
| CATEGORY:    software prerequisite                                          |
| ASSERT:      assert-database-administrator-credentials-p                    |
| ABSTRACT:    Unable to find database administrator credentials.             |
| DESCRIPTION: `Wp-mirror' reads through the MediaWiki configuration file     |
|              looking for both `wikiuser' and administrator credentials.     |
|              `Wp-mirror' will not let you start without finding these       |
|              credentials there.  `Wp-mirror' needs database administrator   |
|              credentials for just two purpose:  1) to grant permissions to  |
|              the `wikiuser' account---permissions to submit create, read,   |
|              update, delete (CRUD) queries to the `wpmirror' database, and  |
|              2) to submit the query 'SHOW ENGINE INNODB STATUS\\G'.         |
|              `Wikiuser' credentials for all other database access.          |
|              Indirectly, one of the MediaWiki maintenance scripts invoked   |
|              by `wp-mirror', namely, `rebuildImages.php', needs database    |
|              administrator credentials.                                     |
| NOTE:        MySQL's out-of-the-box configuration has no password set for   |
|              the database administrator account.  Some people think this is |
|              safe, others deem it unwise.  Please set a password, and enter |
|              it into the MediaWiki configuration file.                      |
| GNU/LINUX:   1) Set the database administrator password:                    |
|              shell$ mysql -u root                                           |
|              mysql> UPDATE mysql.user SET password=PASSWORD('new_pwd')      |
|                     WHERE user='root';                                      |
|              mysql> FLUSH PRIVILEGES;                                       |
|              2) The MediaWiki configuration file is                         |
|                `/etc/mediawiki/AdminSettings.php'                           |
|              Edit the administrator credentials in this file                |
|                 $wgDBadminuser=\"root\";                                    |
|                 $wgDBadminpassword=\"new_pwd\";                             |
| REFERENCE:   See the `wp-mirror' README for more on planning your system.   |
|              See MySQL 5.1 Reference Manual, available on-line.             |
+-----------------------------------------------------------------------------+
"))

(setf (gethash 'assert-database-administrator-account-p *error-messages*)
      (_ "
+-----------------------------------------------------------------------------+
| CATEGORY:    software prerequisite                                          |
| ASSERT:      assert-database-administrator-account-p                        |
| ABSTRACT:    Unable to access database administrator account.               |
| DESCRIPTION: `Wp-mirror' directly accesses the database administrator       |
|              for just one purpose:  to grant permissions to the `wikiuser'  |
|              account---permissions to submit create, read, update, delete   |
|              (CRUD) queries to the `wpmirror' database.                     |
|              `Wp-mirror' uses `wikiuser' credentials for all other database |
|              access.  Indirectly, one of the MediaWiki maintenance          |
|              scripts used by `wp-mirror', namely, `rebuildImages.php',      |
|              needs database administrator credentials.                      |
| REFERENCE:   See the `wp-mirror' README for more on planning your system.   |
+-----------------------------------------------------------------------------+
"))

(setf (gethash 'assert-database-template-or-create *error-messages*)
      (_ "
+-----------------------------------------------------------------------------+
| CATEGORY:    software prerequisite                                          |
| ASSERT:      assert-database-template-or-create                             |
| ABSTRACT:    Unable to find the `wikidb' database.                          |
| DESCRIPTION: `Wp-mirror' uses the newly created `wikidb' database (and its  |
|              mostly empty tables) as a template for creating a separate     |
|              database for each member of *mirror-languages*.                |
|              The template is a `mysqldump' of `wikidb', and it is stored as |
|              `/usr/share/mediawiki/maintenance/template.sql'.               |
| REFERENCE:   See the `wp-mirror' README for more on planning your system.   |
+-----------------------------------------------------------------------------+
"))

(setf (gethash 'assert-database-wikidb-p *error-messages*)
      (_ "
+-----------------------------------------------------------------------------+
| CATEGORY:    software prerequisite                                          |
| ASSERT:      assert-database-wikidb-p                                       |
| ABSTRACT:    Unable to find the `wikidb' database.                          |
| DESCRIPTION: `Wp-mirror' uses the newly created `wikidb' database (and its  |
|              mostly empty tables) as a template for creating a separate     |
|              database for each member of *mirror-languages*.                |
| REFERENCE:   See the `wp-mirror' README for more on planning your system.   |
+-----------------------------------------------------------------------------+
"))

(setf (gethash 'assert-database-wpmirror-p *error-messages*)
      (_ "
+-----------------------------------------------------------------------------+
| CATEGORY:    software prerequisite                                          |
| ASSERT:      assert-database-wpmirror-p                                     |
| ABSTRACT:    Unable to find the `wpmirror' database.                        |
| DESCRIPTION: `Wp-mirror' creates state information about the progress of the|
|              mirroring, and uses InnoDB to store, retrieve and share it.    |
|              This information is stored as records (one per file) in the    |
|              database table `wpmirror.file'.                                |
| REFERENCE:   See the `wp-mirror' README for more on planning your system.   |
+-----------------------------------------------------------------------------+
"))

(setf (gethash 'assert-database-wpmirror-or-create *error-messages*)
      (_ "
+-----------------------------------------------------------------------------+
| CATEGORY:    software prerequisite                                          |
| ASSERT:      assert-database-wpmirror-or-create                             |
| ABSTRACT:    Unable to create the `wpmirror' database.                      |
| DESCRIPTION: `Wp-mirror' creates state information about the progress of the|
|              mirroring, and uses InnoDB to store, retrieve and share it.    |
|              This information is stored as records (one per file) in the    |
|              database table `wpmirror.file'.                                |
| REFERENCE:   See the `wp-mirror' README for more on planning your system.   |
+-----------------------------------------------------------------------------+
"))

(setf (gethash 'assert-database-wpmirror-table-p *error-messages*)
      (_ "
+-----------------------------------------------------------------------------+
| CATEGORY:    software prerequisite                                          |
| ASSERT:      assert-database-wpmirror-table-p                               |
| ABSTRACT:    Unable to find tables in `wpmirror' database.                  |
| DESCRIPTION: `Wp-mirror' creates state information about the progress of the|
|              mirroring, and uses InnoDB to store, retrieve and share it.    |
|              This information is stored as records (one per file) in the    |
|              database table `wpmirror.file'.                                |
| REFERENCE:   See the `wp-mirror' README for more on planning your system.   |
+-----------------------------------------------------------------------------+
"))

(setf (gethash 'assert-database-xxwiki-or-create *error-messages*)
      (_ "
+-----------------------------------------------------------------------------+
| CATEGORY:    software prerequisite                                          |
| ASSERT:      assert-database-xxwiki-or-create                               |
| ABSTRACT:    Unable to find or create one of the `xxwiki' databases.        |
| DESCRIPTION: `Wp-mirror' mirrors needs to be sure that a separate database  |
|              exists for each member of *mirror-languages*.  These databases |
|              must exist before running MediaWiki maintenance scripts, such  |
|              as `importDump.php' which inserts pages from a dump file into  |
|              the database.                                                  |
|              `Wp-mirror' monitors also access these databases to determine  |
|              how many pages, images, etc. have been imported.               |
| REFERENCE:   See the `wp-mirror' README for more on planning your system.   |
+-----------------------------------------------------------------------------+
"))

(setf (gethash 'assert-dbms-mysql-p *error-messages*)
      (_ "
+-----------------------------------------------------------------------------+
| CATEGORY:    software prerequisite                                          |
| ASSERT:      assert-dbms-mysql-p                                            |
| ABSTRACT:    MediaWiki is not using MySQL which `wp-mirror' requires.       |
| DESCRIPTION: `Wp-mirror' creates state information about the progress of the|
|              mirroring, and uses InnoDB to store, retrieve and share it.    |
|              The database is called `wpmirror' and it is accessed using     |
|              the same crediantials (host, user, and password info) that     |
|              MediaWiki uses for its own database `wikidb'.  Indeed,         |
|              `wp-mirror' reads through the MediaWiki configuration file to  |
|              obtain these credentials.  So, at least for time being,        |
|              `wp-mirror' expects MediaWiki to use MySQL.                    |
| DESIGN:      MediaWiki is the software that will maintain and serve the     |
|              pages that you mirror.  MediaWiki is usually configured to     |
|              store its pages and articles in one of five database management|
|              systems (DBMS):  MySQL, `postgres', `sqlite', `mssql', and     |
|              `ibm_db2'.  `Wp-mirror' however uses MySQL only, for reasons   |
|              of performance.                                                |
|              `Wp-mirror' consists of two main parts: mirror and  monitor.   |
|              The mirror maintains state, mostly about the files in its      |
|              working directory, while the monitor(s) read state to present a|
|              set of progress bars.                                          |
|              The design issue is that of concurrency---how to let the mirror|
|              write and the monitor(s) read without data loss or resource    |
|              contention.  Semaphores or locks are used, and one may lock at |
|              different levels:  row, page, extent, table, or database (in   |
|              order from fine to course granularity).  Course granularity    |
|              locks can cause deadlocks and writer starvation.  Not all      |
|              storage engines offer locks of fine granularity.               |
|                o `bdb'      - page-level                                    |
|                o `ibm db2'  - page-level and extent-level                   |
|                o `innodb'   - row-level (MVCC)                              |
|                o `mssql'    - row-level with escalation to page-level and   |
|                               table-level                                   |
|                o `myisam'   - table-level                                   |
|                o `oracle'   - row-level (MVCC)                              |
|                o `postgres' - row-level (MVCC)                              |
|                o `sqlite'   - table-level (actually file-level)             |
|              `Wp-mirror' uses InnoDB because it offers multi-version        |
|              concurrency control (MVCC), that is, row-level locking without |
|              escalation to coarser granularity locks, together with Oracle  |
|              style consistent non-locking reads.                            |
|              Another design issue is that of scalability---how well the     |
|              storage engine performs as the number of records increases     |
|              greatly.  If the `wp-mirror' option `*xchunk-page-count*' is   |
|              left at its default value of 1000, then `wpmirror' will have   |
|              only tens of thousands of records.  If the option is set to 1, |
|              then `wpmirror' will have tens of millions of records.  So it  |
|              is important that the storage engine scale well over several   |
|              orders of magnitude.                                           |
| GNU/LINUX:   The configuration files for `wp-mirror' are                    |
|                `/etc/wp-mirror/default.conf' and `/etc/wp-mirror/local.conf'|
| NOTE:        Support for `postgres' could be added later given evidence of  |
|              significant demand.                                            |
| REFERENCE:   See the `wp-mirror' README for more on planning your system.   |
+-----------------------------------------------------------------------------+
"))

(setf (gethash 'assert-disk-space-p *error-messages*)
      (_ "
+-----------------------------------------------------------------------------+
| CATEGORY:    hardware prerequisite                                          |
| ASSERT:      assert-disk-space-p                                            |
| ABSTRACT:    Insufficient disk space for building mirror of biggest wiki's. |
| DESCRIPTION: For the largest wiki's (the top ten as of 2011), `wp-mirror'   |
|              will not let you start without at least 250 G free disk space. |
| PLANNING:    Disk space requirements for mirroring the latest revisions of  |
|              `en' wiki are approximately (as of 2011):                      |
|                o working directory -  100G                                  |
|                o InnoDB database   -  150G                                  |
|                o images directory  - 1500G                                  |
|              Given the size requirements, it would be best to put this all  |
|              on a separate disk.                                            |
| GNU/LINUX:   The `images' directory is usually                              |
|                `/var/lib/mediawiki/images/'                                 |
|              This can (and should) be a symbolic link to a separate disk.   |
|              The working directory for `wp-mirror' should be placed under   |
|              the images directory.                                          |
| REFERENCE:   See the `wp-mirror' README for more on planning your disk space|
+-----------------------------------------------------------------------------+
"))

(setf (gethash 'assert-hdd-write-cache-disabled-p *error-messages*)
      (_ "
+-----------------------------------------------------------------------------+
| CATEGORY:    hardware prerequisite                                          |
| ASSERT:      assert-hdd-write-cache-disabled-p                              |
| ABSTRACT:    Data could lost during a power failure.                        |
| DESCRIPTION: MediaWiki stores its pages and articles in InnoDB, which is    |
|              MySQL's ACID compliant storage engine that is used for         |
|              transactions.  The issue here regards Durability (the `D' in   |
|              `ACID').  A committed transaction should be stored in a way    |
|              that is resistant to many kinds of failure, including power    |
|              outage.  Transactions that `commit' must actually be written to|
|              disk, and not remain in the hard-disk-drive's write cache,     |
|              where is may be lost during system failure.  It is therefore   |
|              important that the disk's write cache be disabled.             |
| GNU/LINUX:   A hard-drive's write cache can be diabled with a command like: |
|                # hdparm -W0 /dev/sdb                                        |
|              To avoid forgetting, this command should be made part of the   |
|              system boot process.  This can be done by appended it to       |
|                /etc/rc.local                                                |
| REFERENCE:   See the `wp-mirror' README for more on planning your disk space|
+-----------------------------------------------------------------------------+
"))

(setf (gethash 'assert-image-directory-p *error-messages*)
      (_ "
+-----------------------------------------------------------------------------+
| CATEGORY:    software prerequisite                                          |
| ASSERT:      assert-image-directory-p                                       |
| ABSTRACT:    Unable to find image directory for `wp-mirror'.                |
| DESCRIPTION: `Wp-mirror' needs access to the `images' directory so that the |
|              image files that it downloads can found by MediaWiki.          |
|              MediaWiki is the software that will maintain and help serve    |
|              the pages that you mirror.  MediaWiki maintains a directory    |
|              tree where images, thumbs (resized images), and math (formulae |
|              rendered by TeX as PNG images) are stored.                     |
| PLANNING:    Storage requirements for the `images' directory for mirroring  |
|              the latest revisions of `en' wiki are approximately (as of     |
|              2011):                                                         |
|                o image files    -  1T (over a million image files)          |
|                o thumbs         - 50G (several million files)               |
|                o math           -  1G (half a million files)                |
|              The `images' directory should probably be put on a separate    |
|              disk.                                                          |
| GNU/LINUX:   The images directory is usually `/var/lib/mediawiki/images/'.  |
|              This can (and should) be a symbolic link to a separate disk.   |
| REFERENCE:   See the `wp-mirror' README for more on planning your system.   |
+-----------------------------------------------------------------------------+
"))

(setf (gethash 'assert-images-bad-directory-or-create *error-messages*)
      (_ "
+-----------------------------------------------------------------------------+
| CATEGORY:    software prerequisite                                          |
| ASSERT:      assert-images-bad-directory-or-create                          |
| ABSTRACT:    Unable to create bad-images directory for `Wp-mirror'.         |
| DESCRIPTION: Many of the downloaded images are incomplete or corrupt.  A    |
|              directory is needed to sequester these image files.            |
| GNU/LINUX:   The bad-images directory is usually                            |
|                `/var/lib/mediawiki/images/bad-images/'                      |
| REFERENCE:   See the `wp-mirror' README for more on planning your system.   |
+-----------------------------------------------------------------------------+
"))

(setf (gethash 'assert-internet-access-to-wikimedia-site-p *error-messages*)
      (_ "
+-----------------------------------------------------------------------------+
| CATEGORY:    hardware prerequisite                                          |
| ASSERT:      assert-internet-access-to-wikimedia-site-p                     |
| ABSTRACT:    Unable to download wikipedia pages and images.                 |
| DESCRIPTION: Wikipedia pages and articles are made available in the form of |
|              a large compressed `dump' file.  A million image files may also|
|              be downloaded (individually and not as a single dump).         |
| PLANNING:    Bandwidth requirements for mirroring the latest revisions of   |
|              `en' wiki are approximately (as of 2011):                      |
|                o dump file - 10G (one file containing millions of pages)    |
|                o images    -  1T (a million files each containing one image)|
| NOTE:        If this traffic must go through a caching proxy, there may be  |
|              problems.  Some caching proxies crash if a file exceeds its    |
|              available memory.  In this case, the dump file must be manually|
|              copied into `wp-mirror's working directory.  Some caching      |
|              proxies might not have disk space adequate for caching 1TB of  |
|              images.  In this case, the cached images should be removed each|
|              day (perhaps by using a daily `cron' job).  Best of all, is to |
|              by-pass the proxy altogether.  One good method is to use the   |
|              port-forwarding feature of `ssh' to connect to a box outside   |
|              the network served by the proxy.                               |
| REFERENCE:   See the `wp-mirror' README for more on planning your internet  |
|              access.                                                        |
+-----------------------------------------------------------------------------+
"))

(setf (gethash 'assert-mediawiki-adminsettings-p *error-messages*)
      (_ "
+-----------------------------------------------------------------------------+
| CATEGORY:    software prerequisite                                          |
| ASSERT:      assert-mediawiki-adminsettings-p                               |
| ABSTRACT:    Unable to find the admin config file for MediaWiki.            |
| DESCRIPTION: `Wp-mirror' creates state information about the progress of the|
|              mirroring, and uses InnoDB to store, retrieve and share it.    |
|              The database is called `wpmirror' and it is accessed using     |
|              the same crediantials (host, user, and password info) that     |
|              MediaWiki uses for its own database `wikidb'.  Indeed,         |
|              `wp-mirror' reads through the MediaWiki configuration file to  |
|              obtain these credentials.                                      |
| GNU/LINUX:   The configuration file for MediaWiki is                        |
|                `/etc/mediawiki/AdminSettings.php'                           |
| REFERENCE:   See the `wp-mirror' README for more on planning your system.   |
+-----------------------------------------------------------------------------+
"))

(setf (gethash 'assert-mediawiki-extensions *error-messages*)
      (_ "
+-----------------------------------------------------------------------------+
| CATEGORY:    software prerequisite                                          |
| ASSERT:      assert-mediawiki-extensions                                    |
| ABSTRACT:    MediaWiki extensions not installed or enabled.                 |
| DESCRIPTION: Many wikipedia articles use templates to format special content|
|              such as citations, footnotes, and poems.  These are available  |
|              in the `mediawiki-extensions' package.  Without these          |
|              extensions, the templates will fail to render, and you will get|
|              quite a mess consisting of scores of nested braces like        |
|                `{{...}}'.                                                   |
| GNU/LINUX:   Please install the `mediawiki-extensions' package, then go to  |
|                `/etc/mediawiki-extensions/extensions-available/'.           |
|              This directory contains a set of symbolic links.               |
|              Copy all these links to                                        |
|                `/etc/mediawiki-extensions/extensions-enabled/'.             |
|                # cd /etc/mediawiki-extensions/extensions-enabled/           |
|                # cp -a ../extensions-available/* .                          |
| REFERENCE:   See the `wp-mirror' README for more on planning your system.   |
+-----------------------------------------------------------------------------+
"))

(setf (gethash 'assert-mediawiki-import *error-messages*)
      (_ "
+-----------------------------------------------------------------------------+
| CATEGORY:    software prerequisite                                          |
| ASSERT:      assert-mediawiki-import                                        |
| ABSTRACT:    MediaWiki `Import.php' patched version not installed.          |
| DESCRIPTION: MediaWiki v1.15 has a buggy version of `Import.php' that       |
|              causes thousands of error messages like:                       |
|                                                                             |
|              PHP Warning:  xml_parse(): Unable to call handler in_() in     |
|              /usr/share/mediawiki/includes/Import.php on line 437           |
|              PHP Warning:  xml_parse(): Unable to call handler out_() in    |
|              /usr/share/mediawiki/includes/Import.php on line 437           |
|                                                                             |
| GNU/LINUX:   Wp-mirror should automatically download the patched version.   |
|              If it does not, then you may download it from:                 |
|                                                                             |
|              <http://svn.wikimedia.org/viewvc/mediawiki/trunk/phase3/       |
|              includes/Import.php?view=co>                                   |
|                                                                             |
|              Then copy it to `/usr/share/mediawiki/includes/'.              |
| REFERENCE:   See the `wp-mirror' README for more on planning your system.   |
+-----------------------------------------------------------------------------+
"))

(setf (gethash 'assert-mediawiki-localsettings-p *error-messages*)
      (_ "
+-----------------------------------------------------------------------------+
| CATEGORY:    software prerequisite                                          |
| ASSERT:      assert-mediawiki-localsettings-p                               |
| ABSTRACT:    Unable to find the configuration file for MediaWiki.           |
| DESCRIPTION: `Wp-mirror' creates state information about the progress of the|
|              mirroring, and uses InnoDB to store, retrieve and share it.    |
|              The database is called `wpmirror' and it is accessed using     |
|              the same crediantials (host, user, and password info) that     |
|              MediaWiki uses for its own databases.  Indeed, `wp-mirror'     |
|              reads through the MediaWiki configuration files to obtain      |
|              these credentials.                                             |
| GNU/LINUX:   The configuration files for MediaWiki are                      |
|                `/etc/mediawiki/LocalSettings.php' and                       |
|                `/etc/mediawiki/LocalSettings_wp-mirror.php'                 |
| REFERENCE:   See the `wp-mirror' README for more on planning your system.   |
+-----------------------------------------------------------------------------+
"))

(setf (gethash 'assert-mediawiki-localsettings-image *error-messages*)
      (_ "
+-----------------------------------------------------------------------------+
| CATEGORY:    software prerequisite                                          |
| ASSERT:      assert-mediawiki-localsettings-image                           |
| ABSTRACT:    MediaWiki not configured to use `gm' and `rsvg'.               |
| DESCRIPTION: By default MediaWiki converts images (e.g. makes thumbs) by    |
|              using ImageMagick.  Unfortunately, ImageMagick grabs far       |
|              too much scratch space in main memory, and this results in poor|
|              performance and system failures.  Much better is to use        |
|              GraphicsMagick (aka `gm').  Also, scalable vector graphics     |
|              SVG files are converted into PNG files.  It is beter to use    |
|              `rsvg' than the default.                                       |
| GNU/LINUX:   The configuration file for MediaWiki is                        |
|                `/etc/mediawiki/LocalSettings.php'                           |
|              Please install `gm', and if                                    |
|              `/etc/mediawiki/LocalSettings_wpmirror.php' is not installed,  |
|              edit `LocalSettings.php' by appending the lines                |
|        $wgUseImageMagick = false\;                                           |
|        $wgCustomConvertCommand = '/usr/bin/gm convert %s -resize %wx%h %d'\; |
|        $wgSVGConverter = 'rsvg'\;                                            |
| REFERENCE:   See the `wp-mirror' README for more on planning your system.   |
+-----------------------------------------------------------------------------+
"))

(setf (gethash 'assert-mediawiki-localsettings-tex *error-messages*)
      (_ "
+-----------------------------------------------------------------------------+
| CATEGORY:    software prerequisite                                          |
| ASSERT:      assert-mediawiki-localsettings-tex                             |
| ABSTRACT:    MediaWiki not configured to use TeX.                           |
| DESCRIPTION: Many Wikipedia articles contain mathematical formulae.  These  |
|              are best rendered using TeX.  TeX turns the math into small    |
|              PNG image files.  If TeX is not used, then the math will be    |
|              rendered rather poorly using character strings.                |
| GNU/LINUX:   The configuration file for MediaWiki is                        |
|                `/etc/mediawiki/LocalSettings.php'                           |
|              If `/etc/mediawiki/LocalSettings_wpmirror.php' is not installed|
|              then please install the `mediawiki-math' package (Debian), and |
|              edit the MediaWiki configuration file by appending the line    |
|                $wgUseTeX = true\;                                           |
| REFERENCE:   See the `wp-mirror' README for more on planning your system.   |
+-----------------------------------------------------------------------------+
"))

(setf (gethash 'assert-mediawiki-localsettings-tidy *error-messages*)
      (_ "
+-----------------------------------------------------------------------------+
| CATEGORY:    software prerequisite                                          |
| ASSERT:      assert-mediawiki-localsettings-tidy                            |
| ABSTRACT:    MediaWiki not configured to use `tidy'.                        |
| DESCRIPTION: Many Wikipedia articles contain complicated templates that,    |
|              without `tidy', will produce badly formated pages.  In         |
|              particular, you will find `<p>' and `<pre>' tags in the HTML   |
|              after every citation.  The resulting mess will be hard to read.|
|              `Tidy' is an HTML syntax checker and reformatter, and is needed|
|              for generating readable pages.                                 |
| GNU/LINUX:   The relevant configuration file for MediaWiki is               |
|                `/etc/mediawiki/LocalSettings_wpmirror.php'                  |
|              Please install `tidy', and be sure the config file exists and  |
|              contains the line                                              |
|                $wgUseTidy = true\;                                           |
| REFERENCE:   See the `wp-mirror' README for more on planning your system.   |
+-----------------------------------------------------------------------------+
"))

(setf (gethash 'assert-php5-suhosin-p *error-messages*)
      (_ "
+-----------------------------------------------------------------------------+
| CATEGORY:    software prerequisite                                          |
| ASSERT:      assert-php5-suhosin-p                                          |
| ABSTRACT:    PHP5 not configured to use `suhosin'.                          |
| DESCRIPTION: MediaWiki is written in PHP, and is not entirely safe.         |
|              `Suhosin', to paraphrase its documentation, is an advanced     |
|              protection module for PHP5.  It was designed to protect        |
|              servers and users from known and unknown flaws in PHP          |
|              applications and the PHP core.                                 |
|              When you run `importDump.php', you can expect to find many     |
|              alert messages in the `syslog', such as this:                  |
|                                                                             |
|              suhosin[21288]: ALERT - script tried to disable memory_limit by|
|              setting it to a negative value -1 bytes which is not allowed   |
|              (attacker 'REMOTE_ADDR not set', file 'unknown')               |
|                                                                             |
| GNU/LINUX:   The configuration file for `php5-suhosin' is                   |
|                `/etc/php5/conf.d/suhosin.ini'.                              |
| REFERENCE:   See the `wp-mirror' README for more on planning your system.   |
+-----------------------------------------------------------------------------+
"))

(setf (gethash 'assert-physical-memory-p *error-messages*)
      (_ "
+-----------------------------------------------------------------------------+
| CATEGORY:    hardware prerequisite                                          |
| ASSERT:      assert-physical-memory-p                                       |
| ABSTRACT:    Insufficient physical memory found building biggest wiki's.    |
| DESCRIPTION: For the largest wiki's (the top ten as of 2011), `wp-mirror'   |
|              will not let you start without at least 4G main memory.        |
|              Database performance will slow to a crawl without adequate     |
|              memory.                                                        |
|              MediaWiki stores its pages and articles in InnoDB, which is    |
|              MySQL's ACID compliant storage engine.  InnoDB organizes both  |
|              its disk space and its memory into 16K pages.  InnoDB also     |
|              keeps its records sorted.  Since most records are small,       |
|              several can fit in a page.  Accessing a record entails         |
|              accessing the right page.  In worst case, each time a record is|
|              created, updated, or deleted, a page would have to be read from|
|              disk, modified, and written back to disk.  Moreover, as records|
|              are inserted, over-full pages are split up, and as records are |
|              deleted, under-full pages are merged.  So accessing a record   |
|              could entail accessing several pages.  This could cause a      |
|              tremendous amount of disk I/O, which would hurt performance.   |
|              The answer is to keep a large number of pages in memory, so    |
|              that a page can be read in once, then quickly accessed multiple|
|              times before writing it back to disk.  So InnoDB maintains a   |
|              large buffer pool in main memory.  When a page is read from    |
|              disk, it will be put into the buffer pool, where it will stay a|
|              while.  InnoDB also keeps its own administrative data in the   |
|              buffer pool.  So the larger the buffer pool, the better.       |
| PLANNING:    Install at least 4G DRAM onto your mother board.  Install much |
|              more if you can afford it.  If your mother board does not      |
|              accept that much, then replace it (for it is antiquated).      |
| NOTE:        Modern CPUs and operating systems can organize main memory into|
|              pages of different sizes.  For example, while Intel usually    |
|              organizes memory into 4K pages, it also offers `huge pages'    |
|              that are 2M or 4M.  Modern operating systems will swap the     |
|              smaller pages to a swap space on disk, according to some aging |
|              algorythm, but will leave the `huge pages' alone.  Given that  |
|              InnoDB has its own memory management algorythm, there is no    |
|              advantage to having the OS to swap anything held in the buffer |
|              pool---and actually swapping would hurt performance.           |
|              So it is highly recommended (though not currently required) to |
|              allocate at least 3G of memory for `huge pages', and to        |
|              configure InnoDB to use them for its buffer pool.              |
|              Leave at least 1G memory for other processes.  Importing the   |
|              dump into `wikidb' entails creating thumbs (resized images).   |
|              As some image files are over 100M, the resizing can easily     |
|              occupy memory several times that amount.                       |
| REFERENCE:   See the `wp-mirror' README for more on planning your memory.   |
+-----------------------------------------------------------------------------+
"))

(setf (gethash 'assert-utilities-p *error-messages*)
      (_ "
+-----------------------------------------------------------------------------+
| CATEGORY:    software prerequisite                                          |
| ASSERT:      assert-utilities-p                                             |
| ABSTRACT:    Unable to find all of the software utilities required.         |
| DESCRIPTION: `Wp-mirror' makes use of a couple dozen existing software      |
|              utilities.  This was due to a design decision akin to `Do not  |
|              reinvent the wheel'.  `Wp-mirror' will not let you start if any|
|              of these utilities are missing.                                |
| GNU/LINUX:   A package management system (e.g. `Debian') can provide easy   |
|              installation and maintenance of a consistent set of utilities. |
| REFERENCE:   See the `wp-mirror' README for more on planning your system.   |
+-----------------------------------------------------------------------------+
"))

(setf (gethash 'assert-working-directory-or-create *error-messages*)
      (_ "
+-----------------------------------------------------------------------------+
| CATEGORY:    software prerequisite                                          |
| ASSERT:      assert-working-directory-or-create                             |
| ABSTRACT:    Unable to create working directory for `Wp-mirror'.            |
| DESCRIPTION: `Wp-mirror' uses a directory to generate and maintain tens of  |
|              thousands of files used by the mirroring process.              |
| PLANNING:    Storage requirements for the working directory for mirroring   |
|              the latest revisions of `en' wiki are approximately (as of     |
|              2011):                                                         |
|                o dump file              - 10G                               |
|                o decompressed dump file - 50G                               |
|                o xchunks                - 50G                               |
|                o ichunks                -  5G                               |
| GNU/LINUX:   The working directory is usually                               |
|                `/var/lib/mediawiki/images/wp-mirror/'                       |
| REFERENCE:   See the `wp-mirror' README for more on planning your system.   |
+-----------------------------------------------------------------------------+
"))

(setf (gethash 'assert-virtual-host-p *error-messages*)
      (_ "
+-----------------------------------------------------------------------------+
| CATEGORY:    software prerequisite                                          |
| ASSERT:      assert-virtual-host-p                                          |
| ABSTRACT:    Virtual host `mediawiki.site' not found.                       |
| DESCRIPTION: `Wp-mirror' sets up a virtual host to let you access your new  |
|              mirrors.  For example, you access your mirror the `simple' wiki|
|              by giving your browser <http://simple.mediawiki.site/>.        |
|              The virtual host container should be found in                  |
|              `/etc/apache2/sites-available/mediawiki.site.conf'.            |
|              There should also be a symbol link to it found in              |
|              `/etc/apache2/sites-enabled/mediawiki.site.conf'.              |
|              And the virtual host configuration should be loaded into       |
|              Apache2.                                                       |
|                                                                             |
|              # apache2ctl -S |& grep mediawiki                              |
|                       port 80 namevhost mediawiki.site (/etc/apache2/sites-e|
|              nabled/mediawiki.site.conf:1)                                  |
|                                                                             |
| REFERENCE:   See the `wp-mirror' README for more on planning your system.   |
+-----------------------------------------------------------------------------+
"))

(setf (gethash 'assert-virtual-host-name-resolution-p *error-messages*)
      (_ "
+-----------------------------------------------------------------------------+
| CATEGORY:    software prerequisite                                          |
| ASSERT:      assert-virtual-host-name-resolution-p                          |
| ABSTRACT:    Virtual host `mediawiki.site' not found in `/etc/hosts'.       |
| DESCRIPTION: Apache2 needs to resolve the name `mediawiki.site' into the IP |
|              address `::1' (or 127.0.0.1 if you prefer IPv4).  This can be  |
|              done with a DNS server, such as BIND, but that is atypical for |
|              a PC.  Name resolution is most easily done by appending lines  |
|              to `/etc/hosts' like these:                                    |
|                                                                             |
|              ::1 locahost mediawiki.site www.mediawiki.site                 |
|              ::1 meta.mediawiki.site simple.mediawiki.site                  |
|              ::1 en.mediawiki.site zh.mediawiki.site                        |
|                                                                             |
| REFERENCE:   See the `wp-mirror' README for more on planning your system.   |
+-----------------------------------------------------------------------------+
"))


;;;;--------------------------------------------------------------------------+
;;;; Warning messages detailed:                                               |
;;;;--------------------------------------------------------------------------+


(setf (gethash 'warn-if-detect-proxy *error-messages*)
      (_ "
+-----------------------------------------------------------------------------+
| CATEGORY:    software warning                                               |
| ASSERT:      warn-if-detect-proxy                                           |
| ABSTRACT:    Web proxy detected.  Manually check if it can handle traffic.  |
| DESCRIPTION: Caching web proxies offer many benefits: economize band-width, |
|              reduce latency, improve security, log web traffic, block       |
|              inappropriate sites, and bridge between IPv6 and IPv4 networks.|
|              They also have weaknesses: unable to handle downloading very   |
|              large files, and need large disk space for cache.              |
| PLANNING:    1) Dump file.  Dump files containing the latest revisions of   |
|              `en' wiki are 8G (as of 2011).  Some caching web proxies crash |
|              when downloading a file larger than their available memory.    |
|              You may need to obtain the dump file some other way, then      |
|              manually copy it to `wp-mirror's working directory.            |
|              2) Images.  `En' wiki references over a million image files,   |
|              which collectively occupy about 1T (as of 2011).  Downloading  |
|              the image files may stuff your proxy's cache.  This can be     |
|              managed by deleting image files from the cache on a daily basis|
|              (using a `cron' job say).                                      |
| GNU/LINUX:   The `wp-mirror' working directory is usually                   |
|                `/var/lib/mediawiki/images/wp-mirror/'                       |
| REFERENCE:   See the `wp-mirror' README for more on caching web proxies.    |
+-----------------------------------------------------------------------------+
"))


;;;;--------------------------------------------------------------------------+
;;;; Assert prerequisite hardware:                                            |
;;;;--------------------------------------------------------------------------+


(defun assert-disk-space-p ()
  "Asserting adequate disk space"
  (let* ((fs (posix:stat-vfs *whereis-directory-images*)) ; this is a struct
	 (bfree        (stat-vfs-bfree fs))               ; accessor
	 (bsize        (stat-vfs-bsize fs))               ; accessor
	 (disk-space   (* bfree bsize))
	 (large-wiki-p (and (intersection *mirror-languages*
					  *wikimedia-large-wiki-list*
					  :test #'string=)
			    t)))
    (debug-message-value (_ "free blocks") (stat-vfs-bfree fs))
    (debug-message-value (_ "block size")  (stat-vfs-bsize fs))
    (put-message-value (_ "detecting") 
		       (format-integer-for-human disk-space))
    ;; assert if we have one of the largest wiki's
    (put-message-value (_ "our wiki languages are") *mirror-languages*)
    (put-message-value (_ "large wiki's are") *wikimedia-large-wiki-list*)
    (put-message-value (_ "are any of our wiki's large") large-wiki-p)
    (if large-wiki-p
	(progn
	  (put-message-value (_ "requiring minimum") (format-integer-for-human
						      *system-disk-space-min*))
	  (>= disk-space *system-disk-space-min*))
      t)))

(defun assert-hdd-write-cache-disabled-p ()
  "Asserting IDE/SATA drive write cache is disabled"
  ;;
  ;; Design note:
  ;;
  ;; 1) Hard disks have a write cache in order to schedule the
  ;; movements of the read-write heads more efficiently, using
  ;; something like the `elevator algorythm'.
  ;;
  ;; Disabling disk write cache is however important for an ACID
  ;; compliant DBMS.  The issue here is Durability (the `D' in
  ;; `ACID').  Transactions that `commit' must actually be written to
  ;; disk, and not remain in cache, where they may be lost during
  ;; system failure.
  ;;
  ;; 2) In step 1 below, MySQL is queried as to its InnoDB data
  ;; file path.  This means that the hard drive is identified after
  ;; MySQL credentials have been asserted.  Simply put, the H/W
  ;; feature is asserted after the S/W feature.  An alternative design
  ;; would have been to parse the MySQL configuration file for
  ;; information regarding the InnoDB data file path.  Then the H/W
  ;; feature could have been asserted before MySQL was.  With
  ;; `GNU/Linux', the config file is `/etc/mysql/conf.d/custom.cnf'.
  ;;
  ;; 3) In step 6 below, `/sys' is used to identify the hard drive
  ;; underlying the InnoDB data file.  `/sys' exists to make this
  ;; kind of task much easier.  For a prudent disk storage solution
  ;; features [LVM2 over LUKS over RAID over whole disks], and without
  ;; `/sys' this function could have been quite a chore to write.
  ;;
  (let ((innodb-data-file-path      nil)
	(partition-or-file          nil)
	(absolute-path              nil)
	(device-id                  nil)
	(device-id-major            nil)
	(device-id-minor            nil)
	(device-name                nil)
	(dev-path                   (make-pathname :directory "dev"))
	(physical-block-device-list nil)
	(write-cache-disabled-p     t))
    ;; 1) Get InnoDB data file path, usually `/var/lib/mysql/ibdata0'
    (setq innodb-data-file-path (sql-show-innodb-data-file-path :put t))
    (put-message-value (_ "innodb-data-file-path") innodb-data-file-path)

    ; my database uses raw partition
    ; need code for debugging the :file branch (don't use)
    ;(setq innodb-data-file-path "/var/lib/mysql/ib_logfile0")
    ;(put-message-value (_ "innodb-data-file-path") innodb-data-file-path)

    ;; 2) path may be a symbolic link---chase it
    (setq absolute-path (ext:probe-pathname innodb-data-file-path))
    (put-message-value (_ "absolute path") absolute-path)

    ;; 3) Innodb data file path will be either a file or a partition
    (if (zerop (posix:file-size absolute-path))
	(setq partition-or-file :partition)
      (setq  partition-or-file :file))
    
    (case partition-or-file
      (:file 
       ;; 4a) find device ID of device containing file (but it may be mapped)
       (setq device-id (file-stat-dev (posix:file-stat absolute-path)))
       (put-message-value (_ "device-id") device-id)
       (multiple-value-setq (device-id-major device-id-minor)
	 (floor device-id 256))
       ;; 4b) set initial block-device and dir-list
       (setq block-device-init (format nil "~d:~d"
				       device-id-major device-id-minor))
       (put-message-value (_ "block-device-init") block-device-init)
       (setq dir-list-init (append '(:absolute "sys" "dev" "block")
				   (list block-device-init)))
       (put-message-value (_ "dir-list-init") dir-list-init))
      (:partition
       ;; 5a) assert `/dev/foo', block device is `foo' (but it may be mapped)
       (debug-message-value (_ "dev directory")  (pathname-directory dev-path))
       (debug-message-value (_ "pathname-directory data")
			    (pathname-directory absolute-path))
       (debug-message-value (_ "pathname-name data")
			    (pathname-name absolute-path))
       (unless (equal (pathname-directory absolute-path)
		      (pathname-directory dev-path))
	 (put-message-value-fail-and-abort (_ "should be block device")
					   absolute-path))
       ;; 5b) set initial block-device and dir-list
       (setq block-device-init (pathname-name absolute-path))
       (put-message-value (_ "block-device-init") block-device-init)
       (setq dir-list-init (append '(:absolute "sys" "block")
				   (list block-device-init)))
       (put-message-value (_ "dir-list-init") dir-list-init)))

    ;; 6) now go to `/sys', find block device, and chase slaves
    (setq physical-block-device-list
	  (sys-pathname-slaves
	   (make-pathname :directory dir-list-init)))
    (put-message (_ "-----sys-pathname-slaves-----done-----"))
    (put-message-value (_ "physical-block-device-list")
		       physical-block-device-list)
    (let* ((device-name-list
	    (mapcar #'(lambda (hdd)
			(format nil "/dev/~a"
				(first (last (pathname-directory hdd)))))
		    physical-block-device-list))
	   (write-cache-disabled-p-list
	    (mapcar #'(lambda (device-name)
			(shell-hdparm-write-cache-disabled-p device-name))
		    device-name-list)))
      (put-message-value (_ "device-name-list") device-name-list)
      (put-message-value (_ "write-cache-disabled-p-list") 
			 write-cache-disabled-p-list)
      (setq write-cache-disabled-p
	    (notany #'null write-cache-disabled-p-list)))
    (put-message-value (_ "are all write caches disabled") write-cache-disabled-p)
    ;(die) ; uncomment when debug
    write-cache-disabled-p))

(defun sys-pathname-slaves (sys-pathname)
  "Determining `/sys' pathnames of slaves below given `/sys' pathname"
  ;;
  ;; Design note:
  ;;
  ;; A given block device may be mapped to several slaves (e.g. RAID).
  ;; This means `/sys/block/<dev>/slaves/<dev>/slaves/...'  has the
  ;; topology of a tree.  The leaves of the tree (i.e. devices with no
  ;; slaves) are the physical block devices.
  ;;
  ;; This function assumes that the given `sys-block-pathname' starts
  ;; somewhere within the `/sys' filesystem, perhaps `/sys/block' or
  ;; `/sys/dev/block'.  The function chases slaves from there by tree
  ;; search.  This tree search is done using recursion.  The function
  ;; returns a list of sys-block-pathnames to physical block devices.
  ;;
  (put-message (_ "-----sys-pathname-slaves-----"))
  (put-message-value (_ "sys-pathname") sys-pathname)
  (let* ((dir-list               (pathname-directory sys-pathname))
	 (block-device           (first (last dir-list)))
	 (dir-list-wild          (append dir-list (list "slaves" :wild)))
	 (sys-pathname-wild      (make-pathname :directory dir-list-wild))
	 (sys-pathname-next-list (directory sys-pathname-wild)))
    (unless (and (eql     (first  dir-list) :absolute)
		 (string= (second dir-list) "sys"))
      (put-message-fail-and-abort (_ "invalid sys-pathname")))
    (put-message-value (_"block-device")           block-device)
    (put-message-value (_"dir-list")               dir-list)
    (put-message-value (_"dir-list-wild")          dir-list-wild)
    (put-message-value (_"sys-pathname-wild")      sys-pathname-wild)
    (put-message-value (_"sys-pathname-next-list") sys-pathname-next-list)
    (if (null sys-pathname-next-list)
	(progn
	  (put-message-value (_ "found physical device") sys-pathname)
	  (list sys-pathname))
      (progn
	(put-message (_ "recursing over slaves"))
	;; gather the leaves of the tree
	(reduce #'append
		(mapcar #'sys-pathname-slaves sys-pathname-next-list))))))

(defun assert-internet-access-to-wikimedia-site-p (&key (put nil))
  "Asserting internet access to wikimedia site"
  ;; Bug: actually this will not detect a failure if: 
  ;; a) your traffic is going through a caching web proxy that, 
  ;; b) has already cached a copy of the page, and 
  ;; c) then suffers a failure lying between the proxy and wikimedia
  (unless (null put)
    (put-message-value (_ "fetching HTTP-header from") *wikimedia-site*))
  (let ((result (shell-curl-http-head *wikimedia-site* *whereis-tmp-http*
				      :put put))
	(tmp-head (format nil "~a.head" *whereis-tmp-http*)))
    (and (null result)
	 (assert-file-exists-p tmp-head :put put))))

(defun assert-physical-memory-p ()
  "Asserting adequate physical memory"
  (let ((phys-mem     (os:physical-memory))
	(large-wiki-p (and (intersection *mirror-languages*
					 *wikimedia-large-wiki-list*
					 :test #'string=)
			   t)))
    (put-message-value (_ "detecting")
		       (format-integer-for-human phys-mem))
    ;; assert if we have one of the largest wiki's
    (put-message-value (_ "our wiki languages are") *mirror-languages*)
    (put-message-value (_ "large wiki's are") *wikimedia-large-wiki-list*)
    (put-message-value (_ "are any of our wiki's large") large-wiki-p)
    (if large-wiki-p
	(progn
	  (put-message-value (_ "requiring minimum")
			     (format-integer-for-human
			      *system-physical-memory-min*))
	  (>= phys-mem *system-physical-memory-min*))
      t)))

(defun count-cpu ()
  "Counting CPU's"
  ;; number of cpu's useful for determining how many mirror processes
  ;; to run concurrently
  (let* ((directory-list '(:absolute "sys" "devices" "system" "cpu"))
	 (file-name      "present")
	 (cpu-pathname   (make-pathname :directory directory-list
					:name file-name))
	 (cpu-line       nil)
	 (cpus           nil))
    ;; "present" will contain a line like "0-1" (meaning two cpus)
    (with-open-file (f cpu-pathname)
      (setq cpu-line (read-line f)))     ; "0-1"
    (with-input-from-string (s (first    ; "1"
				(last
				 (regexp:regexp-split "-" cpu-line))))
      (setq cpus (1+ (read s))))         ; 2
    (if (and (numberp cpus) (>= cpus 1))
	(put-message-value-done (_ "number of CPU's found") cpus)
      (progn
	(put-message-value (_ "number of CPU's found") cpus)
	(setq cpus 1)
	(put-message-value-done (_ "set CPU's to") cpus)))
    (setq *system-cpu* cpus)
    cpus))


;;;;--------------------------------------------------------------------------+
;;;; Assert prerequisite software:                                            |
;;;;--------------------------------------------------------------------------+


(defun assert-clisp-features-p ()
  "Asserting clisp features"
  (assert-test-set-p
   #'(lambda (x) (member x *features*))
   '(:asdf2 :asdf :clisp :clx :gettext :i18n :loop :regexp :screen :syscalls)))

(defun assert-concurrency-limit-xchunk-p ()
  "Limiting concurrency of importing `x-chunk's"
  (let ((innodb-file-format (sql-show-innodb-file-format :put nil)))
    (put-message-value (_ "innodb-file-format") innodb-file-format)
    (setq *mirror-concurrency-limit-xchunk* 
	  (cond ((string= innodb-file-format "Antelope")
		 (put-message (_ "Antelope handles concurrency well (lesser of number of CPUs or 3)"))
		 (min *system-cpu* 3))
		((string= innodb-file-format "Barracuda")
		 (put-message (_ "Barracuda does not handle concurrency well - frequent deadlocks"))
		 1)
		(t nil)))
    (put-message-value (_ "*mirror-concurrency-limit-xchunk*")
		       *mirror-concurrency-limit-xchunk*))
  *mirror-concurrency-limit-xchunk*)

(defun assert-curl-max-time-p ()
  "Asserting `curl' config file has --max-time option"
  ;;
  ;; Design note:
  ;;
  ;; `Wikix' generates image download scripts that incorrectly handle
  ;; the case of partial downloads.  `http://upload.wikimedia.org/'
  ;; has thousands of images that download only part way, and then
  ;; block.  Without a timeout, the mirror would never finish.
  ;;
  ;; `Curl' can be configured with a timeout.  The option is
  ;; `--max-time <seconds>'.  For a 30 minute timeout, we can edit
  ;; `--max-time 1800' into the scripts, or append the option to the
  ;; curl configuration file `/root/.curlrc'.  This author recommends
  ;; appending the following lines:
  ;;
  ;; # We want a 1 ks timeout (initially and for each retry)
  ;; --max-time 1000
  ;;
  (if (assert-file-exists-p *whereis-curlrc*)
      (let ((l (grep-first-non-comment "--max-time" *whereis-curlrc*)))
	(if (null l)
	    (put-message-fail (_ "probing curl configuration file"))
	  (put-message-value (_ "probing curl configuration file") l))
	l)
    (progn
      (put-message-fail (_ "file not found"))
      nil)))

(defmacro assert-directory-file-p (dir-name file-name)
  "Asserting directory/file"
  `(assert-file-exists-p
    (merge-pathnames
     (parse-namestring ,dir-name)
     (parse-namestring ,file-name))))

(defun assert-mediawiki-adminsettings-p ()
  "Asserting mediawiki admin settings"
  (assert-directory-file-p *whereis-directory-mediawiki*
			   *whereis-file-mediawiki-adminsettings*))

(defun assert-mediawiki-import ()
  "Asserting mediawiki `Import.php' exists and is newer than v1.15"
  (and
   (assert-directory-file-p *whereis-directory-mediawiki-includes*
			    *whereis-file-mediawiki-import*)
   (>
    (posix:file-size (merge-pathnames
		      (parse-namestring *whereis-directory-mediawiki-includes*)
		      (parse-namestring *whereis-file-mediawiki-import*)))
    30934)))

(defun assert-mediawiki-localsettings-p ()
  "Asserting mediawiki local settings"
  (assert-directory-file-p *whereis-directory-mediawiki*
			   *whereis-file-mediawiki-localsettings*))

(defun assert-mediawiki-localsettings-wpmirror-p ()
  "Asserting mediawiki local settings supplemented by wp-mirror"
  (assert-directory-file-p *whereis-directory-mediawiki*
			   *whereis-file-mediawiki-localsettings-wpmirror*))

(defun parse-adminsettings (key)
  "Parsing mediawiki admin settings for `key = value'"
  (parse-mediawiki-configuration-file
   *whereis-file-mediawiki-adminsettings* key))

(defun parse-localsettings (key)
  "Parsing mediawiki local settings for `key = value'"
  (parse-mediawiki-configuration-file
   *whereis-file-mediawiki-localsettings* key))

(defun parse-localsettings-wpmirror (key)
  "Parsing mediawiki local settings (`wp-mirror' supplement) for `key = value'"
  (parse-mediawiki-configuration-file
   *whereis-file-mediawiki-localsettings-wpmirror* key))

(defun parse-mediawiki-configuration-file (file-path key)
  "Parsing mediawiki local settings file for `key = value'"
  (parse-dir-file-key *whereis-directory-mediawiki* file-path key))

(defmacro parse-dir-file-key (dir-name file-name key)
  "Parsing dir/file for `key = value'"
  `(let* ((file-path (merge-pathnames
		      (parse-namestring ,dir-name)
		      (parse-namestring ,file-name)))
	  (l   (grep-first-non-comment ,key file-path))
	  (m   (concatenate 'string (_ "probing") " " key))
	  (rhs      nil)
	  (val-type nil)
	  (val      nil))
     (if (null l)
	 (put-message-fail m)
       (progn
	 (setq rhs (second (regexp:regexp-split "=" l))) ; right-hand side
	 ;; rhs can be `"foo";' (where foo is a string)
	 ;; rhs can be  `foo;'  (where foo is `true', `false', numeric)
	 (if (regexp:match "['\"]" rhs)
	     (setq val-type :string)
	   (setq val-type :other))
	 (case val-type
	   (:string (setq val (second (regexp:regexp-split "['\"]" rhs))))
	   (:other  (setq val (string-trim
			       `(#\Space #\Tab #\Newline)
			       (first  (regexp:regexp-split ";"  rhs))))))
	 (if (regexp:match "password" key)
	     (put-message-value m "********")
	   (put-message-value m val))))
     val))

(defun assert-database-administrator-credentials-p ()
  "Asserting mediawiki has database credentials"
  (defparameter *db-type*           (parse-localsettings "$wgDBtype"))
  (defparameter *db-server*         (parse-localsettings "$wgDBserver"))
  (defparameter *db-name*           (parse-localsettings "$wgDBname"))
  (defparameter *db-wiki-user*      (parse-localsettings "$wgDBuser"))
  (defparameter *db-wiki-password*  (parse-localsettings "$wgDBpassword"))
  (defparameter *db-admin-user*     (parse-adminsettings "$wgDBadminuser"))
  (defparameter *db-admin-password* (parse-adminsettings "$wgDBadminpassword"))
  t)

(defun assert-database-administrator-account-p (&key (put nil))
  "Asserting access to database administrator account"
  ;; assert access to DB administrator account
  (not (null (sql-show-databases :user *db-admin-user* :put put))))

(defun assert-database-p (database-name &key (put nil))
  "Asserting database-name exists"
  ;; assert existance of database-name
  (dolist (l (sql-show-databases :user *db-wiki-user* :put put) nil)
    (when (regexp:match database-name l)
      (return t)))) ; break do loop

(defun assert-database-tables-p (db-name &key (put nil))
  "Asserting database.table"
  (and (assert-database-p db-name :put put)
       (sql-show-tables db-name :put put)))

(defun database-name (language-code)
  "Determining database name given language-code"
  (format nil "~awiki" (language-to-file-name language-code)))

(defun assert-database-template-or-create (&key (put nil))
  "Asserting or creating a `mysqldump' of `wikidb'"
  (let ((file-path (merge-pathnames
		    (parse-namestring *whereis-directory-mediawiki-maintenance*)
		    (parse-namestring *whereis-file-wikidb-template*))))
    (if (file-exists-p file-path)
	(put-message (_ "database template found"))
      (let ((message   (_ "creating database template"))
	    (result  nil))
	(put-message-start message)
	(setq result (shell-mysqldump-to-file *db-name* file-path))
	(if (null result)
	    (put-message-done message)
	  (put-message-fail-and-abort message))))
    (file-exists-p file-path)))

(defun assert-database-wikidb-p (&key (put nil))
  "Asserting database `wikidb'"
  (assert-database-p *db-name* :put put))

(defun assert-database-wpmirror-p (&key (put nil))
  "Asserting database `wpmirror'"
  (assert-database-p *db-wp-mirror-name* :put put))

(defun assert-database-wpmirror-or-create (&key (put nil))
  "Asserting database `wpmirror' or create it"
  (if (assert-database-wpmirror-p :put put)
      (put-message (_ "database found"))
    (progn
      ;; grant
      (put-message-start (_ "granting privileges to *db-wiki-user*"))
      (sql-grant-all *db-wp-mirror-name* :put put)
      (put-message-value-done (_ "granting privileges to *db-wiki-user*")
			      *db-wiki-user*)
      ;; database
      (put-message-start (_ "creating *db-wp-mirror-name*"))
      (sql-create-database *db-wp-mirror-name* :put put)
      (put-message-value-done (_ "creating *db-wp-mirror-name*") 
			      *db-wp-mirror-name*)))
  (if (assert-database-wpmirror-table-p :put put)
      (put-message (_ "table found"))
    (progn
      ;; table `file'
      (put-message-start (_ "creating table `file'"))
      (sql-create-table-file :put put)
      (put-message-value-done (_ "creating table") "file")
      ;; table `image'
      (put-message-start (_ "creating table `image'"))
      (sql-create-table-image :put put)
      (put-message-value-done (_ "creating table") "image")))
  (assert-database-wpmirror-table-p :put put))

(defun assert-database-wpmirror-table-p (&key (put nil))
  "Asserting database `wpmirror' has a table"
  (assert-database-tables-p *db-wp-mirror-name* :put put))

(defun assert-database-xxwiki-or-create (&key (put nil))
  "Asserting database exists for each language"
  (loop
    for language-code in *mirror-languages*        ; stepped first
    as  db-name = (database-name language-code)    ; stepped second
    do
    (if (assert-database-p db-name :put put)
	(put-message-value (_ "database found for") db-name)
      (progn
	;; grant
	(put-message-value (_ "granting privileges for") db-name)
	(sql-grant-all db-name :put put)
	(put-message-value-done (_ "granting privileges to") *db-wiki-user*)
	;; database
	(put-message-value (_ "creating database") db-name)
	(sql-create-database db-name :put put)
	(put-message-value-done (_ "creating database") db-name)))
    (if (assert-database-tables-p db-name :put put)
	(put-message-value (_ "tables found for") db-name)
      (progn
	;; tables 
	(put-message-start (_ "loading database template"))
	(shell-mysql-load-database 
	 db-name
	 (format nil "~a~a" *whereis-directory-mediawiki-maintenance*
		 *whereis-file-wikidb-template*))
	(put-message-value-done (_ "loading database template for") db-name)
	t))
    always (assert-database-tables-p db-name :put put)))

(defun assert-dbms-mysql-p ()
  "Asserting DBMS MySQL"
  (string= "mysql" *db-type*))

(defun assert-directory-exists-or-create (directory-name)
  "Asserting directory exists, or create it"
  (put-message-value (_ "probing directory") directory-name)
  (or (ext:probe-pathname directory-name)
      (progn
	(put-message-start (_ "creating directory"))
	(ext:make-directory directory-name)
	(ext:probe-pathname directory-name))))

(defun assert-directory-exists-p (directory-name)
  "Asserting directory exists or die"
  (put-message-value (_ "probing directory") directory-name)
  (ext:probe-pathname directory-name))

(defun file-exists-p (file-name &key (put t))
  "Determining if file exists and not empty."
  (and (ext:probe-pathname file-name)
       (let ((fsize (posix:file-size file-name)))
	 (unless (null put)
	   (put-message-value (_ "File size")
			      (format-integer-for-human fsize)))
	 (> fsize 0))))

(defun assert-file-exists-or-create (file-name obj)
  "Asserting file exists, or create it"
  (put-message-value (_ "probing file") file-name)
  (or (ext:probe-pathname file-name)
    (progn
      (put-message-value (_ "creating file") file-name)
      (with-open-file (f file-name
			 :direction :output
			 :if-exists :supersede)
	(prin1 obj f))
      (ext:probe-pathname file-name))))

(defun assert-file-exists-p (file-name &key (put t))
  "Asserting file exists and not empty, or die."
  (unless (null put)
    (put-message-value (_ "probing") file-name))
  (file-exists-p file-name))

(defun assert-image-directory-p ()
  "Asserting image directory"
  (assert-directory-exists-p *whereis-directory-images*))

(defun assert-images-bad-directory-or-create ()
  "Asserting or creating a bad-images directory"
  (let ((pathname-bad (parse-namestring *whereis-directory-images-bad*)))
    (unless (ext:probe-directory pathname-bad)
      (ext:make-directory pathname-bad)
      (loop 
	with hex = (the simple-string "0123456789abcdef") 
	for c1 of-type standard-char across hex
	as  d1 of-type pathname =
	    (make-pathname :directory (append
				       (pathname-directory pathname-bad) 
				       (list (format nil "~a" c1))))
	do
	  (ext:make-directory d1)
	  (loop
	    for c2 of-type standard-char across hex
	    as  d2 of-type pathname =
	        (make-pathname :directory (append
					   (pathname-directory d1)
					   (list (format nil "~a~a" c1 c2))))
	    do
	      (ext:make-directory d2))))
    (ext:probe-directory pathname-bad)))

(defun assert-mediawiki-extensions ()
  "Asserting mediawiki extensions"
  ;; mediawiki extensions are templates for formatting articles
  (let ((ext-dir          (ext:probe-pathname
			   *whereis-directory-mediawiki-extensions*))
	(ext-available    (ext:probe-pathname
			   *whereis-directory-mediawiki-extensions-available*))
	(ext-enabled      (ext:probe-pathname
			   *whereis-directory-mediawiki-extensions-enabled*))
	(ext-available-wild nil)
	(ext-enabled-wild   nil)
	(num-available      0)
	(num-enabled        0))
    (put-message-value (_ "mediawiki-extensions") 
			*whereis-directory-mediawiki-extensions*)
    (put-message-value (_ "mediawiki-extensions-available") 
			*whereis-directory-mediawiki-extensions-available*)
    (put-message-value (_ "mediawiki-extensions-enabled") 
			*whereis-directory-mediawiki-extensions-enabled*)
    (unless (and ext-dir ext-available ext-enabled)
      (put-message-fail (_ "mediawiki extensions not found")))
    (unless (null ext-available)
      (setq ext-available-wild
	    (make-pathname
	     :directory (pathname-directory ext-available)
	     :name      :wild))
      (put-message-value (_ "ext-available-wild") ext-available-wild)
      (setq num-available (length (directory ext-available-wild)))
      (put-message-value (_ "number of extensions available") num-available))
    (unless (null ext-enabled)
      (setq ext-enabled-wild
	    (make-pathname
	     :directory (pathname-directory ext-enabled)
	     :name      :wild))
      (put-message-value (_ "ext-enabled-wild") ext-enabled-wild)
      (setq num-enabled (length (directory ext-enabled-wild)))
      (put-message-value (_ "number of extensions enabled")   num-enabled))
    (when (or (eql num-available 0)
	      (eql num-enabled   0)
	      (not (eql num-available num-enabled)))
      (put-message-fail (_ "mediawiki-extensions not enabled")))
    (and ext-dir ext-available ext-enabled
	 (not (or (eql num-available 0)
		  (eql num-enabled   0)
		  (not (eql num-available num-enabled)))))))

(defun assert-mediawiki-localsettings-image ()
  "Asserting mediawiki localsettings for image file conversion"
  (let ((val-imagemagick (parse-localsettings-wpmirror "$wgUseImageMagick"))
	(val-custom      (parse-localsettings-wpmirror "$wgCustomConvertCommand"))
	(val-svg         (parse-localsettings-wpmirror "$wgSVGConverter")))
    (and val-imagemagick
	 val-custom
	 val-svg
	 (regexp:match "false" val-imagemagick)
	 (regexp:match "/usr/bin/gm convert %s -resize %wx%h %d" val-custom)
	 (regexp:match "rsvg" val-svg))))

(defun assert-mediawiki-localsettings-tex ()
  "Asserting mediawiki localsettings for `TeX'"
  (let ((val (parse-localsettings-wpmirror "$wgUseTeX")))
    (and val
	 (regexp:match "true" val))))

(defun assert-mediawiki-localsettings-tidy ()
  "Asserting mediawiki localsettings for `tidy'"
  (let ((val (parse-localsettings-wpmirror "$wgUseTidy")))
    (and val
	 (regexp:match "true" val))))

(defun assert-php5-suhosin-p ()
  "Asserting PHP5 suhosin"
  (assert-file-exists-p *whereis-php5-suhosin*))

(defun assert-test (f &rest args)
  "Wrapping evaluation of a boolean valued function"
  (let ((doc (_ (documentation f 'function))))
    (put-message-start doc)
    (let ((res (apply f args)))
      (if res
	  (put-message-done doc)
	(progn
	  (put-message (gethash f *error-messages*))
	  (put-message-fail doc)))
      res)))

(defun assert-test-set (function-list)
  "Running tests"
  (let ((message (_ "logging test set"))
	(all-pass t))
    (dolist (func function-list all-pass)
      (setq all-pass (and (assert-test func) all-pass)))
    all-pass))

(defun assert-test-set-or-die (function-list)
  "Running tests, and die if any fail"
  (unless (assert-test-set function-list)
    (abort-and-die)))

(defun assert-test-set-p (test set)
  "Asserting all elements in set pass test"
  (let ((all-pass t))
    (dolist (element set all-pass)
      (let ((result (funcall test element)))
	(setq all-pass (and result all-pass))
	(if result
	    (put-symbol-value-pass element)
	  (put-symbol-value-fail element))))))

(defun assert-utilities-p ()
  "Asserting prerequisite utilities"
  (assert-test-set-p
   #'(lambda (x) (ext:probe-pathname (symbol-value x)))
   '(*whereis-bunzip2* *whereis-chown* *whereis-chmod* *whereis-cp* 
		       *whereis-cron* *whereis-curl* *whereis-echo*
		       *whereis-gm* *whereis-hdparm*
		       *whereis-md5sum*
		       *whereis-mv* *whereis-mysql* *whereis-mysqldump*
		       *whereis-openssl*
		       *whereis-php* *whereis-rm* *whereis-rsvg*
		       *whereis-wget*))
  (when *mirror-image-use-wikix-p*
    (assert-test-set-p
     #'(lambda (x) (ext:probe-pathname (symbol-value x)))
     '(*whereis-wikix*)))
  (assert-test-set-p
   #'(lambda (x) (ext:probe-pathname
		  (merge-pathnames
		   (parse-namestring *whereis-directory-mediawiki*)
		   (parse-namestring (symbol-value x)))))
   '(*whereis-file-mediawiki-adminsettings*
     *whereis-file-mediawiki-localsettings*
     *whereis-file-mediawiki-localsettings-wpmirror*))
  (assert-test-set-p
   #'(lambda (x) (ext:probe-pathname
		  (merge-pathnames
		   (parse-namestring *whereis-directory-mediawiki-includes*)
		   (parse-namestring (symbol-value x)))))
   '(*whereis-file-mediawiki-import*))
  (assert-test-set-p
   #'(lambda (x) (ext:probe-pathname
		  (merge-pathnames
		   (parse-namestring *whereis-directory-mediawiki-maintenance*)
		   (parse-namestring (symbol-value x)))))
   '(*whereis-file-mediawiki-farm-importdump*
     *whereis-file-mediawiki-farm-rebuildimages*
     *whereis-file-mediawiki-farm-update*
     *whereis-file-mediawiki-importdump*
     *whereis-file-mediawiki-rebuildimages*
     *whereis-file-mediawiki-update*)))

(defun assert-virtual-host-p (&key (put nil))
  "Asserting virtual host `mediawiki.site'"
  (and
   (assert-file-exists-p
    (merge-pathnames
     (parse-namestring *whereis-directory-apache-sites-available*)
     (parse-namestring *whereis-file-virtual-host*)))
   (assert-file-exists-p
    (merge-pathnames
     (parse-namestring *whereis-directory-apache-sites-enabled*)
     (parse-namestring *whereis-file-virtual-host*)))
   (assert-test
    'shell-apache2ctl-grep-vhost-p)))

(defun assert-virtual-host-name-resolution-p (&key (put nil))
  "Asserting name resolution for `mediawiki.site'"
  (let ((lines   (shell-cat-grep *whereis-file-etc-hosts* *virtual-host-name*))
	(results nil))
    (put-message-start (format nil "grepping ~a" *whereis-file-etc-hosts*))
    (loop
      for line in lines
      do (put-message line))
    (put-message-done (format nil "grepping ~a" *whereis-file-etc-hosts*))
    (put-message-start (_ "matching languages"))
    (setq results
	  (loop
	    with languages = (append *mirror-languages* (list "www"))
	    for lang    in languages
	    as  message =  (format nil "~a.~a" lang *virtual-host-name*)
	    as  result  =  (loop
			     for line in lines
			     thereis (regexp:match lang line))
	    do (if result
		   (put-message-value-pass lang message)
		 (put-message-value-fail lang message))
	    collect result))
    (put-message-done (_ "matching languages"))
    (not (member nil results))))

(defun assert-working-directory-or-create ()
  "Asserting working directory, or create it"
  (assert-directory-exists-or-create *whereis-directory-working*)
  (ext:cd *whereis-directory-working*))

(defun release-all-file-semaphores ()
  "Releasing all file semaphores"
  (put-message-start (_ "releasing all file semaphores"))
  (sql-update-file-semaphore-all 1 :put nil)
  (put-message-done (_ "releasing all file semaphores")))

;;;;--------------------------------------------------------------------------+
;;;; Warn user if proxy detected:                                             |
;;;;--------------------------------------------------------------------------+


(defun warn-if-detect-proxy ()
  "Warning if see mention of web proxy"
  (let* ((pattern        "[pP][rR][oO][xX][yY]")
	 (message        (_ "proxy config in"))
	 (result-curlrc (grep-first-non-comment pattern *whereis-curlrc*))
	 (result-wgetrc (grep-first-non-comment pattern *whereis-wgetrc*))
	 (result-bashrc (grep-first-non-comment pattern *whereis-bashrc*)))
    (unless (null result-curlrc)
      (put-message-value (format nil "~a ~a" message *whereis-curlrc*) 
			 result-curlrc))
    (unless (null result-wgetrc)
      (put-message-value (format nil "~a ~a" message *whereis-wgetrc*)
			 result-wgetrc))
    (unless (null result-bashrc)
      (put-message-value (format nil "~a ~a" message *whereis-bashrc*) 
			 result-bashrc))
    (if (or result-curlrc result-wgetrc result-bashrc)
	(put-message-done (_ "might have found proxy"))
      (put-message-done (_ "no proxy found")))
    (not (or result-curlrc result-wgetrc result-bashrc))))

;;;;--------------------------------------------------------------------------+
;;;; Determine mode of operation (mirror v. monitor):                         |
;;;;   Mode can be forced by use of command-line options                      |
;;;;     `--gui'     forces monitor mode, first attempting to use GUI         |
;;;;     `--mirror'  forces mirror mode (overriding any monitor options)      |
;;;;     `--monitor' forces monitor mode, first attempting to use GUI         |
;;;;     `--screen'  forces monitor mode, first attempting to use SCREEN      |
;;;;     `--text'    forces monitor mode, using normal text output            |
;;;;  If no options are given, then the PID file routines (below) come into   |
;;;;  play.  1) If no PID file is found, then `wp-mirror' goes into mirror    |
;;;;  mode. 2) If PID file is found (and not stale), then `wp-mirror' goes    |
;;;;  into monitor mode, first attempting to use GUI.                         |
;;;;  This use of PID files is somewhat different from the way daemons (when  |
;;;;  invoked by `/etc/init.d/*') use them.                                   |
;;;;--------------------------------------------------------------------------+


(defun pidfile-exists-p (&key (put nil))
  "Determining if PID file exists.  If so return PID, else NIL"
  (unless (null put)
    (put-message-start (_ "probing for PID file")))
  (with-open-file (s *whereis-pidfile* :direction :input
		     :if-does-not-exist nil)
    (if s
	(let ((pid (read s)))
	  (unless (null put)
	    (put-message-value-done (_ "reading value") pid))
	  pid)                  ; return pid
      (progn
	(unless (null put)
	  (put-message-done (_ "PID file not found")))
	nil))))                	; return nil

(defun proc-exists-p (pid &key (put nil))
  "Determining existence of process with given PID"
  (unless (null put)
    (put-message-value (_ "probing for process with PID") pid))
  (let ((proc-p  (ext:probe-pathname (format nil "/proc/~d" pid))))
    (unless (null put)
      (if proc-p
	  (put-message-value-done (_ "probing for process with PID") pid)
	(put-message-value-fail (_ "probing for process with PID") pid)))
    proc-p))

(defun first-mirror-exists-p (&key (put nil))
  "Determining if first instance of `wp-mirror' in mirror mode is running."
  (let* ((pid            (pidfile-exists-p :put put))
	 (first-mirror-p (and pid (proc-exists-p pid))))
    (unless first-mirror-p
      (if pid
	  (put-message (_ "PID file found stale - first mirror was probably killed"))
	(put-message (_ "PID file not found"))))
    first-mirror-p))

(defun i-am-the-first-mirror-p ()
  "Determining if this process set the PID file"
  (eql (os:process-id) (pidfile-exists-p)))

(defun set-pidfile ()
  "Creating PID file"
  (let ((pid (os:process-id)))
    (put-message-value (_ "my process id is") pid)
    (with-open-file (s *whereis-pidfile* :direction :output
		       :if-exists :supersede)
    (if s
	(progn
	  (format s "~d~%" pid)
	  (put-message-value-done (_ "creating PID file") *whereis-pidfile*))
      (put-message-fail-and-die (_ "unable to create PID file"))))))

(defun clear-pidfile ()
  "Removing PID file"
  (put-message-start (_ "probing for PID file"))
  (if (ext:probe-pathname *whereis-pidfile*)
      (progn
	(delete-file *whereis-pidfile*)
	(put-message-value-done (_ "removing PID file") *whereis-pidfile*))
    (put-message-done (_ "PID file not found")))
  t)

(defun main-mode-mirror-or-monitor ()
  "Determining if this instance of wp-mirror should run in mirror mode
or in monitor mode"
  ;;
  ;; Design note:
  ;;
  ;; Command-line options should override default PID logic.
  ;;
  ;;                  mirror cmd line
  ;;     Cases        exists options   action                 return
  ;; ----------------+------+---------+----------------------+--------
  ;;  1) first mirror   no     none    clear-pid              :mirror
  ;;                                   set-pidfile
  ;;                                   release-all-semaphores
  ;;  2) first mirror   no   --mirror  clear-pid              :mirror
  ;;                                   set-pidfile 
  ;;                                   release-all-semaphores
  ;;  3) monitor        no   --monitor   none                 :monitor
  ;;  4) monitor       yes     none      none                 :monitor
  ;;  5) next mirror   yes   --mirror    none                 :mirror
  ;;  6) monitor       yes   --monitor   none                 :monitor
  ;;
  ;; The options `--gui', `--screen', and `--text', imply `--monitor'.
  ;;
  (put-message-start (_ "determining mode of operation"))
  (let* ((mirror-exists-p (first-mirror-exists-p :put t))
	 (choice          nil))
    (put-message-value (_ "does mirror process exist") mirror-exists-p)
    (put-message-value (_ "does command line force mirror") *cmd-force-mirror*)
    (put-message-value (_ "does command line force monitor") 
		       *cmd-force-monitor*)
    ;; logic goes here (the six cases mentioned above)
    (setq choice
	  (cond ((and (not mirror-exists-p)
		      (not *cmd-force-mirror*)
		      (not *cmd-force-monitor*)) :first-mirror)
		((and (not mirror-exists-p)
		      *cmd-force-mirror*)        :first-mirror)
		((and (not mirror-exists-p)
		      (not *cmd-force-mirror*)
		      *cmd-force-monitor*)       :monitor)
		((and mirror-exists-p
		      (not *cmd-force-mirror*)
		      (not *cmd-force-monitor*)) :monitor)
		((and mirror-exists-p
		      *cmd-force-mirror*)        :next-mirror)
		((and mirror-exists-p
		      (not *cmd-force-mirror*)
		      *cmd-force-monitor*)       :monitor)
		(t nil)))
    ;; action goes here
    (case choice
      (:first-mirror
       (put-message-start (_ "clearing stale PID file"))
       (clear-pidfile)
       (put-message-done (_ "clearing stale PID file"))
       (set-pidfile)
       (put-message-value (_ "setting mode of operation to")
			  (_ "FIRST-MIRROR")))
      (:next-mirror
       (put-message-value (_ "setting mode of operation to")
			  (_ "NEXT-MIRROR")))
      (:monitor
       (put-message-value (_ "setting mode of operation to")
			  (_ "MONITOR")))
      (otherwise nil))
    ;; return value goes here
    choice))


;;;;--------------------------------------------------------------------------+
;;;; Status utilities:                                                        |
;;;;   State is stored for each file to enable resuming after outage          |
;;;;                                                                          |
;;;;   Globals:                                                               |
;;;;     *input*               - list of valid inputs to finite-state-machine |
;;;;     *state*               - list of valid states                         |
;;;;     *state-transition*    - list of valid state transitions              |
;;;;     *type*                - list of valid file types                     |
;;;;     *type-state-function* - list of functions to be run when a file of a |
;;;;                             given type and state receives the "start"    |
;;;;                             input                                        |
;;;;     *type-state-priority* - prioritized list of type and state (files are|
;;;;                             processed in order of priority)              |
;;;;   Functions:                                                             |
;;;;     fsm-abort             - abort wp-mirror (due to error)               |
;;;;     fsm-boot              - initialize wp-mirror (load `checksum's)      |
;;;;     fsm-file-bash         - run `ichunk' bash script to download images  |
;;;;     fsm-file-count        - count pages and images in `xml' file         |
;;;;     fsm-file-decompress   - decompress `bzip2' file                      |
;;;;     fsm-file-download     - download checksum or dump file from wikimedia|
;;;;     fsm-file-import       - import `xchunk' into `xxwiki'                |
;;;;     fsm-file-md5sum       - compute `md5sum' checksum                    |
;;;;     fsm-file-split        - split decompressed dump into `xchunk's       |
;;;;     fsm-file-validate     - compare `md5sum' with known value            |
;;;;     fsm-file-wikix        - generate `ichunk' with `wikix'               |
;;;;     fsm-images-chown      - chown images to `www-data:www-data'          |
;;;;     fsm-images-count      - count downloaded images                      |
;;;;     fsm-images-rebuild    - import image metadata into `xxwiki'          |
;;;;     fsm-images-validate   - find corrupt files w/ `gm identify -verbose' |
;;;;     fsm-no-op             - do nothing                                   |
;;;;     fsm-transition        - initial state + valid input --> final state  |
;;;;   Design note:                                                           |
;;;;     finite-state-machine  - graphical representation                     |
;;;;                                                                          |
;;;;    __           __         __           __        __              __     |
;;;;   |  |s        |  |s      |  |s        |  |s     |  |s           |  |sdf |
;;;;   v  |         v  |       v  |         v  |      v  |            v  |    |
;;;; +-----+ d  +-------+ d  +-----+ d  +-------+ d  +----+ d        +-----+  |
;;;; |start|--->|created|--->|valid|--->|pending|--->|done|--------->|error|  |
;;;; +-----+    +-------+    +-----+    +-------+    +----+    __    +-----+  |
;;;;     \            \          \            \         \     |  |s    /      |
;;;;      \f           \f         \f           \f        \f   v  |    /df     |
;;;;       \            \          \            \         \  +----+  /        |
;;;;        ------------------------------------------------>|fail|--         |
;;;;                                                         +----+           |
;;;;                                                                          |
;;;;--------------------------------------------------------------------------+


(defconstant  *input*
  '("start" "done" "fail"))
(defconstant  *state*
  '("start" "created" "valid" "pending" "done" "fail" "error"))
(defconstant  *state-transition*
  ;;  state  +  input ->state
  '(("start"   "start" "start"  ) ;resume
    ("start"   "done"  "created")
    ("start"   "fail"  "fail"   )
    ("created" "start" "created") ;resume
    ("created" "done"  "valid"  )
    ("created" "fail"  "fail"   )
    ("valid"   "start" "valid"  ) ;resume
    ("valid"   "done"  "pending")
    ("valid"   "fail"  "fail"   )
    ("pending" "start" "pending") ;resume
    ("pending" "done"  "done"   )
    ("pending" "fail"  "fail"   )
    ("done"    "start" "done"   ) ;resume
    ("done"    "done"  "done"   ) ;no-op
    ("done"    "fail"  "error"  ) ;should not happen
    ("fail"    "start" "fail"   ) ;resume
    ("fail"    "done"  "fail"   ) ;no-op
    ("fail"    "fail"  "error"  ) ;should not happen
    ("error"   "start" "error"  ) ;should not happen
    ("error"   "done"  "error"  ) ;should not happen
    ("error"   "fail"  "error"  ) ;should not happen
    ))
(defconstant  *type*
  '("checksum" "dump" "xml" "xchunk" "ichunk" "images" "error"))
(defconstant  *fsm-function*
  '(fsm-abort
    fsm-boot
    fsm-file-bash
    fsm-file-count
    fsm-file-decompress 
    fsm-file-download 
    fsm-file-import
    fsm-file-md5sum
    fsm-file-parse
    fsm-file-split
    fsm-file-validate
    fsm-file-wikix
    fsm-images-chown
    fsm-images-count
    fsm-images-rebuild
    fsm-images-validate
    fsm-no-op
))
(defconstant *fsm-function-needs-internet-access*
  '(fsm-file-bash
    fsm-file-download
))
(defconstant  *type-state-function*
  ;;  type     state     function
  '(("checksum" "start"   fsm-file-download  ) ; OK
    ("checksum" "created" fsm-file-md5sum    ) ; OK
    ("checksum" "valid"   fsm-file-parse     ) ; OK
    ("checksum" "pending" fsm-no-op          )
    ("checksum" "done"    fsm-no-op          )
    ("checksum" "fail"    fsm-abort          )
    ("checksum" "error"   fsm-abort          )
    ("dump"     "start"   fsm-file-download  )
    ("dump"     "created" fsm-file-validate  ) ; OK
    ("dump"     "valid"   fsm-file-decompress) ; OK
    ("dump"     "pending" fsm-no-op          )
    ("dump"     "done"    fsm-no-op          )
    ("dump"     "fail"    fsm-abort          )
    ("dump"     "error"   fsm-abort          )
    ("xml"      "start"   fsm-file-count     ) ; OK
    ("xml"      "created" fsm-file-md5sum    )
    ("xml"      "valid"   fsm-file-split     ) ; OK
    ("xml"      "pending" fsm-no-op          )
    ("xml"      "done"    fsm-no-op          )
    ("xml"      "fail"    fsm-abort          )
    ("xml"      "error"   fsm-abort          )
    ("xchunk"   "start"   fsm-file-count     )
    ("xchunk"   "created" fsm-file-md5sum    )
    ("xchunk"   "valid"   fsm-file-wikix     ) ; OK
    ("xchunk"   "pending" fsm-file-import    ) ; OK
    ("xchunk"   "done"    fsm-no-op          )
    ("xchunk"   "fail"    fsm-no-op          )
    ("xchunk"   "error"   fsm-abort          )
    ("ichunk"   "start"   fsm-file-count     )
    ("ichunk"   "created" fsm-file-md5sum    )
    ("ichunk"   "valid"   fsm-no-op          )
    ("ichunk"   "pending" fsm-file-bash      ) ; OK
    ("ichunk"   "done"    fsm-no-op          )
    ("ichunk"   "fail"    fsm-no-op          )
    ("ichunk"   "error"   fsm-abort          )
    ("images"   "start"   fsm-images-count   ) ; OK
    ("images"   "created" fsm-images-validate) ; OK
    ("images"   "valid"   fsm-images-rebuild ) ; OK
    ("images"   "pending" fsm-images-chown   ) ; OK
    ("images"   "done"    fsm-no-op          )
    ("images"   "fail"    fsm-no-op          )
    ("images"   "error"   fsm-abort          )
    ("error"    "start"   fsm-abort          )
    ("error"    "created" fsm-abort          )
    ("error"    "valid"   fsm-abort          )
    ("error"    "pending" fsm-abort          )
    ("error"    "done"    fsm-abort          )
    ("error"    "fail"    fsm-abort          )
    ("error"    "error"   fsm-abort          )))
(defconstant  *type-state-priority*
  ;;  type       state    concurrent non-image
  '(
    ("checksum" "pending" t          t        ) ; fsm-no-op
    ("checksum" "valid"   t          t        ) ; fsm-file-parse
    ("checksum" "created" t          t        ) ; fsm-file-md5sum
    ("checksum" "start"   nil        t        ) ; fsm-file-download
    ("dump"     "pending" t          t        ) ; fsm-no-op
    ("dump"     "valid"   t          t        ) ; fsm-file-decompress
    ("dump"     "created" t          t        ) ; fsm-file-validate
    ("dump"     "start"   nil        t        ) ; fsm-file-download
    ("xml"      "pending" t          t        ) ; fsm-no-op
    ("xml"      "valid"   t          t        ) ; fsm-file-split
    ("xml"      "created" t          t        ) ; fsm-file-md5sum
    ("xml"      "start"   t          t        ) ; fsm-file-count
    ("xchunk"   "valid"   t          nil      ) ; fsm-file-wikix
    ("xchunk"   "created" t          t        ) ; fsm-file-md5sum
    ("xchunk"   "start"   t          t        ) ; fsm-file-count
    ("ichunk"   "pending" nil        nil      ) ; fsm-file-bash
    ("ichunk"   "valid"   t          nil      ) ; fsm-no-op
    ("ichunk"   "created" t          nil      ) ; fsm-file-md5sum
    ("ichunk"   "start"   t          nil      ) ; fsm-file-count
    ("images"   "start"   t          nil      ) ; fsm-images-count
    ("images"   "created" t          nil      ) ; fsm-images-validate
    ("images"   "valid"   nil        nil      ) ; fsm-images-rebuild
    ("xchunk"   "pending" t          t        ) ; fsm-file-import
    ("images"   "pending" t          nil      ) ; fsm-images-chown
    )) 


(defun fsm-abort (file-name &key (put nil))
  "Aborting due to error condition"
  ;; Files and states:
  ;;           |input file          generated file(s)
  ;; ----------+-------------------+--------------------------
  ;; file-name |don't care          none
  ;; state-in  |see *type-state-function*
  ;; state-out |no change
  ;; Return: nil means fail, t means done.
  (let ((file-type  (sql-select-file-type  file-name))
	(file-state (sql-select-file-state file-name)))
    (put-message-start (_ "ERROR: in finite state machine"))
    (put-message (_ "this should never happen"))
    (put-message-value (_ "file-name")  file-name)
    (put-message-value (_ "file-type")  file-type)
    (put-message-value (_ "file-state") file-state)
    (put-message-value-fail-and-abort (_ "ERROR:") file-name))
  nil) ; always mark as `fail'

(defun fsm-boot (&key (put nil))
  "Booting the finite state machine"
  ;; Prime the pump
  ;; Insert checksum file names into empty `wpmirror.file'
  (put-message-start (_ "booting finite state machine"))
  (let ((results (loop
		   for lang      in *mirror-languages*
		   as  file-name = (checksums-file-name lang)
		   as  result    = (sql-insert-file lang "latest" 
						    file-name "checksum"
						    :put put)
		   collect result)))
    (put-message-done (_ "booting finite state machine"))
    (not (member nil results))))

(defun fsm-file-bash (ichunk-name &key (put nil))
  "Downloading images for given `ichunk'"
  ;; Files and states:
  ;;           |input file          generated file(s)
  ;; ----------+-------------------+--------------------------
  ;; file-name |ichunk              image files
  ;; state-in  |valid               n/a
  ;; state-out |pending
  ;; Return: nil means fail, t means done.
  ;;
  ;; Design note:
  ;;
  ;; `Wikix' has an important bug.  If you run `wikix' with the `-i'
  ;; option, e.g.
  ;;
  ;;   /home/myname# wikix -i /var/lib/mediawiki/images/ \
  ;;                       -o /var/lib/mediawiki/images/ \
  ;;                       < xchunk.xml > ichunk.sh
  ;;
  ;; you will get a script that does not run as expected.  For example,
  ;;
  ;;   /home/myname# ./ichunk.sh
  ;;
  ;; will correctly check under `/var/lib/mediawiki/images/' for
  ;; existing images.  If a particular image is not found, `curl' will
  ;; be invoked and the image downloaded to the directory where
  ;; `ichunk.sh' is running (`/home/myname/' in this example).  The
  ;; script will then incorrectly check again under
  ;; `/var/lib/mediawiki/images/' to see if the image was successfully
  ;; downloaded, and therefore not find it.  The script will then fail
  ;; to store the image there, and will then mis-report the image in
  ;; `failed.log' rather than `download.log'.
  ;; 
  ;; If you want the scripts to run correctly, then 
  ;;   a) run wikix without the "-i" option, for example
  ;;
  ;;   /home/myname# wikix < xchunk.xml > ichunk.sh
  ;;    
  ;;   b) move (or copy) `ichunk.sh' to `/var/lib/mediawiki/images/', and then
  ;;   c) run the script in that directory, like so
  ;;
  ;;   /var/lib/mediawiki/images# ./ichunk.sh
  ;;
  ;; It is for this reason, that `fsm-file-bash' (this function)
  ;;   o changes directory to the working directory (if not already there),
  ;;   o copies the ichunk to the images directory, 
  ;;   o changes directory to the images directory,
  ;;   o probes that the copy is there,
  ;;   o runs the copied script (which downloads and stores the images),
  ;;   o removes the copy, and finally
  ;;   o changes directory back to the working directory,
  ;;
  (let ((r nil))
    (debug-message ichunk-name)
    (ext:cd *whereis-directory-working*)
    (debug-message-value (_ "copying") ichunk-name)
    (setq r (null (shell-copy-file ichunk-name *whereis-directory-images*)))
    (when r
      (debug-message-value (_ "changing to images directory")
			   *whereis-directory-images*)
      (ext:cd *whereis-directory-images*)
      (debug-message-value (_ "current working directory is now") (ext:cd))
      (debug-message-value (_ "probing") (ext:probe-pathname ichunk-name))
      (setq r (ext:probe-pathname ichunk-name))
      (when r
	(debug-message-value (_ "running") (ext:probe-pathname ichunk-name))
	(setq r (null (shell-bash ichunk-name)))
	(when r
	  (debug-message-value (_ "removing") (ext:probe-pathname ichunk-name))
	  (shell-remove-file ichunk-name))))
    (debug-message-value (_ "changing to working directory")
			 *whereis-directory-working*)
    (ext:cd *whereis-directory-working*)
    (debug-message-value (_ "current working directory is now") (ext:cd))
    r))

(defun fsm-file-count (file-name &key (put nil))
  "Counting pages `<page>' and images `<title>File:' in `.xml' files"
  ;; Files and states:
  ;;           |input file          generated file(s)
  ;; ----------+-------------------+--------------------------
  ;; file-name |any except dump     none
  ;; state-in  |start
  ;; state-out |created
  ;; Return: nil means fail, t means done.
  (sql-update-file-size file-name :put put)
  (let ((file-type   (sql-select-file-type file-name :put put))
	(page-count  0)
	(image-count 0))
    (when (or (string= file-type "xml") (string= file-type "xchunk"))
      (setq page-count (shell-count-page  file-name))
      (sql-update-file file-name :pages page-count :put put))
    (setq image-count (shell-count-image file-name))
    (sql-update-file file-name :images image-count :put put)))

(defun fsm-file-decompress (file-name &key (put nil))
  "Decompressing dump file to yield XML file"
  ;; Files and states:
  ;;           |input file          generated file(s)
  ;; ----------+-------------------+--------------------------
  ;; file-name |dump                dump-xml
  ;; state-in  |valid               n/a
  ;; state-out |pending             start
  ;; Return: nil means fail, t means done.
  ;; 1) decompress dump-file
  (let* ((language  (sql-select-file-language file-name :put put))
	 (file-date (sql-select-file-date     file-name :put put))
	 (xml-name  (string-right-trim ".bz2" file-name)))
    (put-message-value (_ "file dump xml name") xml-name)
    (put-message-start (_ "decompressing"))
    (shell-bunzip2 file-name)
    (put-message-done (_ "decompressing"))
    ;; 2) create record for decompressed file
    (when (ext:probe-pathname xml-name)
      (sql-insert-file language file-date xml-name "xml" :put put)
      (sql-update-file-size xml-name :put put))))

(defun fsm-file-download (file-name &key (put nil))
  "Downloading checksum or dump file"
  ;; Files and states:
  ;;           |input file          generated file(s)
  ;; ----------+-------------------+--------------------------
  ;; file-name |checksum or dump    checksum or dump
  ;; state-in  |start
  ;; state-out |created
  ;; Return: nil means fail, t means done.
  ;;
  ;; Names:
  ;; site     is   `http://dumps.wikimedia.org/'
  ;; checksum like `xxwiki/yyyymmdd/xxwiki-latest-md5sums.txt'
  ;; dump     like `xxwiki/yyyymmdd/xxwiki-yyyymmdd-pages-articles.xml.bz2'
  ;;
  ;; Design note:
  ;;
  ;; Languages such as `zh-classical' (i.e. `xx' contains a dash `-')
  ;; must be carefully handled.  MySQL does not readily accept
  ;; database names containing a dash:
  ;;
  ;; fail:  CREATE DATABASE zh-classicalwiki;
  ;; OK  :  CREATE DATABASE `zh-classicalwiki`;
  ;; OK  :  CREATE DATABASE zh_classicalwiki;
  ;; 
  ;; WikiMedia Foundation usage is:
  ;; 
  ;; underscore : directory     : http://dumps.wikimedia.org/zh_classical/...
  ;;            : checksum file : zh_classicalwiki-latest-md5sums.txt
  ;;            : dump file     : zh_classical-yyyymmdd-pages-articles.xml.bz2
  ;; dash       : URL           : http://zh-classical.wikipedia.org/
  ;;
  ;; Hence: http://dumps.wikimedia.org/zh_classicalwiki/latest/zh_classicalwiki
  ;;
  ;; Wp-mirror attempts to follow WikiMedia Foundation usage:
  ;; languages use dash, while files and databases use underscore.
  ;; See functions `language-to-file-name' and `str_replace'.
  ;;
  (debug-message-value (_ "file-name") file-name)
  (let* ((file-type (sql-select-file-type     file-name :put put))
	 (lang      (sql-select-file-language file-name :put put))
	 (url       nil)
	 (file-path nil))
    (cond ((string= file-type "checksum")
	   (setq url       (checksums-url lang))
	   (setq file-path (checksums-path lang)))
	  ((string= file-type "dump")
	   (setq url       (dump-url lang))
	   (setq file-path (dump-path lang)))
	  (t nil))
    (put-message url)
    (when (download-file url file-path))
      (put-message file-path)
      (when (file-exists-p file-path)
	(sql-update-file-size file-name :put put))))

(defun fsm-file-import (xchunk-name &key (put nil))
  "Importing `xchunk' into `xxwiki'"
  ;; Files and states:
  ;;           |input file          generated file(s)
  ;; ----------+-------------------+--------------------------
  ;; file-name |xchunk              none
  ;; state-in  |pending             n/a
  ;; state-out |done                
  ;; Return: nil means fail, t means done.
  (let* ((language     (sql-select-file-language xchunk-name :put put))
	 (date         (sql-select-file-date     xchunk-name :put put))
	 (page         (sql-select-file-page     xchunk-name :put put))
	 (pages        (sql-select-file-pages    xchunk-name :put put))
	 (ichunk-name  (concatenate 'string
				    (string-right-trim ".xml" xchunk-name)
				    ".sh"))
	 (ichunk-state (sql-select-file-state ichunk-name :put put))
	 (process      nil))
    (put-message xchunk-name)
    (not (shell-mediawiki-farm-importdump xchunk-name))))

(defun fsm-file-md5sum (file-name &key (put nil))
  "Computing md5sum checksum"
  ;; Files and states:
  ;;           |input file          generated file(s)
  ;; ----------+-------------------+--------------------------
  ;; file-name |any except dump     n/a
  ;; state-in  |created             
  ;; state-out |valid               
  ;; Return: nil means fail, t means done.
  (debug-message-value (_ "file-name") file-name)
  (let ((result (shell-md5sum file-name)))
    (if (null result)
	(put-message-value-fail-and-abort (_ "md5sum returned") result)
      (progn
	(debug-message-value (_ "md5sum") result)
	(sql-update-file-md5sum file-name result :put put)))))

(defun fsm-file-parse (file-name &key (put nil))
  "Parsing checksums file to identify dump file"
  ;; Files and states:
  ;;           |input file          generated file(s)
  ;; ----------+-------------------+--------------------------
  ;; file-name |checksum            dump file
  ;; state-in  |valid               n/a
  ;; state-out |pending
  ;; Return: nil means fail, t means done.
  ;;
  (put-message-value (_ "checksums file") file-name)
  (put-message-start (_ "looking for pattern match"))
  (let* ((language       (sql-select-file-language file-name :put put))
	 (file-template  (first (last (regexp:regexp-split 
				       "/" *wikimedia-dump-template*))))
	 (pattern        (join (cddr (regexp:regexp-split "-" file-template))
			       "-"))
	 (line           (grep-first pattern file-name))
	 (dump-checksum  (first  (regexp:regexp-split " " line)))
	 (dump-name      (first  (last (regexp:regexp-split " " line))))
	 (dump-date      (second (regexp:regexp-split "-" dump-name))))
    (put-message-value (_ "dump file name")     dump-name)
    (put-message-value (_ "dump file checksum") dump-checksum)
    (put-message-done  (_ "looking for pattern match"))
    ;; store status in `wpmirror'
    (put-message-start (_ "probing dump file"))
    (let ((r (sql-select-file dump-name :put put)))
      (put-message-value (_ "sql-select-file dump-name returned") r)
      (when (null r)
	(put-message-start (_ "Seeing newer dump so cleaning any old ones"))
	(dump-file-remove-p language)
	(put-message-done (_ "Seeing newer dump so cleaning any old ones"))
	(put-message-start (_ "sql-insert-file dump"))
	(sql-insert-file language dump-date dump-name "dump"
			 :md5sum dump-checksum :put put)
	(put-message-done (_ "sql-insert-file dump"))))
    (put-message-done (_ "probing dump file"))
    t))

(defun fsm-file-split (file-name &key (put nil))
  "Splitting decompressed dump file into `xchunk's"
  ;; Files and states:
  ;;           |input file          generated file(s)
  ;; ----------+-------------------+--------------------------
  ;; file-name |dump-xml            xchunks
  ;; state-in  |valid               n/a
  ;; state-out |pending             start
  ;; Return: nil means fail, t means done.
  (let ((language  (sql-select-file-language file-name :put put))
	(file-date (sql-select-file-date     file-name :put put)))
    (with-open-file (f file-name
		       :direction :input
		       :if-does-not-exist nil)
      ;; generating the set of xchunks
      (loop named xchunk-set
	with mediawiki-tag-start = (read-line f nil nil)
	and mediawiki-tag-end   = "</mediawiki>"
	and file-name-head      = (concatenate 'string
					       (string-right-trim 
						".xml"
						file-name))
	and file-name-tail      = (concatenate 'string
					       (format nil "-c~9,'0d.xml"
						       *xchunk-page-count*))
	and page                = 0
	;; xchunk file-names will look like
	;; `enwiki-yyyymmdd-pages-articles-p000000000-c000001000.xml'
	for file-name-page of-type string
	                 = (format nil "-p~9,'0d" page) ; see <termination>
	as  chunk-name     of-type string
                         = (concatenate 'string         ; below for end-test
					file-name-head
					file-name-page
					file-name-tail)
	;; creating one xchunk
	count chunk-name into chunk-count
	do (debug-message-value (_ "xchunks-generated") chunk-count)
	   ;; 0) create record in `file' table of `wpmirror'
	   (sql-insert-file language file-date chunk-name "xchunk" :page page)
	   ;; 1) create file on disk
	   (with-open-file (x chunk-name
			      :direction :output
			      :if-exists :supersede)
	     (loop named xchunk
	       initially (debug-message-start (_ "creating xchunk"))
	                 (put-message chunk-name)
			 (write-line mediawiki-tag-start x)
	       with xchunk-page-count = 0
	       for l of-type string = (read-line f nil 'eof) 
	                            then (read-line f nil 'eof)
	       when (eq l 'eof) do (return-from xchunk-set t) ; <termination>
	       do   (write-line l x)
	       when (string= l "  </page>") 
	       do   (setq xchunk-page-count (1+ xchunk-page-count))
                    (setq page (1+ page))
	       until (eql xchunk-page-count *xchunk-page-count*)
	       finally   (write-line mediawiki-tag-end   x)
	                 (debug-message-done (_ "creating xchunk")))
	     t)
	   ;; 2) update record in `file' table of `wpmirror'
	   (when (ext:probe-pathname chunk-name)
	     (sql-update-file-size file-name))))))

(defun fsm-file-validate (file-name &key (put nil))
  "Validating downloaded dump file against known checksum"
  ;; Files and states:
  ;;           |input file          generated file(s)
  ;; ----------+-------------------+--------------------------
  ;; file-name |dump                none
  ;; state-in  |created             
  ;; state-out |valid               
  ;; Return: nil means fail, t means done.
  (let* ((dump-checksum-should (sql-select-file-md5sum file-name :put put))
	 (dump-checksum-actual nil))
    ;; 1) update `size' just in case the dump file was manually downloaded
    ;;    into the working directory
    (sql-update-file-size file-name :put put)
    ;; 2) compare known checksum with actual checksum
    (put-message-value (_ "checksum should be") dump-checksum-should)
    (with-open-stream (s (shell-md5sum-and-return-stream file-name))
      (setq dump-checksum-actual
	    (first (regexp:regexp-split " " (read-line s))))
      (put-message-value (_ "checksum actually is") dump-checksum-actual)
      (equal dump-checksum-should dump-checksum-actual))))

(defun fsm-file-wikix (xchunk-name &key (put nil))
  "Generating `ichunk' for given `xchunk' using `wikix'"
  ;; Files and states:
  ;;           |input file          generated file(s)
  ;; ----------+-------------------+--------------------------
  ;; file-name |xchunk              ichunk
  ;; state-in  |valid               n/a
  ;; state-out |pending             start
  ;; Return: nil means fail, t means done.
  (let* ((language     (sql-select-file-language xchunk-name :put put))
	 (date         (sql-select-file-date     xchunk-name :put put))
	 (page         (sql-select-file-page     xchunk-name :put put))
	 (pages        (sql-select-file-pages    xchunk-name :put put))
	 (ichunk-name  (concatenate 'string
				    (string-right-trim ".xml" xchunk-name)
				    ".sh"))
	 (process      nil))
    (debug-message xchunk-name)
    (debug-message ichunk-name)
    (if *mirror-image-use-wikix-p*
	(progn 
	  (shell-wikix     xchunk-name ichunk-name)
	  (ichunk-edit-p   ichunk-name))
      (let ((image-hash-table (make-hash-table)))
	;; to debug pathological file names, set :put t 
	(parse-image-file-names xchunk-name image-hash-table :put put)
	;;(die)
	(filter-exist image-hash-table :put put)
	(put-bash-script ichunk-name image-hash-table :put put)))
    (sql-insert-file language date ichunk-name "ichunk"
		     :page page :pages pages :put put)
    (shell-chmod     ichunk-name)
    t)) ; do not mark `xchunk' as `fail', even if `wikix' fails


(defun known-image-extension-p (string)
  "Matching with known file extensions"
  (or (regexp:match "\\.[jJ][pP][gG]$"     string) ; jpg
      (regexp:match "\\.[jJ][pP][eE][gG]$" string) ; jpeg
      (regexp:match "\\.[pP][nN][gG]$"     string) ; png
      (regexp:match "\\.[gG][iI][fF]$"     string) ; gif
      (regexp:match "\\.[sS][vV][gG]$"     string) ; svg
      (regexp:match "\\.[pP][dD][fF]$"     string) ; pdf
      (regexp:match "\\.[tT][iI][fF]$"     string) ; tif
      (regexp:match "\\.[tT][iI][fF][fF]$" string) ; tiff
      (regexp:match "\\.[oO][gG][gG]$"     string) ; ogg
      (regexp:match "\\.[dD][jJ][vV][uU]$" string) ; djvu
      (regexp:match "\\.[wW][aA][vV]$"     string) ; wav
      (regexp:match "\\.[mM][iI][dD]$"     string) ; mid
      (regexp:match "\\.[fF][iI][gG]$"     string) ; fig
      (regexp:match "\\.[bB][mM][pP]$"     string) ; bmp
      (regexp:match "\\.[xX][cC][fF]$"     string) ; xcf
      (regexp:match "\\.[dD][iI][aA]$"     string) ; dia
      (regexp:match "\\.[mM][pP]3$"        string) ; mp3
      (regexp:match "\\.[pP][sS][pP]$"     string) ; psp
      (regexp:match "\\.[bB][zZ]2$"        string) ; bz2
      (regexp:match "\\.[mM][oO][vV]$"     string) ; mov
      ))

(defun normalize-file-name (string)
  "Normalizing file name"
  ;; trim leading and trailing whitespace, use underscore not space
  (substitute #\_ #\Space (string-trim " " string)))

(defun parse-image-file-names (xchunk-name image-hash-table &key (put nil))
  "Parsing image file names from `x-chunk'"
  ;;
  ;; Design note:
  ;;
  ;; Problem: File names with special characters can open security
  ;; holes.  For example:
  ;;
  ;; shell# curl http://upload.wikimedia.org/wikipedia/en/0/00/foo`bar`baz.jpg
  ;; bash: bar: command not found
  ;;
  ;; Clearly some file names present a security risk.  A malicious
  ;; person could upload image files with names containing shell
  ;; commands (e.g. replace `bar` with any shell command).  As a not
  ;; too hazardous example, try
  ;;
  ;; shell# curl http://upload.wikimedia.org/wikipedia/en/0/00/foo`ls`baz.jpg
  ;;
  ;; which will produce one error message for each file in your
  ;; current working directory.
  ;;
  ;; Now, guess what the next example does.
  ;;
  ;; shell# curl http://upload.wikimedia.org/wikipedia/en/0/00/foo`halt`baz.jpg
  ;;
  ;; Solution: Ignore file names with certain special characters.
  ;; 
  (let ((xchunk-path      (merge-pathnames
			   (parse-namestring *whereis-directory-working*)
			   (parse-namestring xchunk-name))))
    (with-open-file (s xchunk-path :direction :input)
      (loop
	for line  = (read-line s nil nil)
	while line
	do
	;;(put-message line) ; uncomment to debug pathological file name
	(loop
	  with string = (make-array 0
				    :element-type 'character 
				    :fill-pointer 0
				    :adjustable t)
	  with state = 0
	  for char across line
	  do
	  (cond ((and (eql state 0)(equal char #\[)) ; tag start `['
		 (setq state 1))
		((and (eql state 1)(equal char #\[)) ; tag start `[['
		 (setq state 2))
		((and (eql state 2)(equal char #\:)) ; tag start `[[foo:'
		 (setq state 3))
		;; state 3 - potentially an image file name
		((and (eql state 3)
		      (or (equal char #\:)           ; failing on 2nd colon
			  (equal char #\/)           ; failing on URL
			  (equal char #\~)           ; failing on tilde
			  (equal char #\{)           ; failing on braces
			  (equal char #\})           ; 
			  (equal char #\%)           ; failing on percent
			  (equal char #\Newline)     ; failing on newline
			  (equal char #\Rubout)      ; failing on whitespace
			  (equal char #\Page)
			  (equal char #\Tab)
			  (equal char #\Backspace)
			  (equal char #\Tab)
			  (equal char #\Return)
			  (equal char #\Linefeed)
			  ;; delete this when we decide how to deal 
			  ;; with AMP codes (like &amp;)
			  (equal char #\&)           ; failing on ampersand
			  ;; delete this when we can mix backquote with unicode
			  ;; delete this when we can protect shell from attack
			  (equal char #\`)
			  ;; delete this when proxy does not `302 Redirect'
			  (equal char #\')
			  ))
		 (setq state 0)
		 (setq string (make-array 0
					  :element-type 'character
					  :fill-pointer 0
					  :adjustable t)))
		;; state 3 - probably found valid image file name
		((and (eql state 3)                  
		      (or (equal char #\])           ; ending on `]'
			  (equal char #\|)))         ; ending on `|'
		 (when (known-image-extension-p string)
		   ;; storing file-name and md5sum in hash table
		   ;; this has the effect of suppressing duplicates
		   (let* ((norm-string (normalize-file-name string))
			  (no-escape   (remove #\\ norm-string))
			  ;; md5sum computed from unescaped file name
			  (md5sum      (if (find #\` norm-string)
					   ;; backquote OK,  fail on unicode
					   (string-to-md5sum no-escape)
					   ;; unicode OK, fail on backquote
					 (shell-echo-md5sum no-escape)))
			  (d1          (subseq md5sum 0 1))
			  (d2          (subseq md5sum 0 2)))
		     ;(put-message-value md5sum norm-string)
		     (setf (gethash norm-string image-hash-table)
			   (format nil "~a/~a/" d1 d2))))
		 (setq state 0)
		 (setq string (make-array 0
				  :element-type 'character
				  :fill-pointer 0
				  :adjustable t)))
		;; state 3 - building image file name char-by-char
		((eql state 3)
		 ;; escape some characters
		 ;; this does not yet handle the AMP commands like
		 ;; &lt; &gt; &amp; &copy; &quot; &nbsp; and so on
		 (when (or (eql char #\-)            ; escape `-' for cURL
                           (eql char #\')            ; escape `'' for cURL
                           (eql char #\")            ; escape `"' for cURL
                           (eql char #\()            ; escape `(' for cURL
                           (eql char #\))            ; escape `)' for cURL
                           (eql char #\&)            ; escape `&' for cURL
                           (eql char #\;)            ; escape `;' for cURL
                           (eql char #\`)            ; escape ``' for cURL
                           (eql char #\@)            ; escape `@' for cURL
                           (eql char #\*)            ; escape `*' for cURL
			   )
		   (vector-push-extend #\\ string))
		 (vector-push-extend char string))))))))

(defun string-to-md5sum (string)
  "Computing md5sum of a string, return as a 32 char hex string"
  ;; Handles backquote '`'
  ;; Bug: fail if neither simple-string nor (unsigned-byte 8)
  (format nil "~(~{~2,'0x~}~)" 
	  (map 'list #'identity (md5:md5sum-sequence string))))

(defun filter-exist (image-hash-table &key (put nil))
  "Removing from hash-table names of existing image files"
  (loop
    with path = nil
    for key being the hash-keys of image-hash-table
    using (hash-value value)
    do
    (unless (null put) (put-message-value value key))
    (setq path (parse-namestring
		(format nil "~a~a~a"
			*whereis-directory-images* 
			value
			(remove #\\ key))))             ; unescaped file name
    (unless (null put) (put-message path))
    (unless (or (find #\? key)                          ; skip wildcards
		(find #\* key))
      (when (ext:probe-pathname path)
	(remhash key image-hash-table)))))

(defun put-bash-script (ichunk-name image-hash-table &key (put nil))
  "Putting BASH script for downloading images (like output of `wikix')"
  (let ((language    (sql-select-file-language ichunk-name :put put))
	(ichunk-path (merge-pathnames
		      (parse-namestring *whereis-directory-working*)
		      (parse-namestring ichunk-name))))
    (with-open-file (s ichunk-name
		       :direction :output
		       :if-exists :supersede)
    (loop
      initially
      (format s "#!/bin/sh~%~%")
      (format s "IMAGE=~%")
      (format s "OUTPUT=~%")
      (format s "IMAGEPATH=http://upload.wikimedia.org/wikipedia/~a/~%" 
	      language)
      (format s "COMMONSPATH=http://upload.wikimedia.org/wikipedia/commons/~%~%")
      (format s "/bin/mkdir -p $OUTPUT./thumb~%")
      (format s "/bin/chmod 777 $OUTPUT./thumb~%")
      (format s "/bin/mkdir -p $OUTPUT./temp~%")
      (format s "/bin/chmod 777 $OUTPUT./temp~%")
      (format s "/bin/mkdir -p $OUTPUT./tmp~%")
      (format s "/bin/chmod 777 $OUTPUT./tmp~%")
    for key being the hash-keys of image-hash-table
    using (hash-value value)
    as  path = (format nil "~a~a" value key)
    do
    (format s "~%")
    (format s "if [ -a $IMAGE./~a ]\; then~%" path)
    (format s "        echo ~a already exists >> exists.log~%" path)
    (format s "else~%")
    (format s "        curl --retry 0 -f -O $COMMONSPATH./~a~%" path)
    (format s "        if [ -a $IMAGE./~a ]\; then~%" key)
    (format s "                /bin/mkdir -p $OUTPUT./~a~%" value)
    (format s "                /bin/mv ./~a $OUTPUT./~a~%" key value)
    (format s "                echo ./~a downloaded >> download.log~%" path)
    (format s "        else~%")
    (format s "                curl --retry 0 -f -O $IMAGEPATH./~a~%" path)
    (format s "                if [ -a $IMAGE./~a ]\; then~%" key)
    (format s "                        /bin/mkdir -p $OUTPUT./~a~%" value)
    (format s "                        /bin/mv ./~a $OUTPUT./~a~%" key value)
    (format s "                        echo ./~a downloaded >> download.log~%" path)
    (format s "                else~%")
    (format s "                        echo ./~a failed >> failed.log~%" path)
    (format s "                fi~%")
    (format s "        fi~%")
    (format s "fi~%")
    ))))

(defun fsm-function (file-type file-state &key (put nil))
  "Determining function to use according to *type-state-function* table"
  (debug-message-value (_ "file type, state")
		       (format nil "~a, ~a" file-type file-state))
  (unless (member file-type *type* :test #'string=)
    (put-message-value-fail-and-abort (_ "Error: FSM invalid type") file-type))
  (unless (member file-state *state* :test #'string=)
    (put-message-value-fail-and-abort (_ "Error: FSM invalid state") file-state))
  (dolist (tuple *type-state-function* nil)
    (when (and (string= file-type  (first  tuple))
	       (string= file-state (second tuple)))
      (progn
	(unless (null put)
	  (put-message-value (_ "file type, state, and function")
			     (format nil "~a, ~a, ~a" 
				     (first  tuple)      ; type
				     (second tuple)      ; state
				     (third  tuple))))   ; function
	(return (third tuple)))))) ; break loop        ; <--- return function

(defun fsm-images-count (file-name &key (put nil))
  "Counting images in the images directory"
  (put-message-value (_ "counting images") file-name)
  (let* ((name-split     (regexp:regexp-split "-" file-name))
	 (dir-name-image (format nil "~a~a/~a/"
				 *whereis-directory-images*
				 (first (last name-split 2)) 
				 (first (last name-split  ))))
	 (image-count    (shell-count-dir dir-name-image)))
    (sql-update-file file-name :images image-count :put put)
    (put-message-value-done (_ "counting images") image-count))
  t)

(defun fsm-images-chown (file-name &key (put nil))
  "Setting ownership of image, math, and thumb directories"
  ;;
  ;; Design note:
  ;;
  ;; Wikix downloads image files, and stores them with `root:root'
  ;; ownership.  However, when you browse the mirrored Wikipedia
  ;; articles, MediaWiki will not generate thumbs of image files with
  ;; `root:root' ownership, but will instead render gray rectangles.
  ;;
  ;; MediaWiki expects `www-data:www-data' ownership.  So we `chown'.
  ;;
  (put-message-value (_ "chowning images") file-name)
  (let* ((name-split (regexp:regexp-split "-" file-name))
	 (dir-name-image (format nil "~a~a/~a/" 
				 *whereis-directory-images*
				 (first (last name-split 2)) 
				 (first (last name-split  ))))
	 (dir-name-math  (format nil "~amath/~a/~a/"
				 *whereis-directory-images*
				 (first (last name-split 2)) 
				 (subseq (first (last name-split  )) 1 2)))
	 (dir-name-thumb (format nil "~athumb/~a/~a/"
				 *whereis-directory-images*
				 (first (last name-split 2)) 
				 (first (last name-split  )))))
  ;; 1) directories which hold images
    (shell-chown-dir dir-name-image)
  ;; 2) `math' directory which holds equations formated with TeX 
    (shell-chown-dir dir-name-math)
  ;; 3) `thumb' directory which holds resized images
    (shell-chown-dir dir-name-thumb))
  t)

(defun fsm-images-rebuild (file-name &key (put nil))
  "Running `rebuildImages.php' to load image meta-data into `xxwiki.image'"
  ;;
  ;; Design note:
  ;;
  ;; Problem: It seems that `rebuildImages.php' has a memory leak.
  ;; For every 20,000 new images processed, it captures 1G memory.  
  ;;
  ;; Solution: To deal with this, we impose a time-out to kill the
  ;; forked process before it occupies all memory.  Then we start
  ;; over.
  ;;
  ;; Timing:  
  ;;  o Old images are checked at about 150 files/sec.
  ;;  o New image files are processed at about 5 files/sec.
  ;;
  ;; Problem: Now the `en' wikipedia has about 2 million images.  Each
  ;; time we start over, we must first loop through image files that
  ;; have already been processed, before we can get to the new images.
  ;;
  ;; Solution: We start with a modest time-out, and gradually increase
  ;; it, up to a cap.  The very last file has no time-out and serves
  ;; to finish-up any remaining unprocessed image files.
  ;;
  ;; Assuming
  ;;  o *mirror-rebuildimages-timeout-sec-min* =  6000
  ;;  o *mirror-rebuildimages-timeout-sec-inc* =   200
  ;;  o *mirror-rebuildimages-timeout-sec-max* = 20000  <-- cap for safety
  ;;
  ;; The time-out schedule (for initial build) would look like this:
  ;;
  ;;                             Time-out  Images  Images 
  ;; File-name                      [sec]     Old     New
  ;; ---------------------------+--------+-------+-------
  ;; xxwiki-yyyymmdd-images-0-00     6000       0   30000
  ;; xxwiki-yyyymmdd-images-0-01     6200   30000   30000
  ;; xxwiki-yyyymmdd-images-0-02     6400   60000   30000
  ;; ...
  ;; xxwiki-yyyymmdd-images-4-45    19800 2070000   30000
  ;; xxwiki-yyyymmdd-images-4-46    20000 2100000   30000
  ;; xxwiki-yyyymmdd-images-4-47    20000 2130000   29000
  ;; ...
  ;; xxwiki-yyyymmdd-images-f-ff      nil
  ;;
  ;; Also, if the number of images to be processed falls below some
  ;; threshold, we do nothing (i.e. just skip ahead to the next file),
  ;; the exception to this rule being the last file `...images-f-ff'
  ;; where we process any and all remaining images.
  ;;
  ;; If we are updating a large existing mirror, then we should
  ;; increase the increment.
  ;;
  (let* ((language      (sql-select-file-language file-name :put put))
	 (down-images   (shell-count-images))
	 (import-images (sql-select-count-xxwiki-image language :put put))
         (valid-images  (sql-select-sum-images-pages language :put put))
	 (diff-images   (- down-images import-images))
	 (threshold     *mirror-rebuildimages-threshold*)
	 (threshold-p   (> diff-images threshold))
	 (index         (file-name-to-index file-name))
	 (time-out      (min
			 (+ *mirror-rebuildimages-timeout-sec-min*
			    (* index *mirror-rebuildimages-timeout-sec-inc*))
			 *mirror-rebuildimages-timeout-sec-max*))
	 (last-file-p   (regexp:match "f-ff" file-name)))
    (put-message-value (_ "images down / import / valid") 
		       (format nil "~d / ~d / ~d" 
			       down-images import-images valid-images))
    (if threshold-p
	(progn
	  (put-message-value (_ "threshold exceeded - rebuilding")
			     (format nil "d-i=~d > ~d"
				     diff-images threshold))
	  (shell-mediawiki-farm-rebuildimages file-name :time-out time-out))
      (put-message-value (_ "threshold not exceeded - skipping")
			 (format nil "d-i=~d <= ~d"
				     diff-images threshold)))
    (when last-file-p
      (put-message (_ "rebuilding images - forced"))
      (shell-mediawiki-farm-rebuildimages file-name :time-out nil)))
  t)

(defun file-name-to-index (file-name)
  "Computing number from the last two chars of `xxwiki-yyyymmdd-images-y-yz'"
  (let ((s (first    ; `yz' where {y,z in [0-9a-f]}
	    (last
	     (regexp:regexp-split "-" file-name)))))
    (parse-integer s :radix 16)))

(defun fsm-images-validate (file-name &key (put nil))
  "Validating downloaded images, sequestering corrupt ones"
  ;;
  ;; Design note:
  ;;
  ;; Wikix generates `i-chunk's (image download scripts) that
  ;; incorrectly handle the case of partial downloads.
  ;; `http://upload.wikimedia.org/' has thousands of images that
  ;; download only part way, and then block.  Without a timeout, the
  ;; mirror would never finish.
  ;;
  ;; CURL can be configured with a timeout.  The option is `--max-time
  ;; <seconds>'.  We can edit the option into the scripts, or better
  ;; yet append the option to the curl configuration file
  ;; `/root/.curlrc'.  This author recommends
  ;;
  ;;   # We want a 1 ks timeout (initially and for each retry)
  ;;   --max-time 1000
  ;;
  ;; However, when cURL times out, the `i-chunk's will then store the
  ;; partially downloaded (i.e. corrupt) image files just as if they
  ;; had been downloaded successfully.
  ;;
  ;; We must later hunt for these corrupt files and sequester them.
  ;; For this we use `gm identify -verbose [file]'.  For example,
  ;;
  ;;   /var/lib/mediawiki/images/3/35# gm identify -verbose \
  ;;   Model_of_ancient_Olympia.jpg
  ;;
  ;; When we look at the last line, it reads 
  ;;
  ;;   gm identify: Corrupt JPEG data: premature end of data segment
  ;;  (Model_of_ancient_Olympia.jpg).
  ;;
  ;; We must iterate over all images, sequestering those identified as
  ;; `Corrupt'.
  ;;
  ;; Design note:
  ;;
  ;; We sequester (rather than delete) bad image files, by moving them
  ;; to a `bad-images' directory.  First this lets us examine them
  ;; later.  Second, a programming error during development could
  ;; inadvertently sequester ALL image files.  Sequestration can be
  ;; reversed easily, whereas deletion can only be remedied at a cost
  ;; of redownloading 1TB of image files (and say two months delay).
  ;;
  ;; Design note:
  ;;
  ;; We do not check the SVG files with `gm'.  Some of them cause
  ;; `gm' to hang.  Sometimes error messages and a back-trace is 
  ;; emitted, suggesting that malloc is involved.  Example files 
  ;; include: `/var/lib/mediawiki/images/0/00/IFK_Mariehamnin_logo.svg'
  ;; `/var/lib/mediawiki/images/0/00/Arms_of_Infante_Charles_of_Spain_as_King_of_Naples_and_Sicily.svg'
  ;;
  ;; We check some of the SVG files another way.  If an SVG file is
  ;; small and contains the text `302 Redirected', then it is actually
  ;; an error message returned by the proxy, and is sequestered.
  ;;
  ;; Not checking is unsatisfactory.  Someday we should explore ways
  ;; of letting `gm' timeout.
  ;;
  ;;
  (when *mirror-image-validate-p*
    (put-message (_ "validating images"))
    (put-message "-----") 
    (let* ((language       (sql-select-file-language file-name))
	   (name-split     (regexp:regexp-split "-" file-name))
	   (d1             (first (last name-split 2)))
	   (d2             (first (last name-split  )))
	   (dir-name-image (format nil "~a~a/~a/" 
				   *whereis-directory-images*
				   d1 d2)))
      ;; 1) sequester if file-name contains `%' char
      (put-message-start (_ "looking for invalid names (with `%' char)"))
      (loop
	for file-image of-type pathname in (shell-ls-grep dir-name-image "%")
	do (when (regexp:match "%" (file-namestring file-image))
	     (sequester-bad-image (format nil "~a~a" dir-name-image file-image)
				  :put put)))
      (put-message-done (_ "looking for invalid names (with `%' char)"))
      ;; 2) sequester if file is actually an error message from proxy,
      ;;    such files are small (< 500 bytes)
      (put-message-start (_ "looking for files that are proxy error messages (`302 Redirected')"))
      (loop
	for path-image of-type pathname in (shell-find-1k dir-name-image)
	do (when (< (posix:file-size path-image) 1000)
	     (unless (null put)
	       (put-message path-image))
	     (unless (null (shell-cat-grep path-image "302 Redirected"))
	       (sequester-bad-image path-image :put put))))
      (put-message-done (_ "looking for files that are proxy error messages (`302 Redirected')"))
      ;; 3) sequester if `gm identify -verbose' calls file corrupt,
      ;;    but skip validation if already imported into `xxwiki.image'
      (put-message-start (_ "looking for corrupt image files"))
      (loop
	with images-validated = 0
	for path-image of-type pathname in (directory
					    (format nil "~a*" dir-name-image))
	as  norm-string = (file-namestring path-image)
        as  no-escape   = (remove #\\ norm-string)
	;; md5sum computed from unescaped file name
	as  md5sum      = (if (find #\` norm-string)
			      ;; backquote OK,  fail on unicode
			      (string-to-md5sum no-escape)
			    ;; unicode OK, fail on backquote
			    (shell-echo-md5sum no-escape))
	as  md5-d1      = (subseq md5sum 0 1)
	as  md5-d2      = (subseq md5sum 0 2)
	as  validp      = (string= (sql-select-image-state md5sum :put put)
				  "valid")
	as  result = (unless (or 
			      ;; skip if already validated
			      validp
			      ;; skip certain problem cases
			      (string= (pathname-type path-image) "svg")
			      (string= (pathname-type path-image) "SVG"))
		       (shell-gm-identify path-image :put put))
	do (unless (null put) (put-message path-image))
	;; null result usually means image file is OK
	when (and (null result) (null validp))
	  do (progn
		     (incf images-validated)
		     (sql-update-file file-name
				      :pages images-validated          ; hack
				      :put put)
		     (sql-insert-image md5sum :state 'valid' :put put))
	;; result usually means either trouble or no decode delegate
	when result
	  do (put-message result)
	  and do (if (or
		      ;; usually caused by partial download
		      (regexp:match "Corrupt" result)
		      ;; caused by proxy `302 Redirected' error
		      (regexp:match "Improper image header" result) ;GIF,PNG
		      (regexp:match "Not a JPEG file" result)       ;JPG
		      (regexp:match "Not a TIFF or MDI file" result);TIF
		      (regexp:match "Opening and ending tag mismatch"
				    result)                         ;SVG
		      ;; caused by escaping (or not) the file name
		      (not (string= d1 md5-d1))
		      (not (string= d2 md5-d2))
		      )
		     ;; sequester (do not mark as validated)
		     (sequester-bad-image path-image :put put)
		   (progn
		     ;; keep OGG,PDF,WAV files (do not revalidate)
		     ;; `No decode delegate for this image format'
		     (put-message (_ "keeping file"))
		     (incf images-validated)
		     (sql-update-file file-name
				      :pages images-validated          ; hack
				      :put put)
		     (sql-insert-image md5sum :state 'valid' :put put)
		     (put-message "-----"))))
      (put-message-done (_ "looking for corrupt image files"))
      ))
  t)

(defun sequester-bad-image (pathname-old &key (put nil))
  "Moving bad-image to a bad-images directory"
  (let* ((pathname-bad (parse-namestring *whereis-directory-images-bad*))
	 ;; change directory only (keep host, device, name, type, version)
	 (pathname-new (merge-pathnames
			(make-pathname
			 :directory (append
				     (pathname-directory pathname-bad)
				     (last (pathname-directory pathname-old) 
					   2)))
			pathname-old)))  ; use as default
    (if (wild-pathname-p pathname-old)
	;; eliminate this branch when we learn how to deal with wildcards
	(progn
	  (put-message pathname-old)
	  (put-message (_ "pathname contains a wildcard"))
	  (put-message (_ "keeping file")))
      (progn
	(put-message (_ "sequestering file to `bad-images' directory --->"))
	(put-message pathname-new)
	;; we might have already sequestered this file
	(when (probe-file pathname-new)
	  (delete-file pathname-new))
	(rename-file pathname-old pathname-new)))
    (put-message "-----")
    ))

(defun fsm-no-op (file-name &key (put nil))
  "Doing nothing"
  ;; Files and states:
  ;;           |input file          generated file(s)
  ;; ----------+-------------------+--------------------------
  ;; file-name |don't care          none
  ;; state-in  |don't care             
  ;; state-out |no change               
  ;; Return: nil means fail, t means done.
  t) ; never fails

(defun fsm-process-file (file-name &key (put nil))
  "Processing file with finite-state-machine"
  (let ((message (_ "processing file with finite-state-machine"))
	(file-type (sql-select-file-type  file-name :put put))
	(old-state (sql-select-file-state file-name :put put))
	(new-state "error")
	(input     "error"))
    (put-timestamp-message-start message)
    (debug-message file-name)
    ;; 1) give FSM the `start' input, store new `state', increment `updates'
    (setq input "start")
    (setq new-state (fsm-transition old-state input :put t))
    (sql-update-file-state file-name new-state :put nil)
    (setq old-state new-state)
    ;; 2) process file with appropriate function
    (if (apply (fsm-function file-type new-state :put t)
	       (list file-name :put put))
	(setq input "done")
      (setq input "fail"))
    ;; 3) give FSM the `done' or `fail' input, and store new state
    (setq new-state (fsm-transition old-state input :put t))
    (unless (string= old-state new-state)
      (sql-update-file-state file-name new-state :put put))
    (put-timestamp-message-done message)))

(defun fsm-transition (state input &key (put nil))
  "Changing file state according to *state-transition* table"
  (debug-message-value (_ "initial state and input")
		       (format nil "~a + ~a" state input))
  (unless (member state *state* :test #'string=)
    (put-message-value-fail-and-abort (_ "Error: FSM invalid state") state))
  (unless (member input *input* :test #'string=)
    (put-message-value-fail-and-abort (_ "Error: FSM invalid input") input))
  (dolist (transition *state-transition* nil)
    (when (and (string= state (first transition))
	       (string= input (second transition)))
      (progn
	(unless (null put)
	  (put-message-value (_ "state transition") 
			     (format nil "~a + ~a --> ~a" 
				     (first  transition)
				     (second transition)
				     (third  transition))))
	(return (third transition)))))) ; break loop

(defun fsm-working-chown (file-name &key (put nil))  
  "Setting working directory ownership"
  (null (shell-chown-dir *whereis-directory-working*)))


;;;;--------------------------------------------------------------------------+
;;;; Log utilities:                                                           |
;;;;   Note:  `put-message-*' functions should call `put-and-log-string'.     |
;;;;   Note:  Do not use `dribble', which is intended for logging interactive |
;;;;          sessions, and turns out to be non-portable.                     |
;;;;--------------------------------------------------------------------------+

		     
(defun log-start ()
  "Starting log"
  (put-message-value (_ "creating") *whereis-file-log*)
  (defparameter *log-enabled-p* t)
  (put-timestamp-message-start "Starting log")
  t)

(defun log-stop ()
  "Stopping log"
  (put-message-value (_ "closing") *whereis-file-log*)
  (put-timestamp-message-done "stopping log")
  (defparameter *log-enabled-p* nil)
  t)

(defun log-test (f &rest args)
  "Wrapping evaluation of a boolean valued function with logging"
  (let ((doc (_ (documentation f 'function))))
    (put-timestamp-message-start doc)
    (let ((res (apply f args)))
      (if res
	  (put-timestamp-message-done doc)
	(progn
	  (put-message (gethash f *error-messages*))
	  (put-timestamp-message-fail doc)))
      res)))

(defun log-test-set (function-list)
  "Running tests and logging result"
  (let ((message (_ "logging test set"))
	(all-pass t))
    (put-timestamp-message-start message)
    (dolist (func function-list all-pass)
      (setq all-pass (and (log-test func) all-pass)))
    (if all-pass
	(put-timestamp-message-done message)
      (put-timestamp-message-fail message))
    all-pass))

(defun log-test-set-or-die (function-list)
  "Running tests, logging results, and die if any fail"
  (unless (log-test-set function-list)
    (abort-and-die)))

(defun formatted-timestamp ()
  "Formatting decoded universal time as a human readable string"
  (multiple-value-bind (s m h dd mm yyyy dow dst tz)
      (get-decoded-time)
    (format nil "~4d-~2,'0d-~2,'0d ~2,'0d:~2,'0d:~2,'0d" yyyy mm dd h m s)))

(defun put-timestamp-message-start (message)
  "Logging timestamped message"
  (put-message-start (format nil "~a ~a" (formatted-timestamp) message)))

(defun put-timestamp-message-done (message)
  "Logging timestamped message done"
  (put-message-done (format nil "~a ~a" (formatted-timestamp) message)))

(defun put-timestamp-message-fail (message)
  "Logging timestamped message fail"
  (put-message-fail (format nil "~a ~a" (formatted-timestamp) message)))

(defun put-timestamp-message-fail-and-abort (message)
  "Logging timestamped message fail and abort"
  (put-message-fail-and-abort 
   (format nil "~a ~a" (formatted-timestamp) message)))

;;;;--------------------------------------------------------------------------+
;;;; Mirror mode:                                                             |
;;;;   0) Initialize parameters (command line, configuration file)            |
;;;;   1) Assert prerequisites                                                |
;;;;   2) Run in mirror mode                                                  |
;;;;   2.1) download checksums and dump file, validate, and decompress        |
;;;;   2.2) split dump into x-chunks                                          |
;;;;   2.3) generate i-chunks                                                 |
;;;;   2.4) run i-chunks (download images)                                    |
;;;;   2.5) validate downloaded image files, sequester corrupt ones           |
;;;;   2.6) import x-chunks (import pages)                                    |
;;;;--------------------------------------------------------------------------+

(defun language-to-file-name (language)
  "Replacing dashes with underscores"
  ;; replace `zh-classical' with `zh_classical'
  (substitute #\_ #\- (string-trim " " language)))

(defun checksums-url (language)
  "Generating URL of wikimedia checksums file for a given language"
  (let ((message               (_ "generating URL for checksums file"))
	(replace-this-language "xx")
	(with-this-language    (language-to-file-name language))
	(url                   nil))
    (put-message-start message)
    (put-message-value (_ "*wikimedia-site*")      *wikimedia-site*)
    (put-message-value (_ "*wikimedia-checksums-template*")
		       *wikimedia-checksums-template*)
    (when (regexp:match replace-this-language *wikimedia-checksums-template*)
      (setq url (format nil "~a~a" *wikimedia-site*
			(join
			 (regexp:regexp-split replace-this-language
					      *wikimedia-checksums-template*)
			 with-this-language))))
    (put-message url)
    (put-message-value-done message language)
    url))

(defun dump-url (language)
  "Generating URL of wikimedia dump file for a given language"
  (let* ((message               (_ "generating URL for dump file"))
	 (dummy-date            "yyyymmdd")
	 (replace-this-language "xx")
	 (with-this-language    (language-to-file-name language))
	 (file-name             (dump-file-name language))
	 (file-date             (sql-select-file-date file-name))
	 (url                   nil))
    (put-message-start message)
    (put-message-value (_ "*wikimedia-site*")      *wikimedia-site*)
    (put-message-value (_ "*wikimedia-dump-template*")
		       *wikimedia-dump-template*)
    ;; replace language
    (when (regexp:match replace-this-language *wikimedia-dump-template*)
      (setq url (format nil "~a~a" *wikimedia-site*
			(join
			 (regexp:regexp-split replace-this-language
					      *wikimedia-dump-template*)
			 with-this-language))))
    ;; replace date
    (when (regexp:match dummy-date url)
      (setq url    	(join
			 (regexp:regexp-split dummy-date url)
			 file-date)))
    (put-message url)
    (put-message-value-done message language)
    url))

(defun checksums-file-name (language)
  "Generating file-name of checksums file for a given language"
  (put-message-start (_ "generating checksums file-name"))
  (let* ((replace-this-language  "xx")
	 (with-this-language     (language-to-file-name language))
	 (file-name (join (regexp:regexp-split 
			   replace-this-language
			   *whereis-file-checksums-template*)
			  with-this-language)))
    (put-message-value language file-name)
    (put-message-done (_ "generating checksums file-name"))
    file-name))

(defun dump-file-name (language)
  "Generating file-name of dump file for a given language"
  (sql-select-dump-name language))

(defun checksums-path (language)
  "Generating pathname of checksums file for a given language"
  (put-message-start (_ "generating checksums pathname"))
  (let* ((file-name  (checksums-file-name language))
	 (pathname   (merge-pathnames
		      (parse-namestring *whereis-directory-working*)
		      (parse-namestring file-name))))
    (put-message pathname)
    (put-message-done (_ "generating checksums pathname"))
    pathname))

(defun dump-path (language)
  "Generating pathname of dump file for a given language"
  (put-message-start (_ "generating dump pathname"))
  (let* ((file-name  (dump-file-name language))
	 (pathname   (merge-pathnames
		      (parse-namestring *whereis-directory-working*)
		      (parse-namestring file-name))))
    (put-message pathname)
    (put-message-done (_ "generating dump pathname"))
    pathname))

(defun dump-file-remove-all (language &key (put nil))
  "Removing all dump files and state info"
  (shell-remove-all-working-files language)
  (assert-database-administrator-credentials-p)
  (put-message-start (_ "deleting state info"))
  (sql-delete-from language :put put)
  (put-message-done (_ "deleting state info"))
  t)

(defun dump-file-remove-p (language &key (put nil))
  "Removing any old dump files and state info"
  (shell-remove-old-working-files language)
  (put-message-start (_ "deleting state info"))
  (sql-delete-from language :put put)
  (put-message-value-done (_ "deleting state info") language)
  t)

(defun ichunks-complete-p (file-name &key (put nil))
  "Determining if all `ichunk's are all processed"
  ;; if no `images' records exist, and
  ;; if no `ichunk's remain to be processed,
  ;; then insert `xxwiki-yyyymmdd-images-x-xx' records into the
  ;; `wpmirror.file' table
  (unless (null put)
    (put-message-start (_ "ichunks-complete-p")))
  (let* ((language  (sql-select-file-language file-name :put put))
	 (date      (sql-select-file-date     file-name :put put))
	 (result (and (zerop (sql-select-count-images  language :put put))
		      (zerop (sql-select-count-ichunk-remaining
			      language :put put))
		      (> (sql-select-count-ichunk-done language :put put)
			 0))))
    (when result
      (put-message-value-done (_ "ichunks complete") result)
      (put-message-start (_ "inserting `images' records"))
      (loop named images-validate
	with hex = (the simple-string "0123456789abcdef")
	for c1 of-type standard-char across hex
	do
	(loop
	  for c2 of-type standard-char across hex
	  do
	  (sql-insert-file language date
			   (format nil "~awiki-~a-images-~a-~a~a"
				   (language-to-file-name language)
				   date c1 c1 c2)
			   "images" :put nil))
	(sleep 1))
      (put-message-done (_ "inserting `images' records")))
    result))

(defun ichunk-edit-p (ichunk-name)
  "Editing `ichunk' bash scripts"
  ;;
  ;; Design note:
  ;;
  ;; `Wikix' has an option `-p' (meaning parallel) that lets one
  ;; generate 16 `bash' scripts which may then be run concurrently.
  ;; In these scripts, `curl' is always invoked with a `--retry 7'
  ;; option.  This makes sense when trying to run so many scripts
  ;; concurrently.  However, `wp-mirror' runs its scripts
  ;; sequentially, so `--retry 0' is better.
  ;;
  (let ((temp-file (concatenate 'string ichunk-name ".temp"))
	(replace-this-string "--retry 7")     ; gain speed with fewer retries
	(with-this-string                     ; 0 or 1 retries are good
	 (format nil "--retry ~d" *ichunk-curl-retry*)))
    (with-open-file (f ichunk-name
		       :direction :input
		       :if-does-not-exist nil)
      (with-open-file (g temp-file
			 :direction :output
			 :if-exists :supersede)
	(do ((l (read-line f nil 'eof) (read-line f nil 'eof)))
	    ((eq l 'eof) nil)
	  (cond ((regexp:match replace-this-string l)
		 (write-line (join (regexp:regexp-split replace-this-string l)
				   with-this-string) g))
		(t (write-line l g))))))
    (null (shell-rename-file temp-file (file-namestring ichunk-name)))))

(defun join (list-of-strings join-string)
  "Joining strings"
  (let ((n (length list-of-strings)))
    (cond ((eql n 0) '())
	  ((eql n 1) (first list-of-strings))
	  ((eql n 2) (concatenate 'string
				  (first list-of-strings)
				  join-string
				  (second list-of-strings)))
	  (t (concatenate 'string
			  (first list-of-strings)
			  join-string
			  (second list-of-strings)
			  join-string
			  (join (cddr list-of-strings) join-string))))))

(defun wait-for-internet-access (file-name &key (put nil))
  "Waiting for internet access, if processing the file involves a download"
  ;; We assume lock has been grabbed
  (let* ((message (_ "wait for internet access, if download required"))
	 (file-type (sql-select-file-type  file-name :put put))
	 (old-state (sql-select-file-state file-name :put put))
	 (input     "start")
	 (new-state (fsm-transition old-state input     :put put))
	 (fsm-func  (fsm-function   file-type new-state :put put)))
    ;; 0) when a download is required, wait for internet access.
    ;;    sometimes your Internet Service Provider (ISP) will go down
    ;;    for a few hours during the night
    (when (member fsm-func *fsm-function-needs-internet-access*)
      (put-message message)
      (loop named wait-access
	;; actually this will not detect a failure, if your traffic is
	;; going through a caching web proxy, and the failure lies
	;; between the proxy and wikimedia
	until (assert-internet-access-to-wikimedia-site-p :put put)
	do
	(put-message-value (_ "no access to wikipedia - sleep")
			   *monitor-poll-sec*)
	(sleep *monitor-poll-sec*)))))

(defun main-mirror (&key (main-mode :first-mirror) (put nil))
  "Purpose: Main function when wp-mirror is in Mirror Mode."
  (put-message (_ "-----initializing-begin-----------------------"))
  (log-test-set-or-die '(log-start
			 process-configuration-files-or-die
			 count-cpu
			 put-parameters))
  (put-message (_ "-----initializing-done------------------------"))
  ;; 1) Assert prerequisites
  (put-message (_ "-----asserting-prerequisite-hardware-begin----"))
  (log-test-set-or-die '(assert-disk-space-p))
  (log-test-set-or-die '(assert-physical-memory-p))
  (log-test-set        '(assert-internet-access-to-wikimedia-site-p))
  (put-message (_ "-----asserting-prerequisite-hardware-done-----"))
  (put-message (_ "-----asserting-prerequisite-software-begin----"))
  (log-test-set-or-die '(assert-clisp-features-p))
  (log-test-set-or-die '(assert-utilities-p
			 assert-curl-max-time-p))
  (log-test-set-or-die '(assert-image-directory-p
			 assert-working-directory-or-create
			 assert-images-bad-directory-or-create))
  (log-test-set-or-die '(assert-mediawiki-import
			 assert-mediawiki-extensions))
  (log-test-set-or-die '(assert-mediawiki-adminsettings-p
			 assert-mediawiki-localsettings-p
			 assert-mediawiki-localsettings-image
			 assert-mediawiki-localsettings-tex
			 assert-mediawiki-localsettings-tidy
			 assert-php5-suhosin-p))
  (log-test-set-or-die '(assert-database-administrator-credentials-p
			 ;; `assert-hdd-write-cache-disabled-p' (next)
			 ;; would be under hardware, but, the current
			 ;; implementation queries MySQL as to the
			 ;; location of the InnoDB data file path,
			 ;; and that requires database credentials
			 assert-hdd-write-cache-disabled-p
			 assert-concurrency-limit-xchunk-p
			 assert-dbms-mysql-p
			 assert-database-administrator-account-p))
  (log-test-set-or-die '(assert-database-wikidb-p
			 assert-database-template-or-create
			 assert-database-xxwiki-or-create
			 assert-database-wpmirror-or-create))
  (log-test-set-or-die '(assert-virtual-host-p
			 assert-virtual-host-name-resolution-p))
  (log-test-set        '(warn-if-detect-proxy))
  (put-message (_ "-----asserting-prerequisite-software-done-----"))
  ;; 2) Mirror mode
  (put-message (_ "-----mirror-mode-begin------------------------"))
  (log-test-set-or-die '(fsm-boot))
  (when (eql main-mode :first-mirror)
    (release-all-file-semaphores))
  (loop named mirror
    for i from 0 to 1
    while (first-mirror-exists-p) ; all mirrors exit if the first does
    do
    ;; first pass does all the work
    ;; second pass retries any failed `xchunks' and `ichunks'
    (loop named mirror-pass
      ;;   2a) decide order in which files should be processed,
      ;;       select one, and grab its lock (set its semaphore to 0)
      for file-name of-type string = (sql-select-next-file :put put)
      while (and (first-mirror-exists-p) ; all mirrors exit if the first does
		 file-name)              ; if no file found, bail
        do
          ;; 2b) wait if necessary
          (wait-for-internet-access file-name)
	  ;; 2c) process the file
	  (fsm-process-file file-name :put put) ; <--- heavy lifting
	  ;; 2d) release its lock (set semaphore to 1)
	  (sql-update-file-semaphore file-name 1)
	  ;; 2e) we should routinely delete some rows from the `objectcache'
	  (sql-delete-from-objectcache (sql-select-file-language file-name))
	  ;; 2f) see if `ichunk's all done, if so generate `images'
	  (ichunks-complete-p file-name)
	  (put-message "-----")
	  ;;(die) ; uncomment to `single-step' for debug purposes
	  )
    (when (and (zerop i) (i-am-the-first-mirror-p))
	;; give failed chunks a second chance
	(sql-update-fail-to-pending)   ; give failed chunks a second chance
	;;(sql-update-images-to-pending) ; run `chown' again
      ))
  (put-message (_ "-----mirror-mode-done-------------------------"))
  (put-message (_ "-----cleaning-up-begin------------------------"))
  (log-test-set '(clear-pidfile
		  log-stop))
  (put-message (_ "-----cleaning-up-done-------------------------")))


;;;;--------------------------------------------------------------------------+
;;;; Monitor mode:                                                            |
;;;;   0) Initialize parameters (command line, configuration file)            |
;;;;   1) Assert prerequisites                                                |
;;;;   2) Determine monitor mode (:auto, :gui, :screen, :text)                |
;;;;--------------------------------------------------------------------------+


(defvar       *report*                  nil)

(defstruct report
  ;; timestamp
  timestamp
  ;; wiki
  language
  ;; activities (boolean)
  dump-file-null
  dump-file-downloading
  dump-file-validating
  dump-file-decompressing
  dump-file-page-counting
  dump-file-image-counting
  xchunk-generating
  ichunk-generating
  images-counting
  images-validating
  ;; downloading
  file-dump-name
  file-dump-size
  file-dump-xml-name
  file-dump-xml-size
  ;; estimating
  page-count
  image-count
  chunk-count
  ;; chunk generating
  xchunks-generated
  xchunks-generated-progress
  ichunks-generated
  ichunks-generated-progress
  ;; ichunks
  ichunks-running
  ichunks-done
  ichunks-failed
  ichunks-remaining
  ichunks-progress
  images-count
  images-start
  images-created
  images-valid
  images-imported
  images-sum
  images-count-progress
  images-valid-progress
  images-imported-progress
  ;; xchunks
  xchunks-running
  xchunks-done
  xchunks-failed
  xchunks-remaining
  xchunks-progress
  pages-imported
  ;; InnoDB
  innodb-buffer-pool-pages-total
  innodb-buffer-pool-pages-dirty
  innodb-buffer-pool-pages-dirty-progress
  innodb-buffer-pool-pages-insert
  innodb-buffer-pool-pages-insert-progress
  innodb-buffer-pool-pages-misc
  innodb-buffer-pool-pages-misc-progress
)

(defun monitor-data-collect (lang &key (put nil))
  "Collecting status data suitable for monitors"
  ;; the idea is this:
  ;; `/etc/wp-mirror/default.conf' is needed for *xchunk-page-count*,
  ;; and all else comes from the `wpmirror.file' table
  (unless (null put)
    (put-message-start (_ "collecting status data for monitor")))
  (let* ((file-dump-name   (sql-select-dump-name       lang :put put))
	 (file-dump-xml-name (sql-select-dump-xml-name lang :put put))
	 (chunk-count      (ceiling (/ (sql-select-file-pages
					file-dump-xml-name :put put)
				       *xchunk-page-count*)))
	 (ichunk-count     (sql-select-count-ichunk           lang :put put))
	 (ichunk-done      (sql-select-count-ichunk-done      lang :put put))
	 (ichunk-fail      (sql-select-count-ichunk-fail      lang :put put))
	 (ichunk-remaining (sql-select-count-ichunk-remaining lang :put put))
	 (ichunk-running   (sql-select-count-ichunk-running   lang :put put))
	 (xchunk-count     (sql-select-count-xchunk           lang :put put))
	 (xchunk-done      (sql-select-count-xchunk-done      lang :put put))
	 (xchunk-fail      (sql-select-count-xchunk-fail      lang :put put))
	 (xchunk-remaining (sql-select-count-xchunk-remaining lang :put put))
	 (xchunk-running   (sql-select-count-xchunk-running   lang :put put))
 	 (pages-imported   (sql-select-count-xxwiki-page      lang :put put))
	 (images-count     (sql-select-count-images           lang :put put))
	 (images-start     (sql-select-count-images-start     lang :put put))
	 (images-created   (sql-select-count-images-created   lang :put put))
         (images-valid     (sql-select-count-image-valid           :put put))
	 ;; hack - storing number of validated images in the `pages' field
	 ;(images-valid     (sql-select-sum-images-pages       lang :put put))
	 ;(images-valid     (sql-select-count-images-valid     lang :put put))
 	 (images-imported  (sql-select-count-xxwiki-image     lang :put put))
	 (images-sum       (sql-select-sum-images-images      lang :put put))
	 (innodb-buffer-pool-pages-dirty 
	  (sql-show-innodb-buffer-pool-pages-dirty    :put put))
	 (innodb-buffer-pool-pages-insert
	  (sql-show-innodb-buffer-pool-pages-insert   :put put))
	 (innodb-buffer-pool-pages-misc
	  (sql-show-innodb-buffer-pool-pages-misc     :put put))
	 (innodb-buffer-pool-pages-total
	  (sql-show-innodb-buffer-pool-pages-total    :put put)))
    (unless (null put)
      (put-message-value (_ "chunk-count")     chunk-count)
      (put-message-value (_ "ichunk-count")    ichunk-count)
      (put-message-value (_ "ichunk-done")     ichunk-done)
      (put-message-value (_ "ichunk-fail")     ichunk-fail)
      (put-message-value (_ "xchunk-count")    xchunk-count)
      (put-message-value (_ "xchunk-done")     xchunk-done)
      (put-message-value (_ "xchunk-fail")     xchunk-fail)
      (put-message-value (_ "images-count")    images-count)
      (put-message-value (_ "images-start")    images-start)
      (put-message-value (_ "images-created")  images-created)
      (put-message-value (_ "images-valid")    images-valid)
      (put-message-value (_ "images-imported") images-imported)
      (put-message-value (_ "images-sum")      images-sum)
      (put-message-value (_ "innodb-buffer-pool-pages-dirty") 
			 innodb-buffer-pool-pages-dirty)
      (put-message-value (_ "innodb-buffer-pool-pages-insert") 
			 innodb-buffer-pool-pages-insert)
      (put-message-value (_ "innodb-buffer-pool-pages-misc") 
			 innodb-buffer-pool-pages-misc)
      (put-message-value (_ "innodb-buffer-pool-pages-total") 
			 innodb-buffer-pool-pages-total))
    (setq *report*
	(make-report
	 ;; timestamp
	 :timestamp	        (formatted-timestamp)
	 ;; wiki
	 :language              lang
	 ;; activities (boolean)
	 :dump-file-null          (zerop (sql-select-count-file lang :put put))
	 :dump-file-downloading   (string= (sql-select-file-state
					    file-dump-name :put put)
					   "start")
	 :dump-file-validating    (string= (sql-select-file-state
					    file-dump-name :put put)
					   "created")
	 :dump-file-decompressing (string= (sql-select-file-state
					    file-dump-name :put put)
					   "valid")
	 :dump-file-page-counting (and
				   (string= (sql-select-file-state
					     file-dump-xml-name :put put)
					    "start")
				   (zerop (sql-select-file-pages
					   file-dump-xml-name :put put)))
	 :dump-file-image-counting (and
				    (string= (sql-select-file-state
					      file-dump-xml-name :put put)
					     "start")
				    (zerop (sql-select-file-images
					    file-dump-xml-name :put put)))
	 :xchunk-generating      (sql-select-xchunk-wip lang :put put)
	 :ichunk-generating      (sql-select-ichunk-wip lang :put put)
	 :images-counting        (not
				  (zerop
				   (sql-select-count-images-start lang
								  :put put)))
	 :images-validating      (not
				  (zerop
				   (sql-select-count-images-created lang
								    :put put)))
	 ;; downloading
	 :file-dump-name         file-dump-name
	 :file-dump-size         (sql-select-file-size file-dump-name :put put)
	 :file-dump-xml-name     file-dump-xml-name
	 :file-dump-xml-size     (sql-select-file-size file-dump-xml-name
						       :put put)
	 ;; estimating
	 :page-count             (sql-select-file-pages file-dump-xml-name
							:put put)
	 :image-count            (sql-select-file-images file-dump-xml-name
							 :put put)
	 :chunk-count            chunk-count
	 ;; chunk generating
	 :xchunks-generated      xchunk-count
	 :xchunks-generated-progress (cond ((zerop chunk-count) 0)
					   ((> xchunk-count chunk-count) 1.0)
					   (t (float
					       (/ xchunk-count chunk-count))))
	 :ichunks-generated      ichunk-count
	 :ichunks-generated-progress (cond ((zerop chunk-count) 0)
					   ((> ichunk-count chunk-count) 1.0)
					   (t (float
					       (/ ichunk-count chunk-count))))
	 ;; ichunks
	 :ichunks-running        ichunk-running
	 :ichunks-done           ichunk-done
	 :ichunks-failed         ichunk-fail
	 :ichunks-remaining      (- ichunk-count ichunk-done ichunk-fail)
	 :ichunks-progress       (cond ((zerop chunk-count) 0)
				       ((> (+ ichunk-done ichunk-fail)
					   chunk-count) 1.0)
				       (t (float 
					   (/ (+ ichunk-done ichunk-fail)
					      chunk-count))))
	 ;; xchunks
	 :xchunks-running        xchunk-running
	 :xchunks-done           xchunk-done
	 :xchunks-failed         xchunk-fail
	 :xchunks-remaining      (- xchunk-count xchunk-done xchunk-fail)
	 :xchunks-progress       (cond ((zerop chunk-count) 0)
				       ((> (+ xchunk-done xchunk-fail)
					   chunk-count) 1.0)
				       (t (float
					   (/ (+ xchunk-done xchunk-fail)
					      chunk-count))))
	 :pages-imported         pages-imported
	 ;; images
	 :images-sum                 images-sum
	 :images-count               images-count
	 :images-count-progress      (cond ((zerop images-count) 0)
					   ((> (- images-count images-start)
					       images-count) 1.0)
					   (t (float
					       (/ (- images-count images-start)
						  images-count))))
	 :images-valid               images-valid
	 :images-valid-progress      (cond ((zerop images-sum) 0)
					   ((> images-valid images-sum) 1.0)
					   (t (float (/ images-valid
							images-sum))))
	 :images-imported            images-imported
	 :images-imported-progress   (cond ((zerop images-sum) 0)
					   ((> images-imported images-sum) 1.0)
					   (t (float (/ images-imported
							images-sum))))
	 ;; InnoDB
	 :innodb-buffer-pool-pages-total innodb-buffer-pool-pages-total
	 :innodb-buffer-pool-pages-dirty innodb-buffer-pool-pages-dirty
	 :innodb-buffer-pool-pages-dirty-progress
	 (if (zerop innodb-buffer-pool-pages-total)
	     0
	   (float (/ innodb-buffer-pool-pages-dirty
		     innodb-buffer-pool-pages-total)))
	 :innodb-buffer-pool-pages-insert innodb-buffer-pool-pages-insert
	 :innodb-buffer-pool-pages-insert-progress 
	 (if (zerop innodb-buffer-pool-pages-total)
	     0
	   (float (/ innodb-buffer-pool-pages-insert 
		     innodb-buffer-pool-pages-total)))
	 :innodb-buffer-pool-pages-misc innodb-buffer-pool-pages-misc
	 :innodb-buffer-pool-pages-misc-progress 
	 (if (zerop innodb-buffer-pool-pages-total)
	     0
	   (float (/ innodb-buffer-pool-pages-misc
		     innodb-buffer-pool-pages-total)))
       ))))

(defun windowing-system-running-p ()
  "Determining if a windowing system is running"
  (put-message-start (_ "probing running windowing system"))
  (let ((windowing-system nil))
    (with-open-stream (s (ext:run-shell-command
			  (format nil "ps -wef | ~a ~a"
				  *whereis-grep* *whereis-x*)
			  :output :stream))
      (setq windowing-system (read-line s nil nil)))
    (if (null windowing-system)
	(put-message-fail (_ "probing running windowing system"))
      (put-message-value-done (_ "probing running windowing system")
			      windowing-system))
    windowing-system))

(defun monitor-mode-gui-screen-or-text ()
  "Determining best monitor mode (gui, screen, or text)"
  (put-message-start (_ "choosing monitor mode"))
  (put-message-value (_ "monitor mode") *monitor-mode*)
  (when (eql *monitor-mode* :auto)
    (put-message-value (_ "probing for :gui") *monitor-mode*)
    (if (and (member :clx *features*)
	     (windowing-system-running-p))
	(setq *monitor-mode* :gui)
      (progn
	(put-message-value (_ "probing for :screen") *monitor-mode*)
	(if (member :screen *features*)
	    (setq *monitor-mode* :screen)
	  (progn
	    (put-message-value (_ "falling back on :text") *monitor-mode*)
	    (setq *monitor-mode* :text))))))
  (put-message-value-done (_ "choosing monitor mode") *monitor-mode*)
  *monitor-mode*)

(defun monitor-mode-gui (&key (put nil))
  "Running the monitor in gui mode"
  (put-message-start (_ "attempting gui mode"))
  (let* ((languages   (union *mirror-languages*
			     (sql-select-languages :put put)
			     :test #'string=))
	 (host        '())
	 (display     (xlib:open-display       host))
	 (screen      (first (xlib:display-roots display)))
	 (black       (xlib:screen-black-pixel screen))
	 (white       (xlib:screen-white-pixel screen))
	 (root-window (xlib:screen-root        screen))
	 ;; take the most preferred font available
	 (favorite-font-available
	              (first
		       (reduce #'append
			       (mapcar #'(lambda (x) (xlib:list-font-names
						      display x))
				       *monitor-gui-font-preference-list*))))
	 (font        (xlib:open-font display  favorite-font-available))
	 (gcontexts   (make-hash-table :test 'equal))
	 ;; X11 colors (see `/etc/X11/rgb.txt')
	 (gtext       (xlib:alloc-color
		       (xlib:window-colormap root-window) "black"))
	 (gback       (xlib:alloc-color
		       (xlib:window-colormap root-window) "LightYellow"))
	 (gfill       (xlib:alloc-color
		       (xlib:window-colormap root-window) "RoyalBlue"))
	 (gedge       (xlib:alloc-color
		       (xlib:window-colormap root-window) "YellowGreen"))
	 (my-windows  (mapcar
		       #'(lambda (x)
			   (xlib:create-window
			    :parent        root-window
			    :x             0
			    :y             0
			    :width         (* (+ 1 10 7 26 4 2) 
					      (xlib:max-char-width font))
			    :height        (* (+ 22 1)         ; lines 
					      (floor (* 1.4    ; line spacing
							(xlib:font-ascent font))))
			    :background    gback
			    :event-mask    (xlib:make-event-mask 
					    :button-press
					    :exposure
					    :key-press
					    :structure-notify)))
		       (if *monitor-gui-one-window-per-language*
			   languages
			 (list 1)))))
    (setf (gethash 'text  gcontexts)
	  (xlib:create-gcontext
	   :drawable      root-window
	   :font          font
	   :background    gback
	   :foreground    gtext))
    (setf (gethash 'fill gcontexts)
	  (xlib:create-gcontext
	   :drawable      root-window
	   :background    gback
	   :foreground    gfill))
    (setf (gethash 'edge  gcontexts)
	  (xlib:create-gcontext
	   :drawable      root-window
	   :background    gback
	   :foreground    gedge))
    (loop
      for lang in (if *monitor-gui-one-window-per-language*
		      languages
		    (list 1))
      as  wind in my-windows
      do
      (xlib:change-property wind
			    :wm_name   "WP-MIRROR - Monitor Mode"
			    :string    8
			    :transform #'char-code))
    (unwind-protect
	(loop named gui       ; user terminates with button-press (see below)
	  do
	  (loop
	    for lang in languages
	    as  wind in my-windows by (if *monitor-gui-one-window-per-language*
					  #'cdr
					#'identity)
	    do
	      (monitor-data-collect lang :put put)
	      (debug-message-start "event-case")
	      (xlib:map-window wind)
	      (xlib:event-case (display :discard-p t
					:force-output-p t
			                :timeout 
				(if *monitor-gui-one-window-per-language*
				    (max (ceiling *monitor-poll-sec*
						  (length languages))
					 1)
				   *monitor-poll-sec*))
			       (:button-press () nil)
			       (:destroy-window ()              ; <termination>
						(return-from gui t) t)
			       (:exposure (count)
					  (when (zerop count)
					    (monitor-mode-gui-show wind
								   gcontexts))
					  nil)
			       (:key-press ()                   ; <termination>
					   (return-from gui t) t)
			       )
	      (xlib:send-event wind :exposure nil)
	      (debug-message-value "event-case" (formatted-timestamp))))
      (put-message-start (_ "cleanup"))
      (xlib:display-finish-output display)
      (loop
	for wind in my-windows
	do
	(xlib:destroy-window wind))
      (loop named clean-up
	;for i
	while (xlib:discard-current-event display))
      (put-message-done (_ "cleanup"))
      (xlib:close-display display)
      (put-message-done  (_ "attempting gui mode")))))

(defun monitor-mode-gui-show (my-window gcontexts)
  "Showing status in gui format"
  (let* ((w         my-window)
	 (ww        (xlib:drawable-width   w))
	 (wh        (xlib:drawable-height  w))
	 (wr        (xlib:drawable-root    w))
	 (g         gcontexts)
	 (gt        (gethash 'text  g))
	 (f         (xlib:gcontext-font   gt))
	 (fa        (xlib:font-ascent      f)))
    ;; clear window first (lest new stuff get written on top of old stuff)
    (xlib:clear-area w)
    ;; timestamp
    (let* ((str (report-timestamp *report*))
	   (x   (- ww (xlib:text-extents f str)))
	   (y   (* 1 fa)))
      (xlib:draw-glyphs w gt x y str))

    ;; language code
    (monitor-mode-gui-show-line w g  1
      :d (string-upcase (report-language *report*)))

    ;; dumpfile
    (monitor-mode-gui-show-line w g  2
      :a "Dumpfile:"
      :d (when (null (report-dump-file-null     *report*))
	   (report-file-dump-name *report*)))

    (monitor-mode-gui-show-line w g  3
      :b "downloading"
      :c (format-integer-for-human (report-file-dump-size *report*))
      :d (cond ((report-dump-file-null        *report*) 0)
	       ((report-dump-file-downloading *report*) :pulse)
	       (t 1)))

    (monitor-mode-gui-show-line w g  4
      :b "validating"
      :d (if (report-dump-file-validating *report*) :pulse
	   (if (> (report-file-dump-size *report*) 0) 1 0)))
    
    (monitor-mode-gui-show-line w g  5
      :b "decompressing"
      :c (format-integer-for-human (report-file-dump-xml-size *report*))
      :d (if (report-dump-file-decompressing *report*) :pulse 
	   (if (> (report-file-dump-xml-size *report*) 0) 1 0)))

    ;; estimates
    (monitor-mode-gui-show-line w g  6
      :a "Estimates:")

    (monitor-mode-gui-show-line w g  7
      :b "counting pages"
      :c (format nil "~10:d" (report-page-count *report*))
      :d (if (report-dump-file-page-counting *report*) :pulse
	   (if (> (report-page-count *report*) 0) 1 0)))

    (monitor-mode-gui-show-line w g  8
      :b "counting image files"
      :c (format nil "~10:d" (report-image-count *report*))
      :d (if (report-dump-file-image-counting *report*) :pulse 
	   (if (> (report-image-count *report*) 0) 1 0)))

    ;; chunks
    (monitor-mode-gui-show-line w g  9
      :a "Chunks:")

    (monitor-mode-gui-show-line w g 10
      :b "generating xchunks"
      :c (format nil "~10:d" (report-xchunks-generated *report*))
      :d (report-xchunks-generated-progress *report*))

    (monitor-mode-gui-show-line w g 11
      :b "generating ichunks"
      :c (format nil "~10:d" (report-ichunks-generated *report*))
      :d (report-ichunks-generated-progress *report*))

    ;; images
    (monitor-mode-gui-show-line w g 12
      :a "Images:")

    (monitor-mode-gui-show-line w g 13
      :b "downloading"
      :c (format nil "~10:d" (+ (report-ichunks-done      *report*)
				(report-ichunks-failed    *report*)))
      :d (report-ichunks-progress *report*))

    (monitor-mode-gui-show-line w g 14
      :b "counting"
      :c (format nil "~10:d" (report-images-sum  *report*))
      :d (report-images-count-progress *report*))

    (monitor-mode-gui-show-line w g 15
      :b "validating"
      :c (format nil "~10:d" (report-images-valid  *report*))
      :d (report-images-valid-progress *report*))

    (monitor-mode-gui-show-line w g 16
      :b "importing"
      :c (format nil "~10:d" (report-images-imported  *report*))
      :d (report-images-imported-progress *report*))

    ;; pages
    (monitor-mode-gui-show-line w g 17
      :a "Pages:")

    (monitor-mode-gui-show-line w g 18
      :b "importing          "
;;      :c (format nil "~10:d" (+ (report-xchunks-done      *report*)
;;				(report-xchunks-failed    *report*)))
      :c (format nil "~10:d" (report-pages-imported    *report*))
      :d (report-xchunks-progress *report*))

    ;; InnoDB
    (monitor-mode-gui-show-line w g 19
      :a "InnoDB:")

    (monitor-mode-gui-show-line w g 20
      :b "buffer pool pages dirty" 
      :c (format nil "~10:d" (report-innodb-buffer-pool-pages-dirty
				      *report*))
      :d (report-innodb-buffer-pool-pages-dirty-progress *report*))

    (monitor-mode-gui-show-line w g 21
      :b "buffer pool pages insert"
      :c (format nil "~10:d" (report-innodb-buffer-pool-pages-insert *report*))
      :d (report-innodb-buffer-pool-pages-insert-progress *report*))

    (monitor-mode-gui-show-line w g 22
      :b "buffer pool pages misc"
      :c (format nil "~10:d" (report-innodb-buffer-pool-pages-misc *report*))
      :d (report-innodb-buffer-pool-pages-misc-progress *report*))

    ))

(defun monitor-mode-gui-show-line (w gcontexts l &key 
				     (a nil) (b nil) (c nil) (d nil))
  "Showing one line for monitor screen mode"
  ;; we very carefully define (x,y,w,h) for each cell (a,b,c,d,e)
  (let* ((d        (if (numberp d)
		       (max 0 (min d 1.0))   ; [0,1]
		     d))                     ; string
	 (e        (when (numberp d)
		     (format nil "~5,2f%" (* d 100)))) ; percent progress
	 (g        gcontexts)
	 (gt       (gethash 'text  g))
	 (gf       (gethash 'fill g))
	 (ge       (gethash 'edge  g))
	 (f        (xlib:gcontext-font  gt))
	 (fa       (xlib:font-ascent     f))
	 (fd       (xlib:font-direction  f)) ; used for I18N
	 (fmcw     (xlib:max-char-width  f)) 
	 (fiws     (xlib:text-extents f " ")); inter-word space
	 (lh       (floor (* 1.4 fa)))       ; line spacing
	 (aw       (*  1 fmcw))              ; col a width
	 (bw       (* 10 fmcw))              ; col b width
	 (cw       (*  7 fmcw))
	 (dw       (* 26 fmcw))
	 (dw-bar   (cond ((numberp d) 
			  (floor (* dw d))) ; progress bar - filled
			 ((eql d :pulse)
			  (floor (* dw (/ (get-decoded-time) 60)))))) ; pulsed
	 (ew       (*  4 fmcw))
	 (ax       (+       fiws))           ; col a x
	 (bx       (+ ax aw fiws))           ; col b x
	 (cx       (+ bx bw fiws))
	 (cx-just  (when (stringp c)
		     (- (+ cx cw) (xlib:text-extents f c)))) ; right justify
	 (dx       (+ cx cw fiws))
	 (ex       (+ dx dw fiws))
	 (ex-just  (when (stringp e)
		     (- (+ ex ew) (xlib:text-extents f e)))) ; right justify
	 (ah       lh)                       ; col a height
	 (bh       lh)                       ; col b height
	 (ch       lh)
	 (dh       lh)
	 (dh-bar   fa)        ; progress bar thinner for readability
	 (eh       lh)
	 (ay       (* l lh))                 ; col a y
	 (by       ay)                       ; col b y
	 (cy       ay)
	 (dy       ay)
	 (dy-bar   (- ay fa)) ; progress bar thinner for readablity
	 (ey       ay))
    (unless (null a) (xlib:draw-glyphs    w gt ax      ay a))
    (unless (null b) (xlib:draw-glyphs    w gt bx      by b))
    (unless (null c) (xlib:draw-glyphs    w gt cx-just cy c))       ; justify
    (cond ((stringp d) (xlib:draw-glyphs  w gt dx      dy d))
	  ((and (numberp d) (eql d 1))
	   (xlib:draw-rectangle w gf dx dy-bar dw-bar dh-bar   t))  ; fill
	  ((or (numberp d) (eql d :pulse))
	   (xlib:draw-rectangle w ge dx dy-bar dw     dh-bar nil)   ; edge
	   (xlib:draw-rectangle w gf dx dy-bar dw-bar dh-bar   t))  ; fill
	  (t nil))
    (cond ((eql d :pulse) (xlib:draw-glyphs    w gt ex      ey "pulse"))
          ((and (stringp e) (string= e " 0.00%")) nil)
          ((and (stringp e) (string= e "100.00%"))
	                  (xlib:draw-glyphs    w gt ex      ey "done"))
          ((stringp e)    (xlib:draw-glyphs    w gt ex-just ey e))) ; justify
    ))

(defun monitor-mode-screen (&key (put nil))
  "Running the monitor in text mode"
  (put-message-start (_ "attempting screen mode"))
  (unwind-protect
      (screen:with-window
       (defparameter *monitor-screen-stream* screen:*window*)
       (multiple-value-bind (height width)
	   (screen:window-size *monitor-screen-stream*)
	 (defparameter *monitor-screen-height* height)
	 (defparameter *monitor-screen-width*  width)
	 (when (and (> *monitor-screen-height* 0)
		    (> *monitor-screen-width*  0))
	   (loop named screen    ; user terminates with Control-C 
	     do
	     (loop
	       for lang in (union *mirror-languages*
				  (sql-select-languages :put put)
				  :test #'string=)
	       do
	         (monitor-data-collect lang :put put)
		 (monitor-mode-screen-show)
		 (sleep *monitor-poll-sec*))))))
    (put-message-done (_ "attempting screen mode"))))

(defun monitor-mode-screen-show ()
  "Showing status in screen format"
  (let ((w  *monitor-screen-stream*)
	(wh (1- *monitor-screen-height*))
	(ww (1- *monitor-screen-width* )))

    ;; clear screen first
    (screen:window-cursor-off  w)
    (screen:clear-window       w)

    ;; borders
    (monitor-mode-screen-show-box)

    ;; caption
    (monitor-mode-screen-show-line  0
      :a "WP-MIRROR"
      :d "MONITOR MODE")
    (screen:set-window-cursor-position      w  0 (- ww 1
			(length (report-timestamp *report*))))
    (princ (report-timestamp *report*) w)

    ;; language code
    (monitor-mode-screen-show-line  1
      :d (string-upcase (report-language *report*)))

    ;; dump file
    (monitor-mode-screen-show-line  2
      :a "Dumpfile:            "
      :d (if (null (report-dump-file-null *report*))
	     (report-file-dump-name *report*)
	   (format nil "~awiki" (language-to-file-name
				 (report-language *report*)))))
	   
    (monitor-mode-screen-show-line  3
      :b "downloading        "
      :c (format nil "~10@a" (format-integer-for-human
			      (report-file-dump-size *report*)))
      :d (cond ((report-dump-file-null        *report*) 
		(format-progress-bar-string 0))
	       ((report-dump-file-downloading *report*) :pulse)
	       (t (format-progress-bar-string 1))))

    (monitor-mode-screen-show-line  4
      :b "validating         " 
      :d (if (report-dump-file-validating *report*)
	     (format-progress-bar-pulse)
	   (if (> (report-file-dump-size *report*) 0)
	       (format-progress-bar-string 1)
	     (format-progress-bar-string 0))))

    (monitor-mode-screen-show-line  5
      :b "decompressing      " 
      :c (format nil "~10@a" (format-integer-for-human 
			      (report-file-dump-xml-size *report*)))
      :d (if (report-dump-file-decompressing *report*)
	     (format-progress-bar-pulse)
	   (if (> (report-file-dump-xml-size *report*) 0)
	       (format-progress-bar-string 1)
	     (format-progress-bar-string 0))))

    ;; estimates
    (monitor-mode-screen-show-line  6
      :a "Estimates:           ")

    (monitor-mode-screen-show-line  7
      :b "counting pages     " 
      :c (format nil "~10:d" (report-page-count *report*))
      :d (if (report-dump-file-page-counting *report*)
	     (format-progress-bar-pulse)
	   (if (> (report-page-count *report*) 0)
	       (format-progress-bar-string 1)
	     (format-progress-bar-string 0))))

    (monitor-mode-screen-show-line  8
      :b "counting image files"
      :c (format nil "~10:d" (report-image-count *report*))
      :d (if (report-dump-file-image-counting *report*)
	     (format-progress-bar-pulse)
	   (if (> (report-image-count *report*) 0)
	       (format-progress-bar-string 1)
	     (format-progress-bar-string 0))))

    ;; chunks
    (monitor-mode-screen-show-line  9
      :a "Chunks:              ")

    (monitor-mode-screen-show-line 10
      :b "generating xchunks "
      :c (format nil "~10:d" (report-xchunks-generated *report*))
      :d (format-progress-bar-string
	  (report-xchunks-generated-progress *report*)))

    (monitor-mode-screen-show-line 11
      :b "generating ichunks "
      :c (format nil "~10:d" (report-ichunks-generated *report*))
      :d (format-progress-bar-string 
	  (report-ichunks-generated-progress *report*)))

    ;; images
    (monitor-mode-screen-show-line 12
      :a "Images:              ")

    (monitor-mode-screen-show-line 13
      :b "downloading        "
      :c (format nil "~10:d" (+ (report-ichunks-done      *report*)
				(report-ichunks-failed    *report*)))
      :d (format-progress-bar-string (report-ichunks-progress *report*)))

    (monitor-mode-screen-show-line 14
      :b "counting           "
      :c (format nil "~10:d" (report-images-sum *report*))
      :d (format-progress-bar-string
	  (report-images-count-progress *report*)))

    (monitor-mode-screen-show-line 15
      :b "validating         "
      :c (format nil "~10:d" (report-images-valid *report*))
      :d (format-progress-bar-string
	  (report-images-valid-progress *report*)))
;    (when (string= (report-language *report*) "zu") (die))

    (monitor-mode-screen-show-line 16
      :b "importing           "
      :c (format nil "~10:d" (report-images-imported *report*))
      :d (format-progress-bar-string
	  (report-images-imported-progress *report*)))

    ;; pages
    (monitor-mode-screen-show-line 17
      :a "Pages:               ")

    (monitor-mode-screen-show-line 18
      :b "importing          "
;;      :c (format nil "~10:d" (report-xchunks-done      *report*))
      :c (format nil "~10:d" (report-pages-imported      *report*))
      :d (format-progress-bar-string (report-xchunks-progress *report*)))

    ;; InnoDB
    (monitor-mode-screen-show-line 19
      :a "InnoDB:              ")

    (monitor-mode-screen-show-line 20
      :b "buffer pool dirty  " 
      :c (format nil "~10:d" (report-innodb-buffer-pool-pages-dirty *report*))
      :d (format-progress-bar-string
	  (report-innodb-buffer-pool-pages-dirty-progress *report*)))

    (monitor-mode-screen-show-line 21
      :b "buffer pool insert "
      :c (format nil "~10:d" (report-innodb-buffer-pool-pages-insert *report*))
      :d (format-progress-bar-string
	  (report-innodb-buffer-pool-pages-insert-progress *report*)))

    (monitor-mode-screen-show-line 22
      :b "buffer pool misc   "
      :c (format nil "~10:d" (report-innodb-buffer-pool-pages-misc *report*))
      :d (format-progress-bar-string
	  (report-innodb-buffer-pool-pages-misc-progress *report*)))
    
    ;; park cursor in corner
    (screen:set-window-cursor-position      w wh ww)
))

(defun monitor-mode-screen-show-box ()
  "Showing box bordering the screen"
  (let ((w  *monitor-screen-stream*)
	(wh (1- *monitor-screen-height*))
	(ww (1- *monitor-screen-width* )))
    ;; edges
    (loop 
      for i of-type integer from 0 to ww do 
      (screen:set-window-cursor-position    w  0  i) (princ "-" w)
      (screen:set-window-cursor-position    w wh  i) (princ "-" w))
    (loop 
      for i of-type integer from 1 to wh do
      (screen:set-window-cursor-position    w  i  0) (princ "|" w)
      (screen:set-window-cursor-position    w  i ww) (princ "|" w))
    ;; corners
    (screen:set-window-cursor-position      w  0  0) (princ "+" w)
    (screen:set-window-cursor-position      w wh  0) (princ "+" w)
    (screen:set-window-cursor-position      w  0 ww) (princ "+" w)
    (screen:set-window-cursor-position      w wh ww) (princ "+" w)))

(defun monitor-mode-screen-show-line (y &key (a nil) (b nil) (c nil) (d nil))
  "Showing one line for monitor screen mode"
  ;; left-align four strings (a,b,c,d) on columns (a-col,b-col,c-col,d-col)
  (let ((w  *monitor-screen-stream*)
	(a-col   2)
	(b-col   4)
	(c-col  23)
	(d-col  34))
    (unless (null a) (screen:set-window-cursor-position w y a-col) (princ a w))
    (unless (null b) (screen:set-window-cursor-position w y b-col) (princ b w))
    (unless (null c) (screen:set-window-cursor-position w y c-col) (princ c w))
    (unless (null d) (screen:set-window-cursor-position w y d-col) (princ d w))
    ))

(defun monitor-mode-text (&key (put nil))
  "Running the monitor in text mode"
  (loop                  ; user terminates with Control-C 
    do
    (loop 
      for lang in (union *mirror-languages*
			 (sql-select-languages :put put)
			 :test #'string=)
      do
        (put-message-start (_"monitor mode text loop"))
	(monitor-data-collect lang :put put)
	(monitor-mode-text-show)
	(sleep *monitor-poll-sec*))))

(defun monitor-mode-text-show ()
  "Showing status in text format"
  (put-message-value (report-timestamp               *report*)
		     (format nil "~awiki"
			     (language-to-file-name
			      (report-language       *report*))))
  (put-message-value (_ "dump file name")
		     (report-file-dump-name          *report*))
  (put-message-value (_ "downloading dump file")
		     (report-dump-file-downloading   *report*))
  (put-message-value (_ "validating dump file")
		     (report-dump-file-validating    *report*))
  (put-message-value (_ "decompressing dump file")
		     (report-dump-file-decompressing *report*))
  (put-message-value (_ "counting pages in dump file")
		     (report-dump-file-page-counting *report*))
  (put-message-value (_ "counting images in dump file")
		     (report-dump-file-image-counting *report*))
  (when (> (report-page-count *report*) 0)
    (put-message-value (_ "pages in dump file")
		       (report-page-count            *report*))
    (put-message-value (_ "xchunks expected")  
		       (report-chunk-count           *report*))
    (put-message-value (_ "xchunks generating/generated")
		       (format nil "~d/~d"
			       (report-xchunk-generating     *report*)
			       (report-xchunks-generated     *report*))))
  (when (> (report-ichunks-generated *report*) 0)
    (put-message-value (_ "ichunks generating/generated")
		       (format nil "~d/~d"
			       (report-ichunk-generating     *report*)
			       (report-ichunks-generated     *report*)))
    (put-message-value (_ "ichunks running")   
		       (report-ichunks-running       *report*))
    (put-message-value (_ "ichunks done/failed/remaining")
		       (format nil "~d/~d/~d" 
			       (report-ichunks-done      *report*)
			       (report-ichunks-failed    *report*)
			       (report-ichunks-remaining *report*)))
    (put-message-value (_ "ichunks progress") 
		       (format-progress-bar-string
			(report-ichunks-progress     *report*)))
    (put-message-value (_ "xchunks running")
		       (report-xchunks-running       *report*))
    (put-message-value (_ "xchunks done/failed/remaining")
		       (format nil "~d/~d/~d" 
			       (report-xchunks-done      *report*)
			       (report-xchunks-failed    *report*)
			       (report-xchunks-remaining *report*)))
    (put-message-value (_ "pages imported") 
		       (report-pages-imported            *report*))
    (put-message-value (_ "xchunks progress") 
		       (format-progress-bar-string
			(report-xchunks-progress         *report*)))

    (put-message-value (_ "counting download images")
		       (report-images-counting *report*))
    (put-message-value (_  "images counting")
		       (format nil "~d"
			       (report-images-sum *report*)))
    (put-message-value (_  "images counting")
		       (format-progress-bar-string
			(report-images-count-progress *report*)))
    (put-message-value (_  "images validating")
		       (format nil "~d"
			       (report-images-valid *report*)))
    (put-message-value (_  "images validating")
		       (format-progress-bar-string
			(report-images-valid-progress *report*)))
    (put-message-value (_  "images imported")
			(report-images-imported *report*))
    (put-message-value (_  "images imported")
		       (format-progress-bar-string
			(report-images-imported-progress *report*)))

    (put-message-value (_ "InnoDB buffer pages dirty")
		       (format-progress-bar-string
			(report-innodb-buffer-pool-pages-dirty-progress
				 *report*)))
    (put-message-value (_ "InnoDB buffer pages insert")
		       (format-progress-bar-string
			(report-innodb-buffer-pool-pages-insert-progress
				 *report*)))
    (put-message-value (_ "InnoDB buffer pages misc")
		       (format-progress-bar-string
			(report-innodb-buffer-pool-pages-misc-progress
				 *report*)))
    ))

(defun main-monitor (&key (main-mode :monitor) (put nil))
  "Purpose: Main function when wp-mirror is in Monitor Mode."
  ;; 0) Initialize
  (put-message (_ "-----initializing-begin-----------------------"))
  (process-configuration-files-or-die)
  (count-cpu)
  (put-parameters)
  (put-message (_ "-----initializing-done------------------------"))
  ;; 1) Assert prerequisites
  (put-message (_ "-----asserting-prerequisite-software-begin----"))
  (assert-test-set-or-die '(assert-mediawiki-localsettings-p
			    assert-database-administrator-credentials-p
			    assert-dbms-mysql-p
			    assert-database-administrator-account-p
			    assert-database-wpmirror-or-create
			    ))
  (put-message (_ "-----asserting-prerequisite-software-done-----"))
  ;; 2) Determine monitor mode (:auto, :gui, :screen, :text)
  (put-message (_ "-----monitor-mode-begin-----------------------"))
  (monitor-mode-gui-screen-or-text)
  (case *monitor-mode*
    (:gui    (monitor-mode-gui    :put put))
    (:screen (monitor-mode-screen :put put))
    (:text   (monitor-mode-text   :put put)))
  (put-message (_ "-----monitor-mode-done------------------------")))


;;;;--------------------------------------------------------------------------+
;;;; Main:                                                                    |
;;;;   Determine main mode of operation :first-mirror, :next-mirror, :monitor |
;;;;--------------------------------------------------------------------------+


(defun main ()
  (declare (optimize (safety 3)))
  (debug-message (_ "Hello, Dr. Miller."))
  ;; 0) Initialize parameters
  (process-command-line-arguments-or-die)
  ;; 1) Mirror or monitor
  (let ((main-mode (main-mode-mirror-or-monitor))
	(put       (not (zerop *debug-level*))))
    (case main-mode
      ((:first-mirror :next-mirror)
       (main-mirror  :main-mode main-mode :put put))
      (:monitor                     
       (main-monitor :main-mode main-mode :put put))))
  (debug-message (_ "Goodbye, Dr. Miller."))
  (die))

(main)
